/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "stdafx.h"

using namespace std;
using namespace boost::filesystem;

// Set the current version of CamMark
static Version camMarkVersion(1, 1, 2);

void prepareFilenames(Scenario &scenario,
	string &scenario_file, string &src_file, string &out_file)
{
	// Get canonical scenario file path
	path scenario_path(scenario_file);
	scenario_path = canonical(scenario_path);
	scenario_file = scenario_path.make_preferred().string();

	// Should we read the source filename from the scenario XML?
	if(src_file == "xml")
	{
		cout << "Using source file specified in scenario..." << endl;
		
		if(!scenario.sourceFile.empty())
		{
			path scenario_src_path(scenario.sourceFile);
			if(scenario_src_path.is_relative())
			{
				scenario_src_path = scenario_path.parent_path() / scenario_src_path;
				src_file = scenario_src_path.make_preferred().string();
			}
		}
		else throw runtime_error(
			"The scenario does not specify a source file (or it is empty).");
	}

	// Should we read the output filename from the scenario XML?
	if(out_file == "xml")
	{
		cout << "Using output file specified in scenario..." << endl;

		if(!scenario.outputFile.empty())
		{
			path scenario_out_path(scenario.outputFile);
			if(scenario_out_path.is_relative())
			{
				scenario_out_path = scenario_path.parent_path() / scenario_out_path;
				out_file = scenario_out_path.make_preferred().string();
			}
		}
		else out_file.clear();
	}

	// If no output file name was specified,
	// generate it from the source file name.
	if(out_file.empty())
	{
		cout << "Generating output file name from source file name..." << endl;
		path src_path(src_file), out_path;
		if(src_path.has_parent_path())
			 out_path = src_path.parent_path() / src_path.stem();
		else out_path = src_path.filename();
		out_file = out_path.make_preferred().string() + "_sim.avi";
	}

	// Check existance of source file
	path src_path(src_file);
	if(!exists(src_path)) throw runtime_error(
		"Source file '" + src_file + "' does not exist.");

	// Get output file's parent directory
	path out_path(out_file), out_path_parent;
	if(out_path.has_parent_path())
		 out_path_parent = out_path.parent_path();
	else out_path_parent = current_path();

	// Check exsistance of output file's parent directory
	// and create it if not existant.
	if(!exists(out_path_parent))
	{
		cout << "Creating the specified output file's parent directory, which does not exist." << endl;
		create_directories(out_path_parent);
	}

	// Get the canonical path to the source file
	src_path = canonical(src_path);
	src_file = src_path.make_preferred().string();

	// Get the canonical path to the output file
	out_path_parent = canonical(out_path_parent);
	out_path = out_path_parent / out_path.filename();
	out_file = out_path.make_preferred().string();

	// Output used files
	cout << "Scenario file: " << scenario_file << endl;
	cout << "Source file:   " << src_file << endl;
	cout << "Output file:   " << out_file << endl;
	cout << endl;
}

void applyCamMark(string scenario_file, string src_file, string out_file)
{
	// Read and parse the scenario XML
	Scenario scenario(camMarkVersion);
	scenario.readXMLFromFile(scenario_file);

	prepareFilenames(scenario,
		scenario_file, src_file, out_file);

	// Output scenario overview
	cout << endl << "=== SIMULATION SCENARIO ==========================" << endl << endl;
	scenario.printToStream(cout);
	cout << endl << "==================================================" << endl << endl;

	cout << "Starting simulation... " << endl;
	CamMark cammark;
	float progress;
	
	cammark.startSimulation(src_file, out_file, scenario);
	
	bool lastMsgWasProgress = false;
	while(!cammark.doneProcessing())
	{
		if(cammark.hasPendingLogMessages())
		{
			if(lastMsgWasProgress) cout << endl;
			cammark.showLogMessages();
			
			lastMsgWasProgress = false;
		}

		cammark.rethrowIfPending();
		
		progress = cammark.getProgress();
		if(progress > 0)
		{
			if(lastMsgWasProgress) cout << "\r";
			cout << "progress: " << std::fixed << std::setprecision(2) << progress * 100.0f << "%";
			std::cout.flush();
			
			lastMsgWasProgress = true;
		}

		boost::this_thread::sleep_for( boost::chrono::milliseconds(250) );
	}
	cout << endl;

	cammark.finishSimulation();
	cammark.showLogMessages();
 }

int _tmain(int argc, _TCHAR* argv[])
{
	// Display version info
	cout << "### This is CamMark version " << camMarkVersion << "." << endl;
	cout << endl;

	// Find out the program name
	string program_path, program_name;
	Util::splitFilePathAndName(argv[0],	program_path, program_name);

	if(argc < 3)
	{
		// Display usage
		cout << "Usage: " << program_name << " <scenario-xml> <source-video> [<output-video>]" << endl;
		return 1;
	}

	// Store arguments
	string scenario_file = argv[1];
	string src_file = argv[2];
	string out_file;
	
	// Only try to access the fourth element
	// (output file name) if we know it's there.
	if(argc >= 4) out_file = argv[3];

	try
	{
		// Apply CamMark with given parameters
		applyCamMark(scenario_file, src_file, out_file);
	}
	catch(exception &e)
	{
		cerr << "An error occurred:" << endl;
		cerr << e.what() << endl;
		return 1;
	}
	catch(...)
	{
		cerr << "Sorry, but an unknown error occurred." << endl;
		return 2;
	}

	return 0;
}

