/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "stdafx.h"

#include "CamMark_Calibration.h"
#include "CamMark_Simulation.h"

using namespace std;

// Set the current version of CamMarkCalibration
static Version camMarkCalibVersion(1, 1, 2);

void calibSpatial(string ref_file, string copy_file,
				  bool checkerboardPattern, bool assumeLensDistortion,
				  string result_file)
{
	cout << "Starting spatial calibration... " << endl;

	float progress = 0.0 ;
	Calibration_Spatial spatial;
	
	spatial.startRegistration(ref_file, copy_file,
		checkerboardPattern, assumeLensDistortion,
		result_file);
	
	unsigned int count = 0 ;

	bool lastMsgWasProgress = false;
	while( !spatial.doneProcessing() )
	{
		if(spatial.hasPendingLogMessages())
		{
			if(lastMsgWasProgress) cout << endl;
			spatial.showLogMessages();
			
			lastMsgWasProgress = false;
		}

		spatial.showLogMessages();

		if(0 == count % 10)
		{
			progress = spatial.getProgress() ;

			if(lastMsgWasProgress) cout << "\r";
			cout << "Progress: " << std::fixed << std::setprecision(2) << progress * 100.0f << "%";
			std::cout.flush();

			lastMsgWasProgress = true;
		}

		boost::this_thread::sleep_for( boost::chrono::milliseconds(250) ) ;
		count++ ;
	}
	cout << endl;
	
	spatial.finishRegistration();
	spatial.showLogMessages();
}

void calibAGC(string ref_file, string copy_file,
			  Scenario::AGCMethod method, string result_file)
{
	string ref_file_txt = ref_file + "_luminance.txt";
	string copy_file_txt = copy_file + "_luminance.txt";

	float progress;
	Calibration_AGC agc(method);
		
	if(agc.readRefLuminancesFromFile(ref_file_txt))
		cout << "Using stored luminance values for the REFERENCE video." << endl;
	else
	{
		agc.openRefVideo(ref_file);
		cout << "Extracting luminance values from REFERENCE video..." << endl;
		while(agc.readRefVideo(progress))
			cout << "\rprogress: " << std::fixed << std::setprecision(2) << progress * 100.0f << "%";
		cout << endl << "Saving extracted luminance values for future calls." << endl;
		agc.writeRefLuminancesToFile(ref_file_txt);
	}
		
	if(agc.readCopyLuminancesFromFile(copy_file_txt))
		cout << "Using stored luminance values for the COPY video." << endl;
	else
	{
		agc.openCopyVideo(copy_file);
		cout << "Extracting luminance values from COPY video..." << endl;
		while(agc.readCopyVideo(progress))
			cout << "\rprogress: " << std::fixed << std::setprecision(2) << progress * 100.0f << "%";
		cout << endl << "Saving extracted luminance values for future calls." << endl;
		agc.writeCopyLuminancesToFile(copy_file_txt);
	}
		
	cout << "Starting AGC calibration... " << endl;

	agc.startRegistration();
	
	while(agc.doRegistrationStep(progress, 10000))
	{
		cout << "\rprogress: " << std::fixed << std::setprecision(2) << progress * 100.0f << "%";
		std::cout.flush();
	}
	cout << endl;

	agc.finishRegistration(result_file);
}

void calibTemporal(string ref_file)
{
	cout << "Starting temporal calibration... " << endl;

	float progress ;
	Calibration_Temporal temporal;
	
	temporal.startRegistration(ref_file);
	while( !temporal.doneProcessing() )
	{
		if(temporal.hasPendingLogMessages())
		{
			cout << endl;
			temporal.showLogMessages();
		}
		
		progress = temporal.getProgress() ;
		cout << "\rprogress: " << std::fixed << std::setprecision(2) << progress * 100.0f << "%";
		std::cout.flush();

		boost::this_thread::sleep_for( boost::chrono::milliseconds(250) ) ;
	}
	cout << endl;
	
	temporal.finishRegistration();
	temporal.showLogMessages();
}

void calibTemporalEmbed(string ref_file, string result_file,
						string pattern_file, bool createLuminanceTransferVideo)
{    
	cout << "Starting marker embedding... " << endl ;
	
	float progress ;
	Calibration_Embed embed ;
	
	embed.startRegistration(ref_file, result_file,
		pattern_file, createLuminanceTransferVideo) ;

	while( !embed.doneProcessing() )
	{
		if(embed.hasPendingLogMessages())
		{
			cout << endl;
			embed.showLogMessages() ;
		}

		progress = embed.getProgress() ;
		cout << "\rprogress: " << std::fixed << std::setprecision(2) << progress * 100.0f << "%";
		std::cout.flush();
		
		boost::this_thread::sleep_for( boost::chrono::milliseconds(250) ) ;
	}    
	cout << endl;

	embed.finishRegistration();
	embed.showLogMessages();
}

template<int> void displayWarningIf32Bit()
{	/* Do nothing by default */	}

template<> void displayWarningIf32Bit<4>() 
{	/* If sizeof(size_t) is 4, we compiled in an 32-bit environment */
	cout << "WARNING: The CamMark calibration tool should only be used in its 64-bit version." << endl;
	cout << "         As calibration operations require up to 6 GB of main memory, the 32-bit" << endl;
	cout << "         version is very likely throw a bad_alloc exception (out of memory)." << endl;
	cout << endl;
}

int _tmain(int argc, _TCHAR* argv[])
{
	// Display version info
	cout << "### This is CamMarkCalibration version " << camMarkCalibVersion << "." << endl;
	cout << endl;

	// Display a warning if the 32bit version of this program is used
	displayWarningIf32Bit<sizeof(size_t)>();
	
	// Find out the program name
	string program_path, program_name;
	Util::splitFilePathAndName(argv[0],	program_path, program_name);

	if(argc < 2)
	{
		// Display usage
		cout << "Usage: " << program_name << " <mode> <ref-file> <copy-file> <output-file>" << endl;
		cout << "'temporal' and 'temporal-embed' modes require different arguments." << endl;
		cout << "<mode> can be one of:" << endl;
		cout << "* 'spatial', 'spatial-cb', 'spatial-lens', 'spatial-cb-lens' (spatial calib.)" << endl;
		cout << "* 'spatial-neutral' (create neutral spatial calib.; no file parameters)" << endl;
		cout << "* 'agc-fix', 'agc-perc' (automatic gain control calibration)" << endl;
		cout << "* 'temporal', 'temporal-embed' (temporal calibration)" << endl;
		return 1;
	}

	// Store arguments
	string mode = argv[1];
	
	try
	{
		if( mode == "spatial" || mode == "spatial-cb" ||
			mode == "spatial-lens" || mode == "spatial-cb-lens")
		{
			if(argc < 5)
			{
				// Display usage for this mode
				cout << "Usage: " << program_name << " " << mode << " <ref-file> <copy-file> <output-file>" << endl;
				return 1;
			}

			string ref_file = argv[2];
			string cpy_file = argv[3];
			string out_file = argv[4];

			bool useCheckerboard = false;
			bool assumeLensDistortion = false;

			if(mode.find("cb") != string::npos) useCheckerboard = true;
			if(mode.find("lens") != string::npos) assumeLensDistortion = true;

			calibSpatial(ref_file, cpy_file, useCheckerboard, assumeLensDistortion, out_file);
		}
		else if(mode == "spatial-neutral")
		{
			// Special mode to simply create a neutral spatial transform
			if(argc < 5)
			{
				// Display usage for this mode
				cout << "Usage: " << program_name << " " << mode << " <output-file> <width> <height>" << endl;
				return 1;
			}

			// Parse parameters
			string result_file = argv[2];
			int width = stoi(argv[3]);
			int height = stoi(argv[4]);

			// Create transformation
			ImRegParams identityParams;
			identityParams.initializeNeutral(cv::Size(width, height));

			// Store transformation
			cv::FileStorage fs_result(result_file, cv::FileStorage::WRITE);
			fs_result << "SpatialRegistration" << "[";
			fs_result << identityParams;
			fs_result << "]";
		}
		else if(mode == "agc-fix" || mode == "agc-perc")
		{
			if(argc < 5)
			{
				// Display usage for this mode
				cout << "Usage: " << program_name << " " << mode << " <ref-file> <copy-file> <output-file>" << endl;
				return 1;
			}

			Scenario::AGCMethod method;
			if(mode == "agc-fix") method = Scenario::AGC_MEAN_THRESHOLD_OFFSET_FIX_ADJUST;
			else if(mode == "agc-perc") method = Scenario::AGC_MEAN_THRESHOLD_OFFSET_PERC_ADJUST;
			else { cerr << "Internal error: unknown AGC method." << endl; /* should never happen */ return 1; }

			string ref_file = argv[2];
			string cpy_file = argv[3];
			string out_file = argv[4];

			calibAGC(ref_file, cpy_file, method, out_file);
		}
		else if(mode == "temporal")
		{
			if(argc < 3)
			{
				// Display usage for this mode
				cout << "Usage: " << program_name << " temporal <copy-file>" << endl;
				return 1;
			}

			string cpy_file = argv[2];

			calibTemporal(cpy_file);
		}
		else if(mode == "temporal-embed" /*|| mode == "temporal-embed-luminance"*/)
		{
			if(argc < 5)
			{
				// Display usage for this mode
				cout << "Usage: " << program_name << " temporal-embed <source-file> <output-file> <pattern-image>" << endl;
				return 1;
			}

			string src_file = argv[2];
			string out_file = argv[3];
			string pat_file = argv[4];

			bool createLuminanceTransferVideo = false;
			//if(mode.find("luminance") != string::npos) createLuminanceTransferVideo = true;

			calibTemporalEmbed(src_file, out_file, pat_file, createLuminanceTransferVideo);
		}
		else
		{
			cerr << "'" << mode << "': Invalid mode." << endl;
			cerr << "Valid modes are: 'temporal', 'temporal-embed'." << endl;
		}
	}
	catch(exception &e)
	{
		cerr << "An error occurred:" << endl;
		cerr << e.what() << endl;
		return 1;
	}
	catch(...)
	{
		cerr << "An unknown error occurred." << endl;
		return 2;
	}

	return 0;
}

