/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../stdafx.h"

#include "../utility/VideoReader.h"
#include "../utility/BufferedVideoReader.h"
#include "../utility/Version.h"

#include "../common/Scenario.h"

#include "CalibAGC.h"

using namespace std;
using namespace cv;
using namespace boost;
using namespace boost::this_thread;

Calibration_AGC::Calibration_AGC(Scenario::AGCMethod _method) :
	method(_method), xyzChannels(3),
	numCores(boost::thread::hardware_concurrency())
{
	switch(method)
	{
	case Scenario::AGC_MEAN_THRESHOLD_OFFSET_FIX_ADJUST:
		THRESH_RANGE = 75.0f; THRESH_STEP = 2.5f;
		ADJUST_RANGE = 10.0f; ADJUST_STEP = 0.25f;
		COUNT_RANGE = 8;
		break;
	
	case Scenario::AGC_MEAN_THRESHOLD_OFFSET_PERC_ADJUST:
		THRESH_RANGE = 75.0f; THRESH_STEP = 2.5f;
		ADJUST_RANGE = 0.5f; ADJUST_STEP = 0.01f;
		COUNT_RANGE = 8;
		break;
	
	default:
		throw runtime_error(
			"AGC calibration: invalid AGC method.");
	}
}

Calibration_AGC::~Calibration_AGC(void)
{
}

void Calibration_AGC::openRefVideo(std::string file)
{
	// Open reference file
	ref.open(file); if(!ref.isOpened())
		throw runtime_error("Reference file '" + file + "' could not be opened.");

	// Set properties on the video files
	ref.set(CV_CAP_PROP_CONVERT_RGB, 1);

	// Store frame count
	numFrames = ref.getFrameCount();
	currFrame = 0;
}

bool Calibration_AGC::readRefVideo(float &progress)
{
	// Read frames until the video is over
	ref >> frame;
	if(frame.empty()) return false;

	// Calculate the luminance
	float luminance = calculateLuminance(frame);
	ref_luminance.push_back(luminance);

	currFrame++;

	// Calculate the current progress
	progress = (float)currFrame / numFrames;
	return true;
}

void Calibration_AGC::openCopyVideo(std::string file)
{
	// Open copy file
	copy.open(file); if(!copy.isOpened())
		throw runtime_error("Copy file '" + file + "' could not be opened.");

	// Set properties on the video files
	copy.set(CV_CAP_PROP_CONVERT_RGB, 1);

	// Store frame count
	numFrames = copy.getFrameCount();
	currFrame = 0;
}

bool Calibration_AGC::readCopyVideo(float &progress)
{
	// Read frames until the video is over
	copy >> frame;
	if(frame.empty()) return false;

	// Calculate the luminance
	float luminance = calculateLuminance(frame);
	copy_luminance.push_back(luminance);

	currFrame++;

	// Calculate the current progress
	progress = (float)currFrame / numFrames;
	return true;
}

void Calibration_AGC::writeRefLuminancesToFile(string file)
{
	ofstream ofs(file);	if(!ofs.good())
		throw runtime_error("Luminance file '" + file + "' could not be opened for writing.");

	for(float &lum : ref_luminance)
		ofs << lum << endl;
}

void Calibration_AGC::writeCopyLuminancesToFile(string file)
{
	ofstream ofs(file);	if(!ofs.good())
		throw runtime_error("Luminance file '" + file + "' could not be opened for writing.");

	for(float &lum : copy_luminance)
		ofs << lum << endl;
}

bool Calibration_AGC::readRefLuminancesFromFile(string file)
{
	ifstream ifs(file);
	if(!ifs.good()) return false;

	ref_luminance.clear();

	string line;
	while(getline(ifs, line))
	{
		if(line.empty()) continue;

		stringstream sstr_parser;
		sstr_parser << line;

		float lum;
		sstr_parser >> lum;

		ref_luminance.push_back(lum);
	}

	return true;
}

bool Calibration_AGC::readCopyLuminancesFromFile(string file)
{
	ifstream ifs(file);
	if(!ifs.good()) return false;

	copy_luminance.clear();

	string line;
	while(getline(ifs, line))
	{
		if(line.empty()) continue;

		stringstream sstr_parser;
		sstr_parser << line;

		float lum;
		sstr_parser >> lum;

		copy_luminance.push_back(lum);
	}

	return true;
}

void Calibration_AGC::startRegistration()
{
	// Calculate the number of frames to be processed
	numFrames = (unsigned int)min(ref_luminance.size(), copy_luminance.size());
	currFrame = 0;

	float lowerMidValue = 127 - (THRESH_RANGE / 2);
	float threshLowerStart = lowerMidValue - THRESH_RANGE;
	float threshLowerEnd = lowerMidValue + THRESH_RANGE;

	float upperMidValue = 127 + (THRESH_RANGE / 2);
	float threshUpperStart = upperMidValue - THRESH_RANGE;
	float threshUpperEnd = upperMidValue + THRESH_RANGE;

	float adjustLowerStart = 0.0f;
	float adjustLowerEnd = ADJUST_RANGE;

	float adjustUpperStart = 0.0f;
	float adjustUpperEnd = ADJUST_RANGE;

	cout << "THRESH LOWER: range = " << threshLowerStart << " to " << threshLowerEnd << ", step = " << THRESH_STEP << endl;
	cout << "THRESH UPPER: range = " << threshUpperStart << " to " << threshUpperEnd << ", step = " << THRESH_STEP << endl;
	cout << "ADJUST LOWER: range = " << adjustLowerStart << " to " << adjustLowerEnd << ", step = " << ADJUST_STEP << endl;
	cout << "ADJUST UPPER: range = " << adjustUpperStart << " to " << adjustUpperEnd << ", step = " << ADJUST_STEP << endl;
	cout << "COUNT:        range = 0 to " << COUNT_RANGE << endl;
	
	// Reset any previous data
	paramIndexList.clear();
	
	Parameters currParams;
	float minimumRange = 10.0;

	// Iterate over everything and fill the array
	for (currParams.th_lower = threshLowerStart;
		 currParams.th_lower <= threshLowerEnd;
		 currParams.th_lower += THRESH_STEP)
	{
		for (currParams.th_upper = threshUpperStart;
			 currParams.th_upper <= threshUpperEnd;
			 currParams.th_upper += THRESH_STEP)
		{
			double currThreshRange = currParams.th_upper - currParams.th_lower;
			if(currThreshRange < minimumRange) continue;

			for (currParams.adj_lower = adjustLowerStart;
				 currParams.adj_lower <= adjustLowerEnd;
				 currParams.adj_lower += ADJUST_STEP)
			{
				for (currParams.adj_upper = adjustUpperStart;
					 currParams.adj_upper <= adjustUpperEnd;
					 currParams.adj_upper += ADJUST_STEP)
				{
					for (currParams.th_count = 0;
						 currParams.th_count <= COUNT_RANGE;
						 currParams.th_count++)
					{
						paramIndexList.push_back(currParams);
					}
				}
			}
		}
	}
	
	cout << "numCombinations: " << paramIndexList.size() << endl ;

	// Initialize the lowest diff to the difference of unmodified ref and copy
	cout << "Direct ref/copy diff = " << calculateDiff(ref_luminance, copy_luminance) << endl;

	// Set start index to first element
	lowestDiff = FLT_MAX;
	bestParamsIndex = UINT_MAX;
	params_list_start = 0;
	currStep = 0;

	// Instanciate the vector of difference-iterator pairs
	for( int i = 0 ; i < numCores ; i++ )
		difference.push_back(make_pair(lowestDiff, bestParamsIndex)) ;
}



bool Calibration_AGC::doRegistrationStep(float &progress, int interval)
{
	size_t stop = min(params_list_start + interval,
					  paramIndexList.size()) ;
	
	thread_group parameterThreads ;
	
	for(int i = 0 ; i < numCores ; i++)
	{
		if(stop >= paramIndexList.size()) break;

		parameterThreads.create_thread(boost::bind(
			&Calibration_AGC::threadApplyParameters, this, params_list_start, stop, i)) ;
		
		params_list_start = stop ;
		stop += interval;
	}

	// Wait for all threads to complete
	parameterThreads.join_all() ;

	sort(difference.begin(), difference.end()) ;

	// Extract the smallest difference from the vector, along with it's parameter index.
	float diff = difference[0].first;
	
	// COMPARE TO MAKE SURE THAT THESE VALUES ARE THE GLOBAL MINIMUMS
	if (diff < lowestDiff)
	{
		lowestDiff = diff ;	
		bestParamsIndex = difference[0].second;
	}
	
	// Check if there are any more parameters to process
	if(stop >= paramIndexList.size())
		return false ;
	
	//Calculate the current progress
	progress = (float)currStep / paramIndexList.size();
	return true;
	
}

void Calibration_AGC::threadApplyParameters(size_t start, size_t stop, int index)
{
	LuminanceVec data_mod;
	
	for(size_t idx = start ; idx < stop ; ++idx)
	{
		// Apply current parameters
		applyParameters(ref_luminance, data_mod, paramIndexList[idx]);
		
		// Compute difference
		float diff = calculateDiff(data_mod, copy_luminance);
		
		if (diff < difference[index].first)
			difference[index] = make_pair(diff, idx);
	}
	
	// Take control of the currStep variable to properly update it and check to see if it is at the end.
	boost::mutex::scoped_lock lock(curr_frame_mutex);
	currStep = stop - 1;
}


void Calibration_AGC::finishRegistration(string result_file)
{
	if(bestParamsIndex == UINT_MAX)
	{
		cerr << "Could not find parameters that are better than not applying any distortion." << endl;
		cerr << "Are the reference and copy videos identical?" << endl;
		cerr << "NOT writing output file." << endl;
		return;
	}

	cout << "Final minimal different is " << lowestDiff << endl;
	Parameters bestParams = paramIndexList[bestParamsIndex];

	// Write the result file
	cout << "Writing result file..." << endl;
	ofstream agc_res(result_file);
	if(!agc_res.good()) throw runtime_error(
		"Result file '" + result_file + "' could not be opened for writing.");

	string methodName = Scenario::AutomaticGainControl::methodNames[method];
	agc_res << "<Method name=\"" << methodName << "\">" << endl;

	agc_res	<< "<LowerBound count=\"" << bestParams.th_count
			<< "\" thresh=\"" << bestParams.th_lower
			<< "\" adjust=\"" << bestParams.adj_lower << "\"/>" << endl;
	
	agc_res << "<UpperBound count=\"" << bestParams.th_count
			<< "\" thresh=\"" << bestParams.th_upper
			<< "\" adjust=\"" << bestParams.adj_upper << "\"/>" << endl;

	// Store best matching luminances to a file
	cout << "Writing best matching luminance values to a file..." << endl;
	LuminanceVec data_mod_best;
	applyParameters(ref_luminance, data_mod_best, bestParams);

	string bestMatchFile = result_file + "_bestMatchLum.txt";
	ofstream ofs(bestMatchFile); if(!ofs.good())
		throw runtime_error("Luminance file '" + bestMatchFile + "' could not be opened for writing.");

	for(float &lum : data_mod_best)
		ofs << lum << endl;

	cout << "Done." << endl;
}

float Calibration_AGC::calculateLuminance(Mat &frame)
{
	cvtColor(frame, frame_ycrcb, CV_BGR2YCrCb);
	split( frame_ycrcb, xyzChannels );
	return (float)mean(xyzChannels[0])[0];
}

float Calibration_AGC::calculateDiff(const LuminanceVec &data1,
									  const LuminanceVec &data2)
{
	float diff = 0.0;

	for (unsigned int i = 0; i < numFrames; i++)
		//diff += ( data1[i] - data2[i]) * ( data1[i] - data2[i]);
		diff += fabs( data1[i] - data2[i]);
	
	return diff / numFrames;
}

void Calibration_AGC::applyParameters(
	const LuminanceVec &orig_luminance,
	LuminanceVec &mod_luminance,
	const Parameters &params)
{
	float th_lower = params.th_lower;
	float th_upper = params.th_upper;
	float adj_lower = params.adj_lower;
	float adj_upper = params.adj_upper;
	unsigned int th_count = params.th_count;

	// initialize data to be modified
	mod_luminance.clear();
	mod_luminance.resize(ref_luminance.size());
	mod_luminance[0] = orig_luminance[0];

	unsigned int cnt = 0;
	float gain_offset = 0.0;

	for ( size_t i = 1 ; i < orig_luminance.size() ; i++ )
	{
		float v = mod_luminance[i-1];

		if ( v > th_lower && v < th_upper )
		{
			// do not change current gain
			cnt = 0;
		}
		else // if outside thresholds for several frames: change gain
		{
			if ( cnt >= th_count )
			{
				if ( v < th_lower )
				{
					switch(method)
					{
					case Scenario::AGC_MEAN_THRESHOLD_OFFSET_FIX_ADJUST:
						gain_offset += adj_lower; break;
	
					case Scenario::AGC_MEAN_THRESHOLD_OFFSET_PERC_ADJUST:
						gain_offset += (th_lower - v) * adj_lower; break;
					}
				}

				if ( v > th_upper )
				{
					switch(method)
					{
					case Scenario::AGC_MEAN_THRESHOLD_OFFSET_FIX_ADJUST:
						gain_offset -= adj_upper; break;
	
					case Scenario::AGC_MEAN_THRESHOLD_OFFSET_PERC_ADJUST:
						gain_offset += (th_upper - v) * adj_upper; break;
					}
				}
			}
			else cnt++;
		}

		// modify and store luminance value
		float v_mod = orig_luminance[i] + gain_offset;
		if ( v_mod > 255 ) v_mod = 255 ;
		if ( v_mod < 0 ) v_mod = 0 ;
		mod_luminance[i] = v_mod;
	}
}
