/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../stdafx.h"

#include "../utility/VideoReader.h"
#include "../utility/BufferedVideoReader.h"
#include "../utility/SynchronizedQueue.h"
#include "../utility/ThreadCout.h"
#include "../utility/Util.h"
#include "../utility/Version.h"

#include "../common/Scenario.h"

#include "CalibSpatial.h"

using namespace std;
using namespace cv;
namespace fs = boost::filesystem ;

Calibration_Spatial::Calibration_Spatial()
	: 	checkerboardPattern(false)
{
	maxInQueueSize = 5 ;
	maxOutQueueSize = 5 ;
}

Calibration_Spatial::~Calibration_Spatial(void)
{
}

void Calibration_Spatial::startRegistration(
	string ref_file, string copy_file,
	bool checkerboardPattern, bool assumeLensDistortion,
	string result_file)
{
	// Open reference file
	ref.open(ref_file); if(!ref.isOpened())
		throw runtime_error("Reference file '" + ref_file + "' could not be opened.");
	
	// Open copy file
	copy.open(copy_file); if(!copy.isOpened())
		throw runtime_error("Copy file '" + copy_file + "' could not be opened.");

	// Set properties on the video files
	ref.set(CV_CAP_PROP_CONVERT_RGB, 1);
	copy.set(CV_CAP_PROP_CONVERT_RGB, 1);
	
	// Get the number of frames from both sources
	unsigned int numFrames_ref = ref.getFrameCount();
	unsigned int numFrames_copy = copy.getFrameCount();
	
	TLOG("Number of reference frames: " << numFrames_ref);
	TLOG("Number of copy frames: " << numFrames_copy);
	
	if( 0 == numFrames_ref || 0 == numFrames_copy )
		throw runtime_error("Calibration_Spatial::startRegistration: Error: Unable to detect any frames for input images/videos"
							" to be used for spatial calibration." ) ;
	
	// Determine the number of frames that will be processed and define the appropriate 'pairMode' condition
	if(numFrames_ref == 1 && numFrames_copy == 1)
	{
		numFrames = 1 ;
		pairMode  = 3 ;
	}
	else if(numFrames_ref == 1)
	{
		numFrames = numFrames_copy ;
		pairMode  = 1 ;
	}
	else if(numFrames_copy == 1)
	{
		numFrames = numFrames_ref ;
		pairMode  = 2 ;
	}
	else
	{
		numFrames = min(numFrames_ref, numFrames_copy) ;
		pairMode  = 0 ;
	}
	
	// Get ref/copy frame sizes
	Size ref_size = ref.getSize();
	Size copy_size = copy.getSize();

	double ref_ar = (double)ref_size.width / ref_size.height;
	double copy_ar = (double)copy_size.width / copy_size.height;

	// Log info about frame sizes
	cout << "Reference frame size: " << ref_size.width << " x "  << ref_size.height
		 << " (aspect ratio = " << ref_ar << ")." << endl;
	cout << "Copy frame size:      " << copy_size.width << " x " << copy_size.height
		 << " (aspect ratio = " << copy_ar << ")." << endl;
	cout << endl;

	// Get frame rate of the copy
	double cpy_fps = copy.getFPS();
	
	if(cpy_fps == 0 || ISNAN(cpy_fps))
	{
		cpy_fps = 25;
		
		if(numFrames_copy > 1)
			cout << "Unable to detect copy frame rate, using default of 25 fps." << endl;
	}

	// Log info about the sampling frame rate
	if(numFrames_copy > 1)
	{
		cout << "Sampling frame rate (= copy frame rate) is " << cpy_fps << " fps." << endl;
		cout << endl;
	}

	// Create video file writer for the background
	string bgVideoFilename = result_file + "_background.avi";
	static int nCodec_XVID = CV_FOURCC('X', 'V', 'I', 'D');
	VideoWriter writer( bgVideoFilename, nCodec_XVID, cpy_fps, copy_size ) ;
	if(!writer.isOpened()) throw runtime_error(
		"VideoWriter for background output video could not be initialized.") ;

	// Get filename of background video (without path)
	boost::filesystem::path bgFilePath = bgVideoFilename;
	string bgFilenameOnly = bgFilePath.filename().string();

	// Open a result file storage
	fs_result.open(result_file, FileStorage::WRITE);
	
	// Write the sampling rate of we have more than one copy frame
	if(numFrames_copy > 1)
		fs_result << "SamplingFrameRate" << cpy_fps;

	// Write the background file name
	fs_result << "BackgroundFile" << bgFilenameOnly;

	// Start a list for storing all spatial registration results
	fs_result << "SpatialRegistration" << "[";

	// Store parameters
	this->checkerboardPattern = checkerboardPattern;
	this->assumeLensDistortion = assumeLensDistortion;

	// Reset the current frame counter
	currFrame = 0;    
	
	finishedProcessing = false ;

	// Get number of cores for multithreading
	const int numCores = boost::thread::hardware_concurrency() ;
	cout << "Number of cores: " << numCores << endl ;
	numThreads = numCores ;
	
	// Create a thread for loading the queues with frames. 
	boost::thread loadQueue(boost::bind( &Calibration_Spatial::loadVideoQueues, this)) ;

	// Create a thread to take the data created in the processing threads and write it to a file
	writeResults.create_thread(boost::bind( &Calibration_Spatial::writeRegistrationResults, this, writer)) ;

	// Create threads for processing the frames and doing the registration
	// Be careful adding too many. Memory will become an issue (on some machines, at least)
	for(int i = 0 ; i < numThreads ; i++)
	{
		processFrames.create_thread(boost::bind( &Calibration_Spatial::doRegistrationStep, this )) ;    
		cout << "Thread " << i << " created." << endl ;
	}
}

void Calibration_Spatial::doRegistrationStep()
{
	static atomic<int> thId(0);
	string tname = "spatialRegistration:" + to_string(thId++) ;
	ThreadCout::tcout.setThreadName(tname);

	FeatureType features_perspective = checkerboardPattern ?
		CHECKERBOARD_31x20 : SIFT_Dbl_GM_RAN;

	FeatureType features_lens;
	unsigned int maxNumFeatures;
	
	if(assumeLensDistortion)
	{
		 features_lens = features_perspective;
		 maxNumFeatures = maxNumFeatures_Lens;
	}
	else
	{
		features_lens = NoRegistration;
		maxNumFeatures = maxNumFeatures_Persp;
	}

	vector<Mat> ref_frames(1), copy_frames(1);
	Mat &ref_frame = ref_frames.back();
	Mat &copy_frame = copy_frames.back();

	Mat ref_mask, cpy_mask;

	while( true )
	{
		FramePair inFrames = queueIn.Dequeue();
		
		unsigned int currFrameIdx = inFrames.idx ;

		ref_frame = inFrames.refFrame ;
		copy_frame = inFrames.cpyFrame ;
		
		if( ref_frame.empty() || copy_frame.empty())
			TLOG( "ERROR: doRegistrationStep(): Received empty frame." );
		
		// Do the spatial registration
		ImRegParams reg_params;
		try {
			// Currently hard-coded
			const int gridCols = 1;
			const int gridRows = 1;

			reg_params = register_frames(ref_frames, copy_frames,
				features_lens, features_perspective, 0 /*flags*/,
				maxNumFeatures * gridCols * gridRows,
				gridCols, gridRows);
		
		} catch (...) {
			TLOG("Caught exception during call to register_frames.");
		}
		
		Mat background;

		if( !reg_params.valid() )
		{
			TLOG("WARNING: Failed to spatially register the reference with copy for frame #" << currFrameIdx);
			background = copy_frame;
		}
		else
		{
			// Extract the background
			if(ref_mask.empty()) ref_mask.create(ref_frame.size(), CV_8UC3);
			ref_mask.setTo(Scalar::all(255));
			remap_frame(ref_mask, cpy_mask, reg_params, copy_frame.size(), INTER_LINEAR);            
			dilate(cpy_mask, cpy_mask, Mat(), Point(-1, -1), 3);
			background = copy_frame - cpy_mask;
		}

		// Pack the data into the structure and queue it.
		FrameData frData ;
		frData.idx   = currFrameIdx ;
		frData.frame = background ;
		frData.sr    = reg_params ;
		queueOut.Enqueue(frData) ;

		// Atomically increase the current frame count
		currFrame++ ;
		
		boost::this_thread::interruption_point();
	}
}

void Calibration_Spatial::finishRegistration()
{
	// Close list and file storage
	fs_result << "]";
	fs_result.release();
}

void Calibration_Spatial::loadVideoQueues()
{
	// Used to fill a queue with sets of corresponding source and copy frames. This eliminates threading issues.
	
	currentInputFrame = 0 ;
	
	FramePair currentFrames ;
	
	// Get the next frame
	ref  >> currentFrames.refFrame;
	copy >> currentFrames.cpyFrame;
	currentFrames.idx = currentInputFrame++ ;

	bool processingDone = false;

	while(true)
	{
		int currQueueSize = queueIn.getSize() ;
		int currOutQueueSize = queueOut.getSize() ;
		
		if( currQueueSize < maxInQueueSize && currOutQueueSize < maxOutQueueSize)
		{
			// When all frames have been processed prepare to exit
			if( currentFrames.refFrame.empty() || currentFrames.cpyFrame.empty())
			{
				// Don't load empty data into the queue. This may seem like a redundant step but it will keep
				// the code from entering the 'else' statement which would write an empty frame into the queue
				// and cause the program to crash.
				
				TLOG( "Input video completely processed, finishing tasks." )
				
				if( !processingDone && currQueueSize == 0 )
				{
					// Interrupt threads here
					processFrames.interrupt_all() ;
					
					// Join threads here
					processFrames.join_all();

					// Set flag
					processingDone = true;

					TLOG( "Processing done, finishing writing." )
				}
				
				// Update out queue size
				currOutQueueSize = queueOut.getSize() ;

				if( processingDone && currOutQueueSize == 0 )
				{
					// Now do the same for the encoder. Must be done after the processing threads.
					writeResults.interrupt_all() ;
					writeResults.join_all() ;
					
					TLOG( "Done." )
					
					finishedProcessing = true ;
					
					break ;
				}
			}
			else
			{
				queueIn.Enqueue(currentFrames) ;
			}
			
			// Determine what frames to read next
			switch (pairMode)
			{
				case 0: // Two Videos
					// Empty the frames so that the check above will still work.
					currentFrames.refFrame = Mat() ;
					currentFrames.cpyFrame = Mat() ;
					
					// Get the next frame
					ref  >> currentFrames.refFrame;
					copy >> currentFrames.cpyFrame;
					currentFrames.idx = currentInputFrame++ ;
					
					break;
					
				case 1: // RefImg/CpyVideo
					// Empty the frames so that the check above will still work.                    
					currentFrames.cpyFrame = Mat() ;
					// Get the next frame
					copy >> currentFrames.cpyFrame;
					currentFrames.idx = currentInputFrame++ ;
					
					break;
					
				case 2: // Ref Video/ Cpy Image
					// Empty the frames so that the check above will still work.
					currentFrames.refFrame = Mat() ;
					
					// Get the next frame
					ref  >> currentFrames.refFrame;
					currentFrames.idx = currentInputFrame++ ;
					
					break;
					
				case 3: // Ref Image/ Cpy Image
					// Empty the frames so that the check above will still work.
					currentFrames.refFrame = Mat() ;
					currentFrames.cpyFrame = Mat() ;
					
					currentFrames.idx = currentInputFrame++ ;
					
					break;
					
				default:
					break;
			}
		}
		
		// Also acts as an interruption point
		boost::this_thread::interruption_point();
	}
}

void Calibration_Spatial::writeRegistrationResults(VideoWriter& writer)
{
	// Write the data to a file and the frames to a video
	
	// Pulls a frame from the queue and writes it to a video
	unsigned int sizeOut = 0 ;
	int nextFrame = 0 ;
		
	vector<FrameData> frameStore ;
	
	while(true)
	{
		sizeOut = queueOut.getSize() ;
		
		while(sizeOut > 0)
		{
			FrameData frameDat = queueOut.Dequeue() ;
			
			if( frameDat.idx == nextFrame )
			{
				while(frameDat.idx == nextFrame)
				{
					// Write to video
					writer << frameDat.frame ;

					// Write spatial registration data
					fs_result << frameDat.sr;
					
					nextFrame++ ;
					
					if(frameStore.size())
					{
						for(auto it = frameStore.begin() ; it != frameStore.end() ; ++it)
						{
							if(it->idx == nextFrame)
							{
								frameDat = *it ;
								frameStore.erase(it) ;
								break ; // break out of for loop
							}
						}
					}
				}
			}
			else
				frameStore.push_back(frameDat) ;
			
			sizeOut = queueOut.getSize() ;
		}
		
		// Call interruption point here so the thread can be killed later
		boost::this_thread::interruption_point();
	}
}

float Calibration_Spatial::getProgress()
{
	return (float)currFrame / numFrames ;
}

bool Calibration_Spatial::doneProcessing()
{
	return finishedProcessing ;     
}

