/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

//#define EXTRACT_LUMINANCES

#include "../stdafx.h"

#include "../utility/CRC8.h"
#include "../utility/VideoReader.h"
#include "../utility/SynchronizedQueue.h"
#include "../utility/ThreadCout.h"
#include "../utility/Util.h"
#include "../utility/Img2Cpp.h"

#include "../resources/marker_xo_images.h"

#include "CalibSpatial.h"
#include "CalibTemporal.h"

using namespace std;
using namespace cv;
namespace fs = boost::filesystem ;

Calibration_Temporal::Calibration_Temporal(void)
{
	firstFrameAfterSyncPattern = 0;

	maxValidFrameJump = 100 ;
	lastValidFrameNumber = 0 ;

	// Look-of-table for gamma expansion
	double gamma = 2.2;
	lut_gamma.create(1, 256, CV_8UC1);
	uchar *ptr = lut_gamma.ptr();
	for( int i = 0; i < 256; i++ )
		ptr[i] = (int)( pow( (double) i / 255.0, gamma ) * 255.0 );
	
	// Detect CUDA support
	int numCUDADevs = cv::gpu::getCudaEnabledDeviceCount();
	if(numCUDADevs > 0) hasCUDASupport = true;
	else hasCUDASupport = false;
}


Calibration_Temporal::~Calibration_Temporal(void)
{
}
	
void Calibration_Temporal::loadLUT_pixelVal2Exp(std::string lut_file)
{
	ifstream ifs(lut_file, ios::binary);
	if(!ifs.good())	throw runtime_error(
		"Look-up-table '" + lut_file + "' could not be opened for reading.");

	ifs.seekg (0, ios::beg);

	int numHorizontalDivisons;
	ifs.read((char*)&numHorizontalDivisons, sizeof(int));
	if (!ifs.good() || numHorizontalDivisons < 1) throw runtime_error(
		"Look-up-table '" + lut_file + "' is empty or contains invalid data.");

	lut_pixelVal2Exp.resize(numHorizontalDivisons);

	for(int row = 0; row < numHorizontalDivisons; row++)
	{
		lut_pixelVal2Exp[row].resize(256);
		for( int i = 0; i < 256; i++ )
		{
			double exposure;
			ifs.read((char*)&exposure, sizeof(double));
			if (!ifs.good() || exposure < 0) throw runtime_error(
				"Look-up-table '" + lut_file + "' is incomplete or contains invalid data.");

			lut_pixelVal2Exp[row][i] = exposure;
		}
	}
}

void Calibration_Temporal::startRegistration(string ref_file)
{
	// Hard-coded parameters
	string filenameMarkerPositions = "_marker_positions.yml";
	//string filenameLUT = "lookup_table.bin"; // Empty filename means no LUT
	string filenameLUT = "";

	// Open reference file
	input_vid.open(ref_file); if(!input_vid.isOpened())
		throw runtime_error("Input video '" + ref_file + "' could not be opened.");
	
	// Set properties on the file-capture
	input_vid.set(CV_CAP_PROP_CONVERT_RGB, 1);
	
	// Read fps and number of frames properties
	float src_fps           = (float)input_vid.get(CV_CAP_PROP_FPS);
	unsigned int numFrames  = (unsigned int)input_vid.get(CV_CAP_PROP_FRAME_COUNT);

	totalFrames = numFrames ;
	currFrame = 0 ;
	finishedProcessing = false ; 
	
	// BUG: On linux src_fps is returned at nan. Just set it to 24fps if this happens for now.
	if ( ISNAN( src_fps ) )
	{
		src_fps = 24.0 ;
		TLOG("Source fps read incorrectly from the input video. Set to " << src_fps);
	}
	
	// Read frame size property
	Size srcFrameSize;
	srcFrameSize.width  = (int)(input_vid.get(CV_CAP_PROP_FRAME_WIDTH));
	srcFrameSize.height = (int)(input_vid.get(CV_CAP_PROP_FRAME_HEIGHT));

	TLOG("Input video: " << srcFrameSize.width << " x " << srcFrameSize.height << ", "
		 << src_fps << " fps, " << numFrames << " frames.");
	
	// Load X/O marker images
	imgX = Img2Cpp::loadImageFromArray_PNG(Img2Cpp::marker_X, Img2Cpp::marker_X_length);
	imgO = Img2Cpp::loadImageFromArray_PNG(Img2Cpp::marker_O, Img2Cpp::marker_O_length);	

	if(imgX.empty() || imgO.empty())
		throw runtime_error("Unable to load X an O images.");

	// Resize X/O marker images
	Size xoMarkerSize(15, 15);
	resize ( imgX, imgX, xoMarkerSize ) ;
	resize ( imgO, imgO, xoMarkerSize ) ;

	// Extract the source file path
	string input_path, input_name_only;
	Util::splitFilePathAndName(ref_file, input_path, input_name_only);

	// Load marker data
	string filenameSyncPatternImg;
	Size originalMarkerVideoSize;
	loadMarkerData(input_path + filenameMarkerPositions,
		filenameSyncPatternImg, originalMarkerVideoSize);

	// Load the spatial sync frame    
	boost::filesystem::path syncPatternPath(filenameSyncPatternImg);
	syncPatternPath.make_preferred();
	Mat syncPattern = imread (syncPatternPath.string());

	if( syncPattern.empty() ) throw runtime_error(
		"Unable to open '" + filenameSyncPatternImg + "' (the syncronization frame image).") ;

	// Resize sync pattern to the original marker video size
	resize( syncPattern, syncPattern, originalMarkerVideoSize ) ;

	// Create an output data folder
	boost::filesystem::path outputFolderPath(ref_file + "_frameData");
	outputFolderPath = outputFolderPath.make_preferred();
	outputFolder = outputFolderPath.string();

	// Create if it does not already exists
	if(!exists(outputFolderPath) || !fs::is_directory(outputFolderPath))
		fs::create_directories(outputFolderPath);
	
	// Open blend data output file
	avgBlendData.open( outputFolder + "avgBlendData.csv" ) ;

	// Load look-up-table if provided
	if(!filenameLUT.empty())
	{
		string lut_file = input_path + filenameLUT;
		loadLUT_pixelVal2Exp(lut_file);
	}

	// Set the maximum queue size
	maxQueueOutSize = 10 ;
	maxQueueInSize = 10 ;
		
#ifndef EXTRACT_LUMINANCES
	// Before starting the queues, we must find the spatial alignment from the beginning of the vid.
	string spatialDeformFile = ref_file + "_deform-data.yml";
	string temporalStartFile = ref_file + "_start-frame.txt";
	
	// Try loading previously stored number of first frame after sync
	firstFrameAfterSyncPattern = -1;
	ifstream startFrameFile(temporalStartFile);
	if(startFrameFile.good())
		startFrameFile >> firstFrameAfterSyncPattern;
	startFrameFile.close();

	// If we succeeded...
	if(firstFrameAfterSyncPattern >= 0)
	{
		// ...try loading previously stored spatial deform data
		FileStorage fs(spatialDeformFile, FileStorage::READ);
		fs["deform"] >> deform;
		fs["redeform"] >> redeform;
	}

	// If either is missing, run the spatial calibration
	if(firstFrameAfterSyncPattern < 0 || !deform.valid())
	{
		TLOG("Scanning frames for synchronization pattern...");

		// Find and store spatial calibration
		findSpatialAlignment(syncPattern, spatialDeformFile) ;

		// Store start frame
		int frameIdx = (int)input_vid.get(CV_CAP_PROP_POS_FRAMES);
		firstFrameAfterSyncPattern = frameIdx + 1;

		ofstream startFrameFile(temporalStartFile);
		startFrameFile << firstFrameAfterSyncPattern;
	}
	else
		TLOG("Using previously stored synchronization information.");

	// Log were actualy processing starts
	TLOG("First frame after sync pattern: " << firstFrameAfterSyncPattern);

#else
	// Load existing deform data
	FileStorage fs(input_path + "00000.MTS_deform-data.yml", FileStorage::READ);
	fs["deform"] >> deform;
	fs["redeform"] >> redeform;
#endif

	// Create mask
	Mat mask = Mat::zeros(srcFrameSize, CV_8UC3);
	mask(centerArea).setTo(Scalar::all(255));

	// Apply spatial deformation to mask
	remap_frame(mask, deformedMask, deform);

	// Start the threads
	const int numCores = boost::thread::hardware_concurrency() ;
	numThreads = numCores ;

#ifdef EXTRACT_LUMINANCES
	// Force only one processing thread
	numThreads = 1;
#endif

	TLOG("Detected " << numCores << " cores, starting worker threads.");

	boost::thread loadQueue(boost::bind( &Calibration_Temporal::loadVideoQueues, this)) ;
	
	for(int i = 0 ; i < numThreads ; i++)
		processFrames.create_thread(boost::bind(&Calibration_Temporal::processFrame, this )) ;
}

void Calibration_Temporal::loadMarkerData(const string filename,
	string &syncPatternFile, Size &markerVideoSize)
{
	FileStorage fs(filename.c_str(), FileStorage::READ);
	if(!fs.isOpened()) throw runtime_error(
		"Unable to open '" + filename + "' for reading.");

	fs["snycPatternFile"] >> syncPatternFile;
	fs["markerVideoSize"] >> markerVideoSize;

	int numFrameMarkerBits;
	fs["numFrameMarkerBits"] >> numFrameMarkerBits;
	
	for(int i = 0; i < 3; i++)
	{
		markersTop[i].resize(numFrameMarkerBits);
		markersBottom[i].resize(numFrameMarkerBits);
	}

	fs["centerArea"] >> centerArea;
	
	fs["top1"] >> markersTop[0];
	fs["top2"] >> markersTop[1];
	fs["top3"] >> markersTop[2];
	
	fs["bottom1"] >> markersBottom[0];
	fs["bottom2"] >> markersBottom[1];
	fs["bottom3"] >> markersBottom[2];
	
	fs["left1"] >> markersLeft[0];
	fs["left2"] >> markersLeft[1];
	fs["left3"] >> markersLeft[2];
	fs["left4"] >> markersLeft[3];
	fs["left5"] >> markersLeft[4];
	
	fs["right1"] >> markersRight[0];
	fs["right2"] >> markersRight[1];
	fs["right3"] >> markersRight[2];
	fs["right4"] >> markersRight[3];
	fs["right5"] >> markersRight[4];
}

void Calibration_Temporal::findSpatialAlignment(Mat syncPattern, string spatialDeformFile)
{
	bool syncPatNumFound = false;
	unsigned int maxFoundFeatures = 0;
	
	Mat frame, bestSyncPatternFrame;

	while(!deform.valid())
	{
		// Read input video frame
		input_vid >> frame;
		
		// This should never be reached but is here to clean up in case the video is empty
		if(frame.empty())
		{
			if(!bestSyncPatternFrame.empty())
				syncPatNumFound = true;
			else
			{
				TLOG("ERROR: Scanned the entire video but did not find the spatial synchronization pattern.");
				break;
			}
		}
		
		// Get the current frame index
		int frameIdx = (int)input_vid.get(CV_CAP_PROP_POS_FRAMES);
		
		// Conform with the SyncFrame to determine the synchronization
		if (!syncPatNumFound)
		{
			Mat resizedCopyFrame;
			resize(frame, resizedCopyFrame, syncPattern.size());
	
			vector<Mat> ref_frames, copy_frames;
			ref_frames.push_back(syncPattern);
			copy_frames.push_back(resizedCopyFrame);

			FrameFeaturesVector features;
			find_features(ref_frames, copy_frames,
				SIFT_Dflt_BEST500_GM_RAN, features, 500 /*max num features*/);
					/* Here, we can use a lower number of max num features*/

			// Get the number of matched features
			unsigned int numFoundFeatures =
				(unsigned int)features[0].copy_kpts.size();

			stringstream sstr;
			if(numFoundFeatures > 0) sstr << "Found " << numFoundFeatures << " pattern features in frame " << frameIdx;

			if (numFoundFeatures > maxFoundFeatures)
			{
				frame.copyTo(bestSyncPatternFrame);
				maxFoundFeatures = numFoundFeatures;

				sstr << " (new max)";
			}

			TLOG(sstr.str());

			if (numFoundFeatures == 0 && !bestSyncPatternFrame.empty())
				syncPatNumFound = true;
		}
		
		// If the current frame is the sync frame the deformation is certain
		if (syncPatNumFound && !deform.valid())
		{
			TLOG("Calculating spatial registration ...");
			this->showLogMessages();

			// Currently not configurable
			bool checkerboardPattern = false;
			bool assumeLensDistortion = false;

			FeatureType features_perspective = checkerboardPattern ?
				CHECKERBOARD_31x20 : SIFT_Dbl_GM_RAN;

			FeatureType features_lens;
			unsigned int maxNumFeatures;
	
			if(assumeLensDistortion)
			{
				 features_lens = features_perspective;
				 maxNumFeatures = Calibration_Spatial::maxNumFeatures_Lens;
			}
			else
			{
				features_lens = NoRegistration;
				maxNumFeatures = Calibration_Spatial::maxNumFeatures_Persp;
			}

			vector <Mat> ref_frames, copy_frames;
			ref_frames.push_back(syncPattern);
			copy_frames.push_back(bestSyncPatternFrame);

			deform = register_frames( ref_frames, copy_frames,
									features_lens, features_perspective,
									0 /*flags*/, maxNumFeatures * 4 * 4, 4, 4);
			if (!deform.valid())
				throw runtime_error("Failed to calculate spatial registration.");

			redeform = register_frames( copy_frames, ref_frames, 
									features_lens, features_perspective,
									0 /*flags*/, maxNumFeatures * 4 * 4, 4, 4);
			if (!redeform.valid())
				throw runtime_error("Failed to calculate spatial registration.");
		
			FileStorage fs(spatialDeformFile, FileStorage::WRITE);
			fs << "deform" << deform;
			fs << "redeform" << redeform;

			TLOG("Spatial registration data saved to '" << spatialDeformFile << "'.");
		}
	
		// Atomically increase the current frame
		currFrame++ ;

		this->showLogMessages();
	}
}

void Calibration_Temporal::finishRegistration()
{
#ifdef EXTRACT_LUMINANCES
	// Open CSV file
	stringstream fileNameLuminanceData; fileNameLuminanceData << outputFolder << "luminanceData.csv";
	ofstream luminanceDataCSV (fileNameLuminanceData.str());
	if(!luminanceDataCSV.good())
		throw runtime_error("Calibration_Temporal::processFrame: Unable to open the .csv for writing.") ;

	for(unsigned int i = 0; i < acumulatedBlendData[0].size(); ++i)
	{
		double avgBarValue = acumulatedBlendData[0][i] / acumulatedBlendDataCount;
		double avgBlackValue = acumulatedBlendData[1][i] / acumulatedBlendDataCount;

		// Write results to CSV
		luminanceDataCSV << avgBarValue << ";"
						 << avgBlackValue << ";" << endl;
	}
#endif
}

void Calibration_Temporal::loadVideoQueues()
{
	bool videoEmpty = false;

	while(!videoEmpty)
	{
		int currInQueueSize = frameQueueIn.getSize() ;
		int currOutQueueSize = frameQueueOut.getSize() ;
		
		if( (currInQueueSize < maxQueueInSize) && (currOutQueueSize < maxQueueOutSize) )
		{
			// Get the next frame. IMPORTANT:
			// Do not declare nextFrameBGR outside this scope,
			// it NEEDS to be re-created every time!
			Mat nextFrameBGR;
			input_vid >> nextFrameBGR;

			// When all frames have been processed, set the notifier
			if( nextFrameBGR.empty() /*|| TMPcurrFrameIdx > 25*/)
			{
				// Don't load empty data into the queue. This may seem like a redundant step but it will keep
				// the code from entering the 'else' statement which would write an empty frame into the queue
				// and cause the program to crash.
				
				if(currInQueueSize == 0)
				{
					videoEmpty = true ;
					
					// Interrupt threads here
					processFrames.interrupt_all() ;
					
					// Join threads here
					processFrames.join_all();
					
					finishedProcessing = true ;
					
					TLOG( "Input video completely processed." )
				}
			}
			else
			{
				int currFrameIdx = (int)input_vid.get(CV_CAP_PROP_POS_FRAMES);

				// Skip any sync frames
				if(currFrameIdx < firstFrameAfterSyncPattern)
					continue;

				FrameWithIdx currentFrame ;
				currentFrame.first = currFrameIdx - firstFrameAfterSyncPattern;
				currentFrame.second = nextFrameBGR;
				
				// Load the frame into the queue
				frameQueueIn.Enqueue(currentFrame) ;
			}
		}
	}
}

void Calibration_Temporal::processFrame()
{
	// ---Hard-coded parameters---
	// This threshold was determined as a reasonable choice in a set of experiments.
	// It is the minimum average correlation of the symbols encoding the frame number.
	float confThreshold = 0.35f;

	Mat redeformedFrame;
	Mat frameForMarkers, frameForMarkersGray;

	while( true )
	{
		FrameWithIdx dequeuedFrame = frameQueueIn.Dequeue() ;
		int &currFrameIdx = dequeuedFrame.first;
		
		// Redeform the frame
		remap_frame(dequeuedFrame.second, redeformedFrame, redeform);

		// Perform gamma expansion for marker scanning
		//LUT( redeformedFrame, lut_gamma, frameForMarkers );
		redeformedFrame.copyTo(frameForMarkers);

		// Convert frame to gray for marker scanning
		cvtColor(frameForMarkers, frameForMarkersGray, CV_BGR2GRAY);
		
		// Read all frame number markers
		vector<pair<int, float>> bestFrameNumbers(3);
		for(int position = 0; position < 3; position++)
		{
			// Read top and bottom
			pair<int, float> frameNumbersTop, frameNumbersBottom;
			frameNumbersTop = readFrameNumber(redeformedFrame, markersTop[position], confThreshold);
			frameNumbersBottom = readFrameNumber(redeformedFrame, markersBottom[position], confThreshold);

			// Select the better one of both
			if(frameNumbersTop.second > frameNumbersBottom.second)
				 bestFrameNumbers[position] = frameNumbersTop;
			else bestFrameNumbers[position] = frameNumbersBottom;
		}

		// Find position with highest reliability
		int strongestPos = 0;
		float maxConf = bestFrameNumbers[0].second;
		for(int position = 1; position < 3; position++)
			if(bestFrameNumbers[position].second > maxConf)
			{
				strongestPos = position;
				maxConf = bestFrameNumbers[position].second;
			}
		
		// If another of the three positions have a miss-detected frame number,
		// but a confidence value greater than our threshold, it means a number
		// WAS actually embedded. In this case, infere the missing frame number
		// from the strongest of the three positions if it was decoded correctly.
		int strongestFrameNr = bestFrameNumbers[strongestPos].first;
		if(strongestFrameNr != -1)
		{
			int prevPos = strongestPos - 1;
			int nextPos = (strongestPos + 1) % 3;
			if(prevPos == -1) prevPos = 2;

			// Infere frame number of prev frame's position if it meets our condition
			if(bestFrameNumbers[prevPos].first == -1 && bestFrameNumbers[prevPos].second > confThreshold
				&& strongestFrameNr > 0)
				bestFrameNumbers[prevPos].first = strongestFrameNr - 1;

			// Infere frame number of next frame's position if it meets our condition
			if(bestFrameNumbers[nextPos].first == -1 && bestFrameNumbers[nextPos].second > confThreshold
				&& strongestFrameNr < (totalFrames - 1))
				bestFrameNumbers[nextPos].first = strongestFrameNr + 1;
		}

		int frameNumber1 = bestFrameNumbers[0].first,
			frameNumber2 = bestFrameNumbers[1].first,
			frameNumber3 = bestFrameNumbers[2].first;

		// If frame numbers have a bigger difference then 5, mark all numbers as invalid
		if ((frameNumber1 != -1 && frameNumber2 != -1 && abs(frameNumber1 - frameNumber2) > 5) 
		|| (frameNumber1 != -1 && frameNumber3 != -1 && abs(frameNumber1 - frameNumber3) > 5) 
		|| (frameNumber2 != -1 && frameNumber3 != -1 && abs(frameNumber2 - frameNumber3) > 5))
		{
			TLOG( "Extracted frame numbers " << frameNumber1 << ", " << frameNumber2 << ", " << frameNumber3
				 << " are not plausible: Their distance is more than 5." );
			
			frameNumber1 = frameNumber2 = frameNumber3 = -1;
		}

		// If we encounter a jump that is larger than max. allowed, mark all numbers as invalid
		// CURRENTLY NOT THREAD SAFE
		/*if(frameNumber1 > lastValidFrameNumber + maxValidFrameJump ||
			frameNumber2 > lastValidFrameNumber + maxValidFrameJump ||
			frameNumber3 > lastValidFrameNumber + maxValidFrameJump)
		{
			TLOG( "Extracted frame numbers " << frameNumber1 << ", " << frameNumber2 << ", " << frameNumber3
				 << " are not plausible: Found a jump that is larger than " << maxValidFrameJump << ".");
			
			frameNumber1 = frameNumber2 = frameNumber3 = -1;
		}
		
		int maxFrameNum = max(frameNumber1, max(frameNumber2, frameNumber3));
		if( maxFrameNum != -1 ) lastValidFrameNumber = maxFrameNum;
		*/

		// Now scan blend markers
		vector<vector<double>> markerData = scanMarkers(frameForMarkersGray, 120);
		
		// Open frame CSV file
		stringstream fileNameMarkerData; fileNameMarkerData << outputFolder << currFrameIdx << "_markers.csv";
		ofstream markerDataCSV (fileNameMarkerData.str());
		if(!markerDataCSV.good()) throw runtime_error(
			"Unable to open the '" + fileNameMarkerData.str() + "' for writing.") ;
	
		vector<int> frameNumbers(5);
		frameNumbers[0] = frameNumbers[1] = frameNumbers[2] =
		frameNumbers[3] = frameNumbers[4] = currFrameIdx;
		
		// Write data
		writeBlendDataAndAvg(markerData, frameNumbers, currFrameIdx, markerDataCSV);
		
#ifdef EXTRACT_LUMINANCES
		unsigned int numHorizontalDivisions = markerData[0].size();

		if(acumulatedBlendData.empty())
		{
			acumulatedBlendDataCount = 0;
			acumulatedBlendData.resize(2);
			acumulatedBlendData[0].resize(numHorizontalDivisions);
			acumulatedBlendData[1].resize(numHorizontalDivisions);
		}

		for(unsigned int i = 0; i < numHorizontalDivisions; ++i)
		{
			double avgMarkerData = (markerData[0][i] + markerData[2][i] + markerData[4][i]) / 3;
			double avgBlackData = (markerData[1][i] + markerData[3][i]) / 2;
			
			acumulatedBlendData[0][i] += avgMarkerData;
			acumulatedBlendData[1][i] += avgBlackData;
		}

		acumulatedBlendDataCount++;
#endif

		// If we have a look-up-table, apply it to the read marker pixel values
		if(!lut_pixelVal2Exp.empty())
		{
			// Option A: Subtract before applying the LUT
			//subtractBlackValue(markerData);

			// Apply LUT
			applyLUTtoMarkers(markerData);

			// Option B: Subtract after applying the LUT
			//subtractBlackValue(markerData);

			// Open frame CSV file
			stringstream fileNameMarkerLUTData; fileNameMarkerLUTData << outputFolder << currFrameIdx << "_markers_lut.csv";
			ofstream markerLUTDataCSV (fileNameMarkerLUTData.str());
			if(!markerLUTDataCSV.good()) throw runtime_error(
				"Unable to open '" + fileNameMarkerLUTData.str() + "' for writing.") ;

			// Write data
			writeBlendDataAndAvg(markerData, frameNumbers, currFrameIdx, markerLUTDataCSV);
		}
		else
		{
			// If no look-up-table is applied,
			// subtract blackness value form all other values.
			subtractBlackValue(markerData);
		}

		// Now calculate influences
		vector<vector<double>> influences = calcInfluenceFromMarkers(
			markerData, frameNumber1 != -1, frameNumber2 != -1, frameNumber3 != -1);
		
		// Open frame CSV file
		stringstream fileNameBlendData; fileNameBlendData << outputFolder << currFrameIdx << "_data.csv";
		ofstream blendDataCSV (fileNameBlendData.str());
		if(!blendDataCSV.good()) throw runtime_error(
			"Unable to open '" + fileNameBlendData.str() + "' for writing.") ;
	
		frameNumbers[0] = frameNumber1;
		frameNumbers[1] = frameNumber2;
		frameNumbers[2] = frameNumber3;
		frameNumbers[3] = -1;
		frameNumbers[4] = -1;
		
		// Write data and average
		writeBlendDataAndAvg(influences, frameNumbers, currFrameIdx, blendDataCSV);

		// Enqueue frame for encoding
		FrameWithIdx enqueueFrame = dequeuedFrame ;
		enqueueFrame.second = enqueueFrame.second.mul((1.0)/255 * deformedMask) ;
		frameQueueOut.Enqueue( enqueueFrame ) ;
	}
}

pair<int, float> Calibration_Temporal::readFrameNumber(Mat &background, vector<Rect> markers, float confThreshold)
{
	// Read the frame markers
	pair<int, float> readVals = readXOMarkers(background, markers);

	// Split frame number and CRC, store confidence value
	int readFrameWithCRC = readVals.first ;
	int frameNumber = readVals.first & 0x00FFFFFF ;
	float confidence = readVals.second ;
	
	CRC8 checkCRC; // Apply some checks:
	if (!checkCRC.check_24bit_8bitcrc(readFrameWithCRC) ||	// 1) Check the CRC value
		frameNumber > totalFrames + 10 ||					// 2) Verify the frame number is reasonable
		confidence < confThreshold)							// 3) Verify the confidence is above the threshold
	{
		// Mark the extracted frame number as invalid, but keep the confidence.
		// If it is greater than our confThreshold, we might try to infere the correct number later.
		frameNumber = -1;
	}
	
	return make_pair(frameNumber, confidence);
}

pair<unsigned int, float> Calibration_Temporal::readXOMarkers(Mat &background, vector<Rect> positions)
{
	CV_Assert(positions.size() == 32);

	bitset<32> binary;

	vector<Mat> background_bgr(3); split(background, background_bgr);
	vector<Mat> imgX_bgr(3); split(imgX, imgX_bgr);
	vector<Mat> imgO_bgr(3); split(imgO, imgO_bgr);
	
	Mat resultX, resultO;
	double averageCorr = 0 ;
	
	for (int i = 0; i < 32; i++)
	{
		Rect window = positions[i];
		
		// Extend the search window by 3 pixels to each side
		window.x -= 3 ; window.y -= 3 ;
		window.width  += 6 ; window.height += 6 ;

		// Calculate correlation within the serach window
		matchTemplate(background_bgr[1](window), imgX_bgr[1], resultX, CV_TM_CCOEFF_NORMED);
		matchTemplate(background_bgr[2](window), imgO_bgr[2], resultO, CV_TM_CCOEFF_NORMED);

		// Calculate average correlation within search window
		//float xCorr = (float)mean(resultX)[0];
		//float oCorr = (float)mean(resultO)[0];

		double xCorr, oCorr, dummy;
		minMaxLoc(resultX, &dummy, &xCorr);
		minMaxLoc(resultO, &dummy, &oCorr);
		
		// Decide with symbol was detected
		if (xCorr > oCorr) {
			binary[i] = true ;
			averageCorr += xCorr ;
		}
		else {
			binary[i] = false ;
			averageCorr += oCorr ;
		}

		// If you want to see where frame numbers are read
		//rectangle (background, window, Scalar(0,0,255));

		/*Mat tmpImg;
		background.copyTo(tmpImg);
		rectangle (tmpImg, window, Scalar(0,0,255));
		string msg = to_string(meanValGreen - meanValRed);
		putText(tmpImg, msg, Point(positions[31].x + 15, positions[31].y), FONT_HERSHEY_SIMPLEX, 0.5, Scalar(255, 255, 255), 1);
		imshow("XO", tmpImg);
		waitKey(0);*/
	}
	
	averageCorr /= 32.0;

	/*string msg = to_string(averageCorr);
	putText(background, msg, Point(positions[31].x + 15, positions[31].y + 15), FONT_HERSHEY_SIMPLEX, 0.5, Scalar(255, 255, 255), 1);
	imshow("XO", background);
	waitKey(0);*/

	// Calculate decimal value from X/O bits
	unsigned int decVal = binary.to_ulong();
	
	return make_pair(decVal, (float)averageCorr);
}

vector<vector<double>>
Calibration_Temporal::scanMarkers (Mat background, unsigned int numHorSamplingSteps)
{
	CV_Assert(background.type() == CV_8UC1);

	vector<vector<double>> markerData(5);
	for(int pos = 0; pos < 5; pos++)
	{
		markerData[pos].resize(numHorSamplingSteps);
		
		Rect currBarLeft = markersLeft[pos];
		Rect currBarRight = markersRight[pos];

		CV_Assert(currBarLeft.y == currBarRight.y &&
				  currBarLeft.height == currBarRight.height);

		// Reduce the size of bar by 3 pixels on each side to avoid 
		// reading the background in case of inaccurate spatial alignment
		currBarLeft.x += 3; currBarLeft.y += 3;
		currBarLeft.width -= 6; currBarLeft.height -= 6;
		
		currBarRight.x += 3; currBarRight.y += 3;
		currBarRight.width -= 6; currBarRight.height -= 6;

		// Sample the marker bar
		double samplingHeight = double(currBarLeft.height) / numHorSamplingSteps;

		for (unsigned int i = 0; i < numHorSamplingSteps; i++)
		{
			int roiStartY = cvRound(samplingHeight * i);
			int roiHeight = cvRound(samplingHeight);

			Rect roiLeft(currBarLeft.x, currBarLeft.y + roiStartY, currBarLeft.width, roiHeight);
			Rect roiRight(currBarRight.x, currBarRight.y + roiStartY, currBarRight.width, roiHeight);

			// Calculate average value per ROI/marker
			double meanLeft = mean(background(roiLeft))[0];				
			double meanRight = mean(background(roiRight))[0];
	
			// Calculate and store average of left and right values
			markerData[pos][i] = (meanLeft + meanRight) / 2;

			// If not commented out, the following code shows where the markers are read
			//rectangle(background, roiLeft, Scalar(255));
			//rectangle(background, roiRight, Scalar(255));
		}
	}

	//imshow ("BARS",background);
	//waitKey(0);

	return markerData;
}

void Calibration_Temporal::subtractBlackValue(
	vector<vector<double>> &data)
{
	size_t numHorizontalDivisions = data[0].size();

	for(unsigned int i = 0; i < numHorizontalDivisions; ++i)
	{
		// Black value is in column 4 (= index 3)
		double blackVal = data[3][i];
		
		// Subtract black value, but never fall below 0
		for(unsigned int col = 0; col < data.size(); ++col)
		{
			double subtracted = data[col][i] - blackVal;
			data[col][i] = 0 > subtracted ? 0 : subtracted;
		}
	}
}

void Calibration_Temporal::applyLUTtoMarkers(
	vector<vector<double>> &data)
{
	size_t numHorizontalDivisions = data[0].size();

	for(unsigned int i = 0; i < numHorizontalDivisions; ++i)
	{
		for(unsigned int col = 0; col < data.size(); ++col)
		{
			// Round marker luminance values to integer numbers
			unsigned int valueRounded = cvRound(data[col][i]);

			// Use the look-up-table to get an actual exposure time from the luminance value
			data[col][i] = lut_pixelVal2Exp[i][valueRounded];
		}
	}
}

vector<vector<double>>
Calibration_Temporal::calcInfluenceFromMarkers (vector<vector<double>> markerData,
												bool frame1Valid, bool frame2Valid, bool frame3Valid)
{
	size_t numHorizontalDivisions =
		markerData[0].size();

	vector<vector<double>> influences(3);
	influences[0].resize(numHorizontalDivisions);
	influences[1].resize(numHorizontalDivisions);
	influences[2].resize(numHorizontalDivisions);
	
	for (unsigned int i = 0; i < numHorizontalDivisions; i++)
	{
		double m1Val = markerData[0][i];
		double m2Val = markerData[1][i];
		double m3Val = markerData[2][i];
		//double blackVal = markerData[3][i];
		//double whiteVal = markerData[4][i];

		double totalMarkerVal = 0;
		if(frame1Valid) totalMarkerVal += m1Val;
		if(frame2Valid) totalMarkerVal += m2Val;
		if(frame3Valid) totalMarkerVal += m3Val;

		if (totalMarkerVal == 0) {
			influences[0][i] = 0.0;
			influences[1][i] = 0.0;
			influences[2][i] = 0.0;
		}
		else {
			influences[0][i] = frame1Valid ? m1Val / totalMarkerVal : 0.0;
			influences[1][i] = frame2Valid ? m2Val / totalMarkerVal : 0.0;
			influences[2][i] = frame3Valid ? m3Val / totalMarkerVal : 0.0;
		}
	}

	return influences;
}

void Calibration_Temporal::writeBlendDataAndAvg(
	const vector<vector<double>> &data,
	vector<int> frameNumbers, unsigned int realFrameIdx,
	ofstream &dataStream)
{
	size_t numCols = data.size();
	size_t numHorizontalDivisions = data[0].size();

	// Write CSV header
	for(unsigned int col = 0; col < numCols; col++)
		dataStream << frameNumbers[col] << ";";
	dataStream << endl;

	vector<double> sumInfluences(numCols);
	vector<double> avgInfluences(numCols);
	
	for(unsigned int i = 0; i < numHorizontalDivisions; ++i)
	{
		// Write results to CSV
		for(unsigned int col = 0; col < numCols; col++)
			dataStream << data[col][i] << ";";
		dataStream << endl;

		// Add up influence for average calculation
		for(unsigned int col = 0; col < numCols; col++)
			sumInfluences[col] += data[col][i];
	}

	// Stop here if the average out file is not opened
	if(!avgBlendData.good()) return;

	// Calculate and write average influence for column
	for(unsigned int col = 0; col < numCols; col++)
	{
		if(frameNumbers[col] != -1)
		{
			// Calculate average influence
			avgInfluences[col] = sumInfluences[col] / numHorizontalDivisions;
	
			// Write average influence
			avgBlendData << realFrameIdx << "," << frameNumbers[col] << "," << avgInfluences[col] << "\n" ;
		}
	}	
}

float Calibration_Temporal::getProgress()
{
	return (float)currFrame / totalFrames ;
}

bool Calibration_Temporal::doneProcessing()
{
	return finishedProcessing ;
}
