/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../stdafx.h"

#include "../utility/CRC8.h"
#include "../utility/VideoReader.h"
#include "../utility/SynchronizedQueue.h"
#include "../utility/ThreadCout.h"
#include "../utility/Util.h"
#include "../utility/Img2Cpp.h"

#include "../resources/marker_xo_images.h"

#include "CalibTemporalMarker.h"

using namespace std ;
using namespace cv ;

Calibration_Embed::Calibration_Embed(void)
{}

Calibration_Embed::~Calibration_Embed(void)
{}

void Calibration_Embed::startRegistration(
	string src_file, string dst_file,
	string pat_file, bool createLuminanceTransferVideo )
{
	// Hard-coded parameters
	unsigned int numSyncFrames = 5;
	string filenameMarkerPositions = "_marker_positions.yml";
	string filenameSyncPatternImg = pat_file;

	this->createLuminanceTransferVideo = createLuminanceTransferVideo;
	
	static int nCodec_XVID = CV_FOURCC('X', 'V', 'I', 'D') ;
//	static int nCodec_DIVX = CV_FOURCC('D', 'I', 'V', 'X') ;
//	static int nCodec_UYVY = CV_FOURCC('U', 'Y', 'V', 'Y') ;
	
	// Open input video
	input_vid.open(src_file); if(!input_vid.isOpened())
		throw runtime_error("Input video '" + src_file + "' could not be opened.");

	// Set properties on the file-capture
	input_vid.set(CV_CAP_PROP_CONVERT_RGB, 1);
	
	// Read fps and number of frames properties
	float src_fps = (float)input_vid.get(CV_CAP_PROP_FPS);
	unsigned int numFrames = (unsigned int)input_vid.get(CV_CAP_PROP_FRAME_COUNT);

	// BUG: On linux src_fps is returned at nan. I will jsut set it to 24fps if this happens for now.
	if ( ISNAN( src_fps ) )
	{
		src_fps = 24.0 ;
		TLOG("WARNING: Source fps read incorrectly from the input video. Set to " << src_fps);
	}

	// Read frame size property
	Size srcFrameSize;
	srcFrameSize.width = (int)(input_vid.get(CV_CAP_PROP_FRAME_WIDTH)) ;
	srcFrameSize.height = (int)(input_vid.get(CV_CAP_PROP_FRAME_HEIGHT)) ;

	TLOG("Input video: " << srcFrameSize.width << " x " << srcFrameSize.height << ", "
		 << src_fps << " fps, " << numFrames << " frames.");
	
	totalFrames = numFrames ;

	// Load X/O marker images
	imgX = Img2Cpp::loadImageFromArray_PNG(Img2Cpp::marker_X, Img2Cpp::marker_X_length);
	imgO = Img2Cpp::loadImageFromArray_PNG(Img2Cpp::marker_O, Img2Cpp::marker_O_length);	

	if(imgX.empty() || imgO.empty())
		throw runtime_error("Unable to load X an O images.");

	// Resize X/O marker images
	Size xoMarkerSize(15, 15);
	resize ( imgX, imgX, xoMarkerSize ) ;
	resize ( imgO, imgO, xoMarkerSize ) ;

	// Extract the destination file path
	string path, name;
	Util::splitFilePathAndName(dst_file, path, name);

	// Calculate and store all marker positions
	outVideoSize = calculateMarkerPositions(srcFrameSize, 32, xoMarkerSize);

	// Load the spatial sync frame    
	boost::filesystem::path syncPatternPath(filenameSyncPatternImg);
	syncPatternPath.make_preferred();
	Mat syncPattern = imread (syncPatternPath.string());

	if( syncPattern.empty() ) throw runtime_error(
		"Unable to open '" + filenameSyncPatternImg + "' (the syncronization frame image).") ;
	
	// Resize sync pattern to the full output video size
	resize( syncPattern, syncPattern, outVideoSize ) ;

	// Store marker positions
	saveMarkerData(path + filenameMarkerPositions, filenameSyncPatternImg);

	// Create video file writer
	VideoWriter writer(dst_file, nCodec_XVID, src_fps, outVideoSize);
	if(!writer.isOpened())
		throw runtime_error("VideoWriter for output video could not be initialized.") ;
		
	// Write numSyncFrames sync frames to the beginning of the video
	for ( unsigned int i = 0 ; i < numSyncFrames ; i++ )
		writer << syncPattern ;
	
	// Write a single black frame as separation
	writer << Mat::zeros(outVideoSize, CV_8UC3) ;

	/**** Start Threads ****/
	
	// Reset the current frame counter and other class viables
	currFrame = 0;
	finishedProcessing = false ;
	maxQueueInSize = 10 ;
	maxQueueOutSize = 10 ;

	// Get number of cores for multithreading
	const int numCores = boost::thread::hardware_concurrency() ;
	numThreads = numCores ;

	TLOG("Detected " << numCores << " cores, starting worker threads.");

	// Create a thread for loading the queues with frames
	boost::thread loadQueue( boost::bind( &Calibration_Embed::loadVideoQueues, this ) ) ;
	
	// Create a thread to encode the video
	encodeVideo.create_thread( boost::bind( &Calibration_Embed::writeVideo, this, writer ) ) ;
	
	// Create threads for processing the frames and doing the registration
	for( int i = 0 ; i < numThreads ; i++ )
		processFrames.create_thread( boost::bind( &Calibration_Embed::processFrame, this ) ) ;
}

Size Calibration_Embed::calculateMarkerPositions(Size srcFrameSize,
												 unsigned int symbolCount, Size symbolSize)
{
	// Hard-coded parameters
	Size xo_spacing(5, 5);
	int bar_spaceing = 5;

	// Calculated space required for X/O markers
	unsigned int xo_area_height = 3 * symbolSize.height + 6 * xo_spacing.height;

	// Calculate size of generated marker video
	double ratio = (double)srcFrameSize.width / srcFrameSize.height;

	Size outVideoSize;
	outVideoSize.height = srcFrameSize.height + 2 * xo_area_height;
	outVideoSize.width = cvRound( ratio * outVideoSize.height );

	// Calculate center area
	centerArea = Rect((outVideoSize.width - srcFrameSize.width) / 2, xo_area_height,
					  srcFrameSize.width, srcFrameSize.height);

	// Calculate X/O marker positions
	for(int position = 0; position < 3; position++)
	{
		for(unsigned int symbol = 0; symbol < symbolCount; symbol++)
		{
			int markerWidth = symbolSize.width + xo_spacing.width;
			int markerWidthTotal = symbolCount * markerWidth;
			int markerHeight = symbolSize.height + xo_spacing.height;

			int startX = (outVideoSize.width - markerWidthTotal) / 2 + markerWidth * symbol;
			int startY = 2 * xo_spacing.height + markerHeight * position;

			markersTop[position].push_back(Rect(
				startX, startY,
				symbolSize.width, symbolSize.height ));
		
			markersBottom[position].push_back(Rect(
				startX, startY + centerArea.y + centerArea.height,
				symbolSize.width, symbolSize.height ));
		}
	}

	// Calculate bar marker positions
	int availableWidth = (outVideoSize.width - srcFrameSize.width) / 2;
	int bar_width = (availableWidth - 7 * bar_spaceing) / 5;

	for(int position = 0; position < 5; position++)
	{
		int markerStartLeftX = 2* bar_spaceing + (bar_spaceing + bar_width) * position;
		int markerStartRightX = markerStartLeftX + (centerArea.x + centerArea.width);

		markersLeft[position] = Rect(markerStartLeftX, 0, bar_width, outVideoSize.height);
		markersRight[position] = Rect(markerStartRightX, 0, bar_width, outVideoSize.height);
	}

	return outVideoSize;
}

void Calibration_Embed::saveMarkerData(
	const string filename, const string syncPatternFile)
{
	FileStorage fs(filename.c_str(), FileStorage::WRITE);
	if(!fs.isOpened()) throw runtime_error(
		"Unable to open '" + filename + "' for writing.");

	size_t numFrameMarkerBits = markersTop[0].size();
	CV_Assert(markersBottom[0].size() == numFrameMarkerBits);
	CV_Assert(markersTop[1].size() == numFrameMarkerBits && markersTop[2].size() == numFrameMarkerBits);
	CV_Assert(markersBottom[1].size() == numFrameMarkerBits && markersBottom[2].size() == numFrameMarkerBits);

	fs << "snycPatternFile" << syncPatternFile;
	fs << "markerVideoSize" << outVideoSize;
	fs << "numFrameMarkerBits" << (int)numFrameMarkerBits;
	
	fs << "centerArea" << centerArea;
	
	fs << "top1" << markersTop[0];
	fs << "top2" << markersTop[1];
	fs << "top3" << markersTop[2];
	
	fs << "bottom1" << markersBottom[0];
	fs << "bottom2" << markersBottom[1];
	fs << "bottom3" << markersBottom[2];
	
	fs << "left1" << markersLeft[0];
	fs << "left2" << markersLeft[1];
	fs << "left3" << markersLeft[2];
	fs << "left4" << markersLeft[3];
	fs << "left5" << markersLeft[4];
	
	fs << "right1" << markersRight[0];
	fs << "right2" << markersRight[1];
	fs << "right3" << markersRight[2];
	fs << "right4" << markersRight[3];
	fs << "right5" << markersRight[4];
}
	
void Calibration_Embed::finishRegistration()
{
	
}

void Calibration_Embed::loadVideoQueues()
{
	string tname = "loadVideoQueues" ;
	ThreadCout::tcout.setThreadName(tname);
	
	bool videoEmpty = false ;
	int currInQueueSize ;
	int currOutQueueSize ;
	int currentInputFrame = 0 ;
	int frameIdx ;
	
	while( !videoEmpty )
	{
		currInQueueSize = frameQueueIn.getSize() ;
		currOutQueueSize = frameQueueOut.getSize() ;
		
		if( (currInQueueSize < maxQueueInSize) && (currOutQueueSize < maxQueueOutSize) )
		{       
			FrameWithIdx currentFrame ;
			
			// Get the next frame
			input_vid >> currentFrame.second ;
			
			// When all frames have been processed, set the notifier
			if( currentFrame.second.empty() )
			{
				// Don't load empty data into the queue. This may seem like a redundant step but it will keep
				// the code from entering the 'else' statement which would write an empty frame into the queue
				// and cause the program to crash.
				
				if( currInQueueSize == 0 && currOutQueueSize == 0 )
				{
					videoEmpty = true ;
					
					// Interrupt threads here
					processFrames.interrupt_all() ;
					
					// Join threads here
					processFrames.join_all();
					
					// Now do the same for the encoder. Must be done after the processing threads.
					encodeVideo.interrupt_all() ;
					encodeVideo.join_all() ;
					
					finishedProcessing = true ;
					
					TLOG( "Input video completely processed." )
				}
			}
			else
			{
				// Get the current frame index
				frameIdx = currentInputFrame++ ;
				
				currentFrame.first = frameIdx ;

				// Load the frame into the queue
				frameQueueIn.Enqueue( currentFrame ) ;
				
			}
		}
	}
}


void Calibration_Embed::writeVideo( VideoWriter &_writer )
{
	// Write the data to a file and the frames to a video
	
	string tname = "writeVideo"  ;
	ThreadCout::tcout.setThreadName(tname);
	
	// Pulls a frame from the queue and writes it to a video
	unsigned int sizeOut = 0 ;
	int nextFrame = 0 ;
	VideoWriter &writer = _writer ;
	
	
	FrameWithIdx frameDat ;
	vector<FrameWithIdx>::iterator it ;
	vector<FrameWithIdx> frameStore ;
	
	while(true)
	{
		sizeOut = frameQueueOut.getSize() ;
		
		while(sizeOut > 0)
		{
			frameDat = frameQueueOut.Dequeue() ;
			
			if( frameDat.first == nextFrame )
			{
				while(frameDat.first == nextFrame)
				{
					writer << frameDat.second ;
					
					nextFrame++ ;
					
					if(frameStore.size())
					{
						for(it = frameStore.begin() ; it != frameStore.end() ; ++it)
						{
							if(it->first == nextFrame)
							{
								frameDat = *it ;
								frameStore.erase(it) ;
								break ; // break out of for loop
							}
						}
					}
				}
			}
			else
				frameStore.push_back(frameDat) ;
			
			sizeOut = frameQueueOut.getSize() ;
		}
		
		// Call interruption point here so the thread can be killed later
		boost::this_thread::interruption_point();
	}
}

void Calibration_Embed::processFrame()
{
	// This function is bound to a thread. It just repeats the same loop over and over.
	// It will dequeue a frame and them apply the temporal and all subsequent distortions.
	// No actually processing is done here. It just calls other functions that actually
	// manipulate the frames.
	
	static atomic<int> thId(0);
	string tname = "processFrame:" + to_string(thId++) ;
	ThreadCout::tcout.setThreadName(tname);

	// Iterate over all frames 
	while(true)
	{
		// Create a black background
		Mat background = Mat::zeros( outVideoSize.height, outVideoSize.width, CV_8UC3 ) ;

		// Read input video frame
		FrameWithIdx deqframe = frameQueueIn.Dequeue() ;
		int frameIdx = deqframe.first ;

		// -->> COMMENT OUT THE FOLLOWING FOR A MARKER-ONLY VIDEO <<--
		Mat background_roi (background, centerArea);
		deqframe.second.copyTo(background_roi);
		
		// Crc der aktuellen Framenummer berechnen und anh‰ngen
		CRC8 curCRC;
		unsigned int frameIdx_crc = curCRC.append_24bit_8bitcrc(frameIdx);
		
		int position;
		if (frameIdx % 3 == 0)      position = 0;
		else if (frameIdx % 3 == 1) position = 1;
		else                        position = 2;

		// Draw frame number via X/O markers
		drawXOMarkers( background, markersTop[position], frameIdx_crc ) ;
		drawXOMarkers( background, markersBottom[position], frameIdx_crc ) ;
		
		// Draw frame number (human-readable)
		string strFrameNo = to_string(frameIdx);
		
		Point textPos(markersTop[position][0].x - 100, markersTop[position][0].y + markersTop[position][0].height);
		putText(background, strFrameNo, textPos, FONT_HERSHEY_SIMPLEX, 0.5, Scalar(255, 255, 255), 1);
		
		textPos = Point(markersBottom[position][0].x - 100, markersBottom[position][0].y + markersBottom[position][0].height);
		putText(background, strFrameNo, textPos, FONT_HERSHEY_SIMPLEX, 0.5, Scalar(255, 255, 255), 1);

		// Marker rows
		unsigned char grayValue = 128;
		if(createLuminanceTransferVideo)
		{
			for(int i = 0; i < 5; i++)
			{
				// Draw every second marker row
				if(i % 2 == 0)
				{
					rectangle( background, markersLeft[i], Scalar::all(grayValue), -1 ) ;
					rectangle( background, markersRight[i], Scalar::all(grayValue), -1 ) ;
				}
			}
		}
		else
		{
			// Draw marker rows at current position
			rectangle( background, markersLeft[position], Scalar::all(grayValue), -1 ) ;
			rectangle( background, markersRight[position], Scalar::all(grayValue), -1 ) ;

			// Draw the 'alway on'-marker rows
			rectangle( background, markersLeft[4], Scalar::all(grayValue), -1 ) ;
			rectangle( background, markersRight[4], Scalar::all(grayValue), -1 ) ;
		}

		FrameWithIdx outputFrame ;
		outputFrame.first = frameIdx ;
		outputFrame.second = background ;
		
		frameQueueOut.Enqueue( outputFrame ) ;
		
		// Atomically increase the current frame
		currFrame++ ;
	}
}

void Calibration_Embed::drawXOMarkers(Mat &background, vector<Rect> positions, unsigned int decimal)
{
	CV_Assert(positions.size() == 32);
	
	bitset<32> binary(decimal);	

	for (int i = 0; i < 32; i++)
	{
		Mat background_roi(background, positions[i]);
		
		if (binary[i])	imgX.copyTo(background_roi);
		else			imgO.copyTo(background_roi);
	}
}

float Calibration_Embed::getProgress()
{
	return (float)currFrame / totalFrames ;
}

bool Calibration_Embed::doneProcessing()
{
	return finishedProcessing ;
}
