/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#ifndef __FRAMESEQUENCE__
#define __FRAMESEQUENCE__

#include "../utility/sorted_interval_vector.h"

template <class FrameType>
class FrameSequence
{
public:
	// Helper struct typedefs
	typedef typename sorted_interval_vector<FrameType, double>::SeqAccessCache SeqAccessCache;

	// Iterator typedefs
	typedef typename sorted_interval_vector<FrameType, double>::iterator iterator;
	typedef typename sorted_interval_vector<FrameType, double>::const_iterator const_iterator;

public:
	// Insertion
	void add(const FrameType &frame)
	{
		// Use frame start/end times as interval bounds
		frames.insert(	frame.getStartTime(),
						frame.getEndTime(),
						frame	);
	}

	// Validation
	void validate()
	{
		// Sort and validate that frame intervals meet
		// the conditions for the sorted_interval_vector.
		if(!frames.validate()) throw std::runtime_error(
			"Internal error: Invalid frame intervals detected.");
	}

	// Returns the number of frames
	size_t size() const	{	return frames.size();	}

	// Access through iterators
	iterator begin()	{	return frames.begin();	}
	iterator end()		{	return frames.end();	}

	// Access frame by index
	FrameType operator[] (size_t idx) const
	{	return frames[idx];		}

	// Returns the number of frames
	unsigned int getNumFrames() const
	{	return (unsigned int)frames.size(); }

	// Accesses (possibly overlapping) frames by time
	std::vector<FrameType> getFrames(double time) const
	{	return frames.get(time);	}

	// Accesses (possibly overlapping) frames by time, optimized for sequencial access
	std::vector<FrameType> getFrames_seq(double time, SeqAccessCache &cache) const
	{	return frames.get_seq(time, cache);	}

	// Gets a list of frames within the given time range
	std::vector<FrameType> getFrameRange(double start_time, double end_time) const
	{	return frames.get_range(start_time, end_time);	}
	
	// Gets a list of frames within the given time range, optimized for sequencial access
	std::vector<FrameType> getFrameRange_seq(double start_time, double end_time, SeqAccessCache &cache) const
	{	return frames.get_range_seq(start_time, end_time, cache);	}
	
protected:
	// Store frames in a sorted interval vector
	sorted_interval_vector<FrameType, double> frames;
};

#endif //__FRAMESEQUENCE__