/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../stdafx.h"

#include "../utility/Util.h"
#include "../utility/Version.h"

#include "Scenario.h"

using namespace std;
using namespace cv;
using namespace pugi;
using namespace boost::filesystem;

// Static member initialization
const char* Scenario::AutomaticGainControl::methodNames[] = { "Mean-Threshold-Offset-FixAdjustment", "Mean-Threshold-Offset-PecentualAdjustment" };
const char* Scenario::AutomaticWhiteBalance::methodNames[] = { "Grayworld", "Weng" };
const char* Scenario::ColorFilterArray::methodNames[] = { "SimpleBayer" };
const char* Scenario::Defocus::methodNames[] = { "GaussianBlur" };

Scenario::Scenario(Version _camMarkVersion)
	: camMarkVersion(_camMarkVersion)
{
	applyTemporalDistortion = false;
	applySpatialDistortion = false;
	applyAWBDistortion = false;
	applyAGCDistortion = false;
	applyCFADistortion = false;
	applyDefocus = false;

	override_src_FPS = 0;
	cam_start_time = 0;
	cam_end_time = 0;

	spatial.display_resolution = Size(0,0);
	spatial.camera_resolution = Size(0,0);
	spatial.ar_mode = AR_LETTERBOX;
	spatial.extendBorderPx = 0;
	spatial.fadeBorderPx = 0;
	spatial.backgroundInsertion = false;

	temporal.cam_FPS = 0;
	temporal.display_refresh_rate = 0;
	temporal.display_g2g_time = 0;
	temporal.display_pwm_duty_cycle = 1.0;
	temporal.display_pwm_frequency = 180.0;
	temporal.exposure_time_ms = 0;
	temporal.readout_time_ms = 0;
	temporal.vsync = false;

	agc.method = AGC_MEAN_THRESHOLD_OFFSET_FIX_ADJUST;
	agc.frameCountLower = agc.frameCountUpper = 0;
	agc.threshLower = agc.threshUpper = 0;
	agc.adjustLower = agc.adjustUpper = 0;

	awb.method = AWB_GRAYWORLD;
	
	cfa.method = CFA_SIMPLE_BAYER;
	cfa.aafilter = false;

	defocus.method = DEFOCUS_GAUSS;
	defocus.kernel_size = 0;
	defocus.sigma = 0;

	// Add the current path as search path for referenced files
	addFileSearchPath(current_path());
}

Scenario::~Scenario(void)
{
}

void Scenario::addFileSearchPath(path searchPath)
{
	// Add specified search path
	searchPaths.push_back(searchPath);
}

void Scenario::readXMLFromFile(string filename)
{
	xml_document doc;
	xml_parse_result result = doc.load_file(filename.c_str());

	// Check success
	if(!result) throw runtime_error(
		"Error while opening '" + filename + "' XML scenario file:\n" + result.description());

	// Determine the folder in with the scenario file resides
	// and add it as a possible search path for referenced files.
	path settingsFileFolder = path(filename).parent_path();
	addFileSearchPath(settingsFileFolder);
	
	// Parse content
	readXML(doc);
}

/*void Scenario::readXMLFromString(std::string xmlString)
{
	xml_document doc;
	

	// Parse content
	readXML(doc, path("."));
}*/

void Scenario::readXML(xml_document &doc)
{
	// Get scenario root node
	xml_node scenario = doc.child("Scenario");
	if(!scenario) throw runtime_error(
		"The specified XML does not contain a CamMark scenario.\n"
		"(The required <Scenario> root tag was not found.)");

	// Check if a source/output filename was specified
	// in the scenario file (uncommon, but possible)
	sourceFile = scenario.child("Source").attribute("file").as_string();
	outputFile = scenario.child("Output").attribute("file").as_string();

	// GENERAL
	xml_node src_fps = scenario.child("OverrideSourceFPS");
	override_src_FPS = src_fps.attribute("value").as_double();
	
	xml_node timespan = scenario.child("CaptureTimespan");
	cam_start_time = timespan.attribute("start-time").as_double();
	cam_end_time = timespan.attribute("end-time").as_double();

	// SPATIAL DISTORTION
	xml_node node_spatial = scenario.child("SpatialDistortion");
	if(node_spatial && node_spatial.attribute("enabled").as_bool() == true)
	{
		applySpatialDistortion = true;
		spatial.presetMeta = loadAndMergePreset(node_spatial);
		readSpatial(node_spatial);
	}
	else applySpatialDistortion = false;

	// TEMPORAL DISTORTION
	xml_node node_temporal = scenario.child("TemporalDistortion");
	if(node_temporal && node_temporal.attribute("enabled").as_bool() == true)
	{
		applyTemporalDistortion = true;
		temporal.presetMeta = loadAndMergePreset(node_temporal);
		readTemporal(node_temporal);
	}
	else applyTemporalDistortion = false;

	// AUTOMATIC GAIN CONTROL
	xml_node node_agc = scenario.child("AutomaticGainControl");
	if(node_agc && node_agc.attribute("enabled").as_bool() == true)
	{
		applyAGCDistortion = true;
		agc.presetMeta = loadAndMergePreset(node_agc);
		readAGC(node_agc);
	}
	else applyAGCDistortion = false;

	// AUTOMATIC WHITE BALANCE
	xml_node node_awb = scenario.child("AutomaticWhiteBalance");
	if(node_awb && node_awb.attribute("enabled").as_bool() == true)
	{
		applyAWBDistortion = true;
		awb.presetMeta = loadAndMergePreset(node_awb);
		readAWB(node_awb);
	}
	else applyAWBDistortion = false;

	// BAYER PATTERN SIMULATION
	xml_node node_cfa = scenario.child("ColorFilterArray");
	if(node_cfa && node_cfa.attribute("enabled").as_bool() == true)
	{
		applyCFADistortion = true;
		cfa.presetMeta = loadAndMergePreset(node_cfa);
		readCFA(node_cfa);
	}
	else applyCFADistortion = false;

	// DEFOCUS
	xml_node node_defocus = scenario.child("Defocus");
	if(node_defocus && node_defocus.attribute("enabled").as_bool() == true)
	{
		applyDefocus = true;
		defocus.presetMeta = loadAndMergePreset(node_defocus);
		readDefocus(node_defocus);
	}
	else applyDefocus = false;
}

void Scenario::readSpatial(pugi::xml_node &node)
{
	// Display resolution tag
	xml_node disp_resolution = node.child("DisplayResolution");
	xml_attribute disp_width_attr = disp_resolution.attribute("width");
	xml_attribute disp_height_attr = disp_resolution.attribute("height");

	Util::ci_string disp_width_string = disp_width_attr.as_string();
	if(disp_width_string == "source") spatial.display_resolution.width = -1;
	else if(disp_width_string == "transform") spatial.display_resolution.width = -2;
	else spatial.display_resolution.width = disp_width_attr.as_uint();

	Util::ci_string disp_height_string = disp_height_attr.as_string();
	if(disp_height_string == "source") spatial.display_resolution.height = -1;
	else if(disp_height_string == "transform") spatial.display_resolution.height = -2;
	else spatial.display_resolution.height = disp_height_attr.as_uint();
			
	xml_attribute ar_mode_attr = disp_resolution.attribute("aspect-ratio-mode");
	if(hasDefinedValue(ar_mode_attr))
	{
		Util::ci_string ar_mode_str = ar_mode_attr.as_string();
		if(ar_mode_str == "resize") spatial.ar_mode = AR_RESIZE;
		else if(ar_mode_str == "letterbox") spatial.ar_mode = AR_LETTERBOX;
		else if(ar_mode_str == "crop") spatial.ar_mode = AR_CROP;
		else throw runtime_error(
			"Scenario: DisplayResolution: Invalid aspectRatioMode.\n"
			"Valid modes are: 'resize', 'letterbox', and 'crop'.");
	}
	// Default is letterboxing
	else spatial.ar_mode = AR_LETTERBOX;

	// Camera resolution tag
	xml_node cam_resolution = node.child("CameraResolution");
	xml_attribute cam_width_attr = cam_resolution.attribute("width");
	xml_attribute cam_height_attr = cam_resolution.attribute("height");

	Util::ci_string cam_width_string = cam_width_attr.as_string();
	if(cam_width_string == "source") spatial.camera_resolution.width = -1;
	else if(cam_width_string == "calib") spatial.camera_resolution.width = -2;
	else spatial.camera_resolution.width = cam_width_attr.as_uint();

	Util::ci_string cam_height_string = cam_height_attr.as_string();
	if(cam_height_string == "source") spatial.camera_resolution.height = -1;
	else if(cam_height_string == "calib") spatial.camera_resolution.height = -2;
	else spatial.camera_resolution.height = cam_height_attr.as_uint();

	// Transform file tag
	xml_node transformFile = node.child("TransformationFile");
	if(transformFile)
	{
		path configFilePath = transformFile.attribute("path").as_string();
		spatial.transform_file = findFileOrThrow(configFilePath);

		spatial.backgroundInsertion =
			transformFile.attribute("background-insertion").as_bool();
	}
	else
	{
		// An empty config file path will generate an identity transformation
		spatial.transform_file.clear();
		spatial.backgroundInsertion = false;
	}

	// Soften borders tag
	xml_node softenBorders = node.child("SoftenBorders");
	if(softenBorders)
	{
		xml_attribute extendBorderPxAttr = softenBorders.attribute("extend-border-px");
		xml_attribute fadeBorderPxAttr = softenBorders.attribute("fade-border-px");
			
		if(hasDefinedValue(extendBorderPxAttr))
			 spatial.extendBorderPx = extendBorderPxAttr.as_uint();
		else spatial.extendBorderPx = 2; // If attribute is not specified, use 3 as default

		if(hasDefinedValue(fadeBorderPxAttr))
			 spatial.fadeBorderPx = fadeBorderPxAttr.as_uint();
		else spatial.fadeBorderPx = spatial.extendBorderPx + 1; // If attribute is not specified, use default
	}
	else
	{
		spatial.extendBorderPx = 0;
		spatial.fadeBorderPx = 0;
	}
}

void Scenario::readTemporal(pugi::xml_node &node)
{
	// Display tag
	xml_node display = node.child("Display");
	xml_attribute display_rr, display_fading, display_vsync;
	if(display)
	{
		display_rr = display.attribute("refresh-rate");
		display_fading = display.attribute("fading-time");
		display_vsync = display.attribute("vsync");
	}

	temporal.display_refresh_rate = display_rr.as_double();
	temporal.display_g2g_time = display_fading.as_double();
	
	if(hasDefinedValue(display_vsync))
		 temporal.vsync = display_vsync.as_bool();
	else temporal.vsync = true;

	// LCDBacklightDimming tag
	xml_node lcdBacklightDimming = node.child("LCDBacklightDimming");
	xml_attribute pwmDutyCycle, pwmFrequency;
	if(lcdBacklightDimming)
	{
		pwmDutyCycle = lcdBacklightDimming.attribute("pwm-duty-cycle");
		pwmFrequency = lcdBacklightDimming.attribute("pwm-frequency");
	}

	if(hasDefinedValue(pwmDutyCycle))
		 temporal.display_pwm_duty_cycle = pwmDutyCycle.as_double();
	else temporal.display_pwm_duty_cycle = 1.0;

	if(temporal.display_pwm_duty_cycle > 1.0 || temporal.display_pwm_duty_cycle < 0.0)
		throw runtime_error("Scenario: LCDBacklightDimming: pwm-duty-cycle must be between 0.0 and 1.0 (corresponding to 0% and 100%)");

	if(hasDefinedValue(pwmFrequency))
		 temporal.display_pwm_frequency = pwmFrequency.as_double();
	else temporal.display_pwm_frequency = 180.0;

	// Capture tag
	xml_node capture = node.child("Capture");
	xml_attribute cap_fps, cap_exposure_time, cap_readout_time;
	if(capture)
	{
		cap_fps = capture.attribute("fps");
		cap_exposure_time = capture.attribute("exposure-time");
		cap_readout_time = capture.attribute("readout-time");
	}

	temporal.cam_FPS = cap_fps.as_double();
	temporal.exposure_time_ms = cap_exposure_time.as_double();
	temporal.readout_time_ms = cap_readout_time.as_double();

	// BlendFile tag
	xml_node configFile = node.child("BlendFile");
	if(configFile)
	{
		path configFilePath = configFile.attribute("path").as_string();
		temporal.blend_file = findFileOrThrow(configFilePath);
	}
}

void Scenario::readAGC(pugi::xml_node &node)
{
	// Method
	xml_node method = node.child("Method");

	Util::ci_string agc_method_string;
	if(method) agc_method_string = method.attribute("name").as_string();
	else throw runtime_error("Scenario: AutomaticGainControl: The <Method> tag must be specified.");

	if(agc_method_string == agc.methodNames[AGC_MEAN_THRESHOLD_OFFSET_FIX_ADJUST]) agc.method = AGC_MEAN_THRESHOLD_OFFSET_FIX_ADJUST;
	else if(agc_method_string == agc.methodNames[AGC_MEAN_THRESHOLD_OFFSET_PERC_ADJUST]) agc.method = AGC_MEAN_THRESHOLD_OFFSET_PERC_ADJUST;
	else throw runtime_error("Scenario: AutomaticGainControl: Method name must be '"
		+ string(agc.methodNames[AGC_MEAN_THRESHOLD_OFFSET_FIX_ADJUST]) + "' or '"
		+ string(agc.methodNames[AGC_MEAN_THRESHOLD_OFFSET_PERC_ADJUST]) + "'.");

	// Parameters
	xml_node lower_bound = node.child("LowerBound");
	xml_node upper_bound = node.child("UpperBound");

	if(!lower_bound) throw runtime_error("Scenario: AutomaticGainControl: no 'LowerBound' specified");
	if(!upper_bound) throw runtime_error("Scenario: AutomaticGainControl: no 'UpperBound' specified");

	xml_attribute attr_count, attr_threshLum, attr_adjustBy;

	// Lower bound
	attr_count = lower_bound.attribute("count");
	if(hasDefinedValue(attr_count)) agc.frameCountLower = attr_count.as_uint();
	else throw runtime_error("Scenario: LowerBound: no 'count' specified, or value is empty");

	attr_threshLum = lower_bound.attribute("thresh");
	if(hasDefinedValue(attr_threshLum)) agc.threshLower = attr_threshLum.as_float();
	else throw runtime_error("Scenario: LowerBound: no 'threshLuminance' specified, or value is empty");

	attr_adjustBy = lower_bound.attribute("adjust");
	if(hasDefinedValue(attr_adjustBy)) agc.adjustLower = attr_adjustBy.as_float();
	else throw runtime_error("Scenario: LowerBound: no 'adjustBy' specified, or value is empty");

	// Upper bound
	attr_count = upper_bound.attribute("count");
	if(hasDefinedValue(attr_count)) agc.frameCountUpper = attr_count.as_uint();
	else throw runtime_error("Scenario: UpperBound: no 'count' specified, or value is empty");

	attr_threshLum = upper_bound.attribute("thresh");
	if(hasDefinedValue(attr_threshLum)) agc.threshUpper = attr_threshLum.as_float();
	else throw runtime_error("Scenario: UpperBound: no 'threshLuminance' specified, or value is empty");

	attr_adjustBy = upper_bound.attribute("adjust");
	if(hasDefinedValue(attr_adjustBy)) agc.adjustUpper = attr_adjustBy.as_float();
	else throw runtime_error("Scenario: UpperBound: no 'adjustBy' specified, or value is empty");
}

void Scenario::readAWB(pugi::xml_node &node)
{
	xml_node method = node.child("Method");

	Util::ci_string awb_method_string;
	if(method) awb_method_string = method.attribute("name").as_string();
	else throw runtime_error("Scenario: AutomaticWhiteBalance: The <Method> tag must be specified.");

	if(awb_method_string == AutomaticWhiteBalance::methodNames[AWB_GRAYWORLD]) awb.method = AWB_GRAYWORLD;
	else if(awb_method_string == AutomaticWhiteBalance::methodNames[AWB_WENG]) awb.method = AWB_WENG;
	else throw runtime_error("Scenario: AutomaticWhiteBalance: Method name must be '"
		+ string(AutomaticWhiteBalance::methodNames[AWB_GRAYWORLD]) + "' or '"
		+ string(AutomaticWhiteBalance::methodNames[AWB_WENG]) + "'.");
}

void Scenario::readCFA(pugi::xml_node &node)
{
	xml_node method = node.child("Method");

	Util::ci_string cfa_method_string;
	if(method) cfa_method_string = method.attribute("name").as_string();
	else throw runtime_error("Scenario: ColorFilterArray: The <Method> tag must be specified.");

	if(cfa_method_string == ColorFilterArray::methodNames[CFA_SIMPLE_BAYER]) cfa.method = CFA_SIMPLE_BAYER;
	else throw runtime_error("Scenario: ColorFilterArray: Currently, '"
		+ string(ColorFilterArray::methodNames[CFA_SIMPLE_BAYER])
		+ "' is the only supported method.");

	xml_node aafilter = node.child("AAFilter");
	if(aafilter) cfa.aafilter = aafilter.attribute("enabled").as_bool();
}

void Scenario::readDefocus(pugi::xml_node &node)
{
	const char* gaussMethodName = Defocus::methodNames[DEFOCUS_GAUSS];
	xml_node gblur = node.child(gaussMethodName);
	if(gblur)
	{
		defocus.method = DEFOCUS_GAUSS;
		defocus.kernel_size = gblur.attribute("kernel-size").as_uint();
		defocus.sigma = gblur.attribute("sigma").as_double();

		if(defocus.kernel_size == 0 && defocus.sigma == 0) throw runtime_error(
			"Scenario: GaussianBlur: Either the kernel size, or sigma, or both must be specified.");

		if(defocus.kernel_size > 0 && defocus.kernel_size % 2 == 0) throw runtime_error(
			"Scenario: GaussianBlur: The specified kernel size must be odd, or zero for automatic calculation from sigma.");
	}
	else
	{
		defocus.kernel_size = 0;
		defocus.sigma = 0;
		applyDefocus = false;
	}
}

string Scenario::findFileOrThrow(path file)
{
	// Check specified file path directly
	if(exists(file) && !is_directory(file))
		return file.make_preferred().string();

	// Try using each of the specified search paths
	for(path searchPath : searchPaths)
	{
		// Make an absolute path using the current search path
		path absolutePath = absolute(file, searchPath);
		if(exists(absolutePath) && !is_directory(absolutePath))
		{
			// Make path canonical and return system-preferred string
			path canonicalPath = canonical(absolutePath);
			return canonicalPath.make_preferred().string();
		}
	}

	// If not found, throw an exception
	throw runtime_error("CamMark scenario: file '"
		+ file.string() + "' cannot be found.");
}

Scenario::PresetMetaData Scenario::loadAndMergePreset(xml_node &node)
{
	// Find preset attribute
	xml_attribute preset_attribute = node.attribute("preset");
	if(!preset_attribute) return PresetMetaData();

	// Get the file name as string
	string preset_file = preset_attribute.as_string();
	if(preset_file.empty()) return PresetMetaData();

	// Find the referenced preset file or fail
	preset_file = findFileOrThrow(preset_file);

	// Get the folder the preset resides in and
	// add it as search path for referenced files.
	path presetFileFolder = path(preset_file).parent_path();
	addFileSearchPath(presetFileFolder);

	// Load and parse preset file
	xml_document preset_doc;
	xml_parse_result result = preset_doc.load_file(preset_file.c_str());

	// Get the filename only for error messages
	string presetFileNameOnly = path(preset_file).filename().string();

	// Check success
	if(!result) throw runtime_error(
		"Error while opening preset file '" + presetFileNameOnly + "':\n" + result.description());

	// Check for the "Preset" root element
	xml_node preset_root = preset_doc.child("Preset");
	if(!preset_root) throw runtime_error(
		"The specified preset file '" + presetFileNameOnly + "' does not contain a <Preset> root element.");

	// Read preset meta data
	PresetMetaData meta;
	meta.presetFile = preset_file;

	string version = preset_root.attribute("version").as_string();
	string compatibility = preset_root.attribute("compatibility").as_string();

	if(!version.empty()) meta.version = Version(version);
	if(!compatibility.empty()) meta.compatibility = Version(compatibility);

	xml_node desc = preset_root.child("Description");
	if(desc)
	{
		string desc_string(desc.child_value());
		meta.desc = Util::trim(desc_string);
	}

	// Check preset compatibility
	if(camMarkVersion < meta.compatibility)
	{
		stringstream compatibleVersion;
		compatibleVersion << meta.compatibility;

		throw runtime_error(
			"The specified preset file '" + presetFileNameOnly + "' is not compatible with this version of CamMark. " +
			"Please use CamMark version " + compatibleVersion.str() + " or newer in order to used the specified preset.");
	}

	// Append all child nodes that are not already in the node.
	// This way, any preset tag can be overwritten.
	for(const xml_node &preset_child : preset_root)
	{
		if(!node.child(preset_child.name()))
			node.append_copy(preset_child);
	}

	// Return read meta data
	return meta;
}

void Scenario::printToStream(std::ostream &os)
{
	os << "*** GENERAL ***" << endl;

	if(override_src_FPS > 0)
		os << "    Source frame rate override: " << override_src_FPS << " fps" << endl;

	string startTime = cam_start_time > 0 ? to_string(cam_start_time) + " sec" : "<start of source video>";
	string endTime = cam_end_time > 0 ? to_string(cam_end_time) + " sec" : "<end of source video>";
	os << "    Capture start time: " << startTime << endl;
	os << "    Capture end time:   " << endTime << endl;
	
	os << endl;

	//----------------------------------------------------
	os << "*** SPATIAL ***" << endl;
	if(applySpatialDistortion == false) os << "    NOT applied." << endl;
	else
	{
		printUsedPresetInfo(spatial.presetMeta, os);

		string displayWidth = spatial.display_resolution.width > 0 ? to_string(spatial.display_resolution.width) : "<unspecified>";
		string displayHeight = spatial.display_resolution.height > 0 ? to_string(spatial.display_resolution.height) : "<unspecified>";
		os << "    Simulated display resolution: " << displayWidth << " x " << displayHeight << endl;
		
		string aspectRatioMode;
		switch(spatial.ar_mode)
		{
		case AR_RESIZE: aspectRatioMode = "Resize"; break;
		case AR_LETTERBOX: aspectRatioMode = "Letterbox"; break;
		case AR_CROP: aspectRatioMode = "Crop"; break;
		default: aspectRatioMode = "[unknown]"; 
		}
		os << "    Simulated playback aspect ratio mode: " << aspectRatioMode << endl;

		string camWidth = spatial.camera_resolution.width > 0 ? to_string(spatial.camera_resolution.width) : "<unspecified>";
		string camHeight = spatial.camera_resolution.height > 0 ? to_string(spatial.camera_resolution.height) : "<unspecified>";
		os << "    Simulated camera resolution: " << camWidth << " x " << camHeight << endl;

		string extendBorder = spatial.extendBorderPx > 0 ? "extend border by " + to_string(spatial.extendBorderPx) + " px" : "no border extension";
		string fadeBorder = spatial.fadeBorderPx > 0 ? "fade border by " + to_string(spatial.fadeBorderPx) + " px" : "no border fading";
		os << "    Border softening: " << extendBorder << ", " << fadeBorder << endl;

		string transformation;
		if(spatial.transform_file.empty()) transformation = "no transformation";
		else transformation = path(spatial.transform_file).filename().string();

		string insertBg = spatial.backgroundInsertion ? "yes" : "no";
		os << "    Spatial transformation: " << transformation << endl;
		os << "    Background insertion : " << insertBg << endl;
	}
	
	os << endl;

	//----------------------------------------------------
	os << "*** TEMPORAL ***" << endl;
	if(applyTemporalDistortion == false) os << "    NOT applied." << endl;
	else
	{
		printUsedPresetInfo(temporal.presetMeta, os);

		string displayRR = temporal.display_refresh_rate > 0 ?  to_string(temporal.display_refresh_rate) + " Hz" : "<as frame rate of source>";
		os << "    Simulated display refresh rate: " << displayRR << endl;
		os << "    Simulated display fade time: " << temporal.display_g2g_time << " ms" << endl;
		os << "    Simulate v-sync mechanism: " << (temporal.vsync ? "yes" : "no") << endl;
		
		bool simLCDBacklightDimming = temporal.display_pwm_duty_cycle < 1.0;
		os << "    Simulate LCD backlight dimming: " << (simLCDBacklightDimming ? "yes" : "no") << endl;
		if(simLCDBacklightDimming)
		{
			os << "    LCD backlight PWM frequency: " << temporal.display_pwm_frequency << " Hz" << endl;
			os << "    LCD backlight PWM duty-cycle: " << temporal.display_pwm_duty_cycle * 100 << "%" << endl;
		}

		string camFPS = temporal.cam_FPS > 0 ?  to_string(temporal.cam_FPS) + " fps" : "<as frame rate of source>";
		os << "    Simulated capture frame rate: " << camFPS << endl;

		string exposure = temporal.exposure_time_ms > 0 ?  to_string(temporal.exposure_time_ms) + " ms" : "<calculated as 1/2 of frame duration>";
		os << "    Simulated exposure time: " << exposure << endl;

		bool simRollingShutter = temporal.readout_time_ms > 0.0;
		os << "    Simulate rolling shutter mechanism: " << (simRollingShutter ? "yes" : "no") << endl;
		if(simRollingShutter)
			os << "    Simulated rolling shutter readout time: " << temporal.readout_time_ms << " ms" << endl;
	}
	
	os << endl;

	//----------------------------------------------------
	os << "*** AUTOMATIC GAIN CONTROL ***" << endl;
	if(applyAGCDistortion == false) os << "    NOT applied." << endl;
	else
	{
		printUsedPresetInfo(agc.presetMeta, os);

		string agcMethod = AutomaticGainControl::methodNames[agc.method];
		string adjustByLower, adjustByUpper;

		switch(agc.method)
		{
		case AGC_MEAN_THRESHOLD_OFFSET_FIX_ADJUST:
			adjustByLower = to_string(agc.adjustLower);
			adjustByUpper = to_string(agc.adjustUpper);
			break;
		
		case AGC_MEAN_THRESHOLD_OFFSET_PERC_ADJUST:
			adjustByLower = to_string(agc.adjustLower * 100) + "%";
			adjustByUpper = to_string(agc.adjustUpper * 100) + "%";
			break;
		}
		
		os << "    Applied method: " << agcMethod << endl;
		os << "    If for " << agc.frameCountLower << " frames below brightness of " << agc.threshLower << ", adjust by " << adjustByLower << endl;
		os << "    If for " << agc.frameCountUpper << " frames above brightness of " << agc.threshUpper << ", adjust by " << adjustByUpper << endl;
	}
	
	os << endl;

	//----------------------------------------------------
	os << "*** AUTOMATIC WHITE BALANCE ***" << endl;
	if(applyAWBDistortion == false) os << "    NOT applied." << endl;
	else
	{
		printUsedPresetInfo(awb.presetMeta, os);

		string awbMethod = AutomaticWhiteBalance::methodNames[awb.method];
		os << "    Applied method: " << awbMethod << endl;
	}
	
	os << endl;

	//----------------------------------------------------
	os << "*** COLOR FILTER ARRAY ***" << endl;
	if(applyCFADistortion == false) os << "    NOT applied." << endl;
	else
	{
		printUsedPresetInfo(cfa.presetMeta, os);

		string cfaMethod = ColorFilterArray::methodNames[cfa.method];
		os << "    Applied method: " << cfaMethod << endl;
		os << "    Simulate AA-filter: " << (cfa.aafilter ? "yes" : "no") << endl;
	}
	
	os << endl;

	//----------------------------------------------------
	os << "*** DEFOCUS ***" << endl;
	if(applyDefocus == false) os << "    NOT applied." << endl;
	else
	{
		printUsedPresetInfo(defocus.presetMeta, os);

		string defocusMethod = Defocus::methodNames[defocus.method];
		os << "    Applied method: " << defocusMethod << endl;
		
		string kernelSize = defocus.kernel_size > 0 ? to_string(defocus.kernel_size) : "<calculated from sigma>";
		os << "    Kernel size: " << kernelSize << endl;
		
		string sigma = defocus.sigma > 0 ? to_string(defocus.sigma) : "<calculated from kernel size>";
		os << "    Sigma: " << sigma << endl;
	}
}

void Scenario::printUsedPresetInfo(PresetMetaData &presetMeta, std::ostream &os)
{
	if(!presetMeta.presetFile.empty())
	{
		string presetFileNameOnly = path(presetMeta.presetFile).filename().string();
		os << "    Settings based on preset: " << presetFileNameOnly << endl;

		if(Version() < presetMeta.version)
			os << "    Preset version: " << presetMeta.version << endl;

		if(Version() < presetMeta.compatibility)
			os << "    Preset compatibility: CamMark " << presetMeta.compatibility << " and newer" << endl;

		if(!presetMeta.desc.empty())
			os << "    Preset description: \"" << presetMeta.desc << "\"" << endl;
	}
}