/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#ifndef __SCENARIO__
#define __SCENARIO__

// Forward declaration
namespace pugi {
	class xml_document;
}

class Scenario
{
public:
	enum AspectRatioMode
	{
		AR_RESIZE = 0,
		AR_LETTERBOX,
		AR_CROP
	};

	enum AGCMethod
	{
		AGC_MEAN_THRESHOLD_OFFSET_FIX_ADJUST = 0,
		AGC_MEAN_THRESHOLD_OFFSET_PERC_ADJUST
	};

	enum AWBMethod
	{
		AWB_GRAYWORLD = 0,
		AWB_WENG
	};

	enum CFAMethod
	{
		CFA_SIMPLE_BAYER = 0
	};

	enum DefocusMethod
	{
		DEFOCUS_GAUSS = 0
	};

	struct PresetMetaData
	{
		Version version;
		Version compatibility;

		std::string presetFile;
		std::string desc;
	};

	struct Spatial
	{
		PresetMetaData presetMeta;

		cv::Size display_resolution;
		cv::Size camera_resolution;
		AspectRatioMode ar_mode;
		
		unsigned int extendBorderPx;
		unsigned int fadeBorderPx;
		
		std::string transform_file;
		bool backgroundInsertion;
	};

	struct Temporal
	{
		PresetMetaData presetMeta;

		double cam_FPS;
		double display_refresh_rate;
		double display_g2g_time;
		double display_pwm_duty_cycle;
		double display_pwm_frequency;
		double exposure_time_ms;
		double readout_time_ms;
		bool vsync;
		
		std::string blend_file;
	};

	struct AutomaticGainControl
	{
		PresetMetaData presetMeta;

		AGCMethod method;
		static const char* methodNames[2];

		unsigned int frameCountLower, frameCountUpper;
		float threshLower, threshUpper;
		float adjustLower, adjustUpper;
	};

	struct AutomaticWhiteBalance
	{
		PresetMetaData presetMeta;

		AWBMethod method;
		static const char* methodNames[2];
	};
	
	struct ColorFilterArray
	{
		PresetMetaData presetMeta;

		CFAMethod method;
		static const char* methodNames[2];

		bool aafilter;	
	};

	struct Defocus
	{
		PresetMetaData presetMeta;

		DefocusMethod method;
		static const char* methodNames[2];

		unsigned int kernel_size;
		double sigma;
	};

public:
	Scenario(Version camMarkVersion);
	~Scenario(void);

	// Adds a path to look for files referenced in the scenario
	void addFileSearchPath(boost::filesystem::path searchPath);

	// Read from a file
	void readXMLFromFile(std::string filename);

	// Read from an XML string
	//void readXMLFromString(std::string xmlString);
	
	// Read from a pugi XML document
	void readXML(pugi::xml_document &doc);
	
	// Print an overview to a std stream
	void printToStream(std::ostream &os);

public:
	Version camMarkVersion;

	std::string sourceFile;
	std::string outputFile;

	bool applySpatialDistortion;
	bool applyTemporalDistortion;
	bool applyAWBDistortion;
	bool applyAGCDistortion;
	bool applyCFADistortion;
	bool applyDefocus;

	double override_src_FPS;
	double cam_start_time;
	double cam_end_time;

	Spatial spatial;
	Temporal temporal;
	AutomaticGainControl agc;
	AutomaticWhiteBalance awb;
	ColorFilterArray cfa;
	Defocus defocus;

protected:
	inline bool hasDefinedValue(pugi::xml_attribute &attr)
	{	return attr && !std::string(attr.value()).empty();	}

	void readSpatial(pugi::xml_node &node);
	void readTemporal(pugi::xml_node &node);
	void readAGC(pugi::xml_node &node);
	void readAWB(pugi::xml_node &node);
	void readCFA(pugi::xml_node &node);
	void readDefocus(pugi::xml_node &node);

	void printUsedPresetInfo(PresetMetaData &presetMeta, std::ostream &os);

	// Check if a preset is specified for the specified node,
	// and if so, load and merge its contents into the node.
	// Returns valid preset meta data if a preset was found.
	PresetMetaData loadAndMergePreset(pugi::xml_node &node);

	// Find a file either directly or in one of the search paths
	std::string findFileOrThrow(boost::filesystem::path file);

protected:
	// A vector of file search paths
	std::vector<boost::filesystem::path> searchPaths;
};

#endif // __SCENARIO__