/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

/*------------------------------------------------------------------------------------------*\
   This file contains material supporting chapter 9 of the cookbook:  
   Computer Vision Programming using the OpenCV Library. 
   by Robert Laganiere, Packt Publishing, 2011.

   This program is free software; permission is hereby granted to use, copy, modify, 
   and distribute this source code, or portions thereof, for any purpose, without fee, 
   subject to the restriction that the copyright notice may not be removed 
   or altered from any source or altered source distribution. 
   The software is released on an as-is basis and without any warranties of any kind. 
   In particular, the software is not guaranteed to be fault-tolerant or free from failure. 
   The author disclaims all warranties with regard to this software, any use, 
   and any consequent failure, is purely the responsibility of the user.
 
   Copyright (C) 2010-2011 Robert Laganiere, www.laganiere.name
\*------------------------------------------------------------------------------------------*/

#include <vector>
#include <opencv2/opencv.hpp>

#include "features.h"

#ifdef INCLUDE_NONFREE
#include <opencv2/nonfree/features2d.hpp>
#endif

using namespace std;
using namespace cv;

// It seems that for matching OpenCV internally puts keypoints from multiple images into one vector,
// and thus only allows a maximum number of keypoints per image (otherwise, an assertion fails).
// The source code line in question is: "CV_Assert( trainDescCollection[iIdx].rows < IMGIDX_ONE );"
// Found in "cv::BFMatcher::knnMatchImpl", file "opencv\modules\features2d\src\matchers.cpp". 
// We thus limit the maximum number to the size checked in the assertion in OpenCV's source code.
const int IMGIDX_SHIFT = 18; const int IMGIDX_ONE = (1 << IMGIDX_SHIFT);
const int MAX_NUM_FEATURES = IMGIDX_ONE - 1;

double find_features(const MatVector& ref_frames, const MatVector& copy_frames,
						FeatureType type, FrameFeaturesVector& frame_features,
						int keep_N_best, int gridRows, int gridCols)
{
	double ransacReprojThreshold_Normal = 3.0;
	double ransacReprojThreshold_High = 100.0;

	// Only for shorter names
	const auto &ref = ref_frames;
	const auto &cpy = copy_frames;
	const double RAN_norm = ransacReprojThreshold_Normal;
	const double RAN_high = ransacReprojThreshold_High;

	switch(type)
	{
		// Checkerboard calibration
		case CHECKERBOARD_31x20:return find_features_Checkerboard(ref, cpy, frame_features, Size(30, 19));
			
		// GoodFeaturesToTrack
		case GFTT_RAN:			return find_features_GFTT(ref, cpy, frame_features, true, RAN_norm, keep_N_best, gridRows, gridCols);

#ifdef INCLUDE_NONFREE
		// SURF + RANSAC
		case SURF_Dflt_RAN:		return find_features_SURF(   ref, cpy, frame_features, 0,   false, true, RAN_norm, keep_N_best, gridRows, gridCols);
		case SURF_10_RAN:		return find_features_SURF(   ref, cpy, frame_features, 10,  false, true, RAN_norm, keep_N_best, gridRows, gridCols);
		case SURF_400_RAN:		return find_features_SURF(   ref, cpy, frame_features, 400, false, true, RAN_norm, keep_N_best, gridRows, gridCols);
		case SURF_800_RAN:		return find_features_SURF(   ref, cpy, frame_features, 800, false, true, RAN_norm, keep_N_best, gridRows, gridCols);
		case SURF_AA250_RAN:	return find_features_SURF_AA(ref, cpy, frame_features, 250, false, true, RAN_norm, keep_N_best, gridRows, gridCols);
		case SURF_AA500_RAN:	return find_features_SURF_AA(ref, cpy, frame_features, 500, false, true, RAN_norm, keep_N_best, gridRows, gridCols);
		case SURF_AA750_RAN:	return find_features_SURF_AA(ref, cpy, frame_features, 750, false, true, RAN_norm, keep_N_best, gridRows, gridCols);

		// SURF + GoodMatches + RANSAC
		case SURF_Dflt_GM_RAN:	return find_features_SURF(   ref, cpy, frame_features, 0,   true, true, RAN_high, keep_N_best, gridRows, gridCols);
		case SURF_10_GM_RAN:	return find_features_SURF(   ref, cpy, frame_features, 10,  true, true, RAN_high, keep_N_best, gridRows, gridCols);
		case SURF_400_GM_RAN:	return find_features_SURF(   ref, cpy, frame_features, 400, true, true, RAN_high, keep_N_best, gridRows, gridCols);
		case SURF_800_GM_RAN:	return find_features_SURF(   ref, cpy, frame_features, 800, true, true, RAN_high, keep_N_best, gridRows, gridCols);
		case SURF_AA250_GM_RAN:	return find_features_SURF_AA(ref, cpy, frame_features, 250, true, true, RAN_high, keep_N_best, gridRows, gridCols);
		case SURF_AA500_GM_RAN:	return find_features_SURF_AA(ref, cpy, frame_features, 500, true, true, RAN_high, keep_N_best, gridRows, gridCols);
		case SURF_AA750_GM_RAN:	return find_features_SURF_AA(ref, cpy, frame_features, 750, true, true, RAN_high, keep_N_best, gridRows, gridCols);

		// SIFT + RANSAC
		case SIFT_Dflt_RAN:				return find_features_SIFT(ref, cpy, frame_features, 1.0f, false, true, RAN_norm, keep_N_best, gridRows, gridCols);
		case SIFT_Half_RAN:				return find_features_SIFT(ref, cpy, frame_features, 0.5f, false, true, RAN_norm, keep_N_best, gridRows, gridCols);
		case SIFT_Dbl_RAN:				return find_features_SIFT(ref, cpy, frame_features, 2.0f, false, true, RAN_norm, keep_N_best, gridRows, gridCols);

		// SURF + GoodMatches + RANSAC
		case SIFT_Dflt_GM_RAN:			return find_features_SIFT(ref, cpy, frame_features, 1.0f, true, true, RAN_high, keep_N_best, gridRows, gridCols);
		case SIFT_Dflt_BEST500_GM_RAN:	return find_features_SIFT(ref, cpy, frame_features, 1.0f, true, true, RAN_norm, 500, gridRows, gridCols);
		case SIFT_Half_GM_RAN:			return find_features_SIFT(ref, cpy, frame_features, 0.5f, true, true, RAN_high, keep_N_best, gridRows, gridCols);
		case SIFT_Dbl_GM_RAN:			return find_features_SIFT(ref, cpy, frame_features, 2.0f, true, true, RAN_high, keep_N_best, gridRows, gridCols);
#endif

		case ORB_GM_RAN:			return find_features_TYPE(ref, cpy, frame_features, "ORB",   "ORB",   true, true, RAN_high, keep_N_best, gridRows, gridCols);
		case BRISK_GM_RAN:			return find_features_TYPE(ref, cpy, frame_features, "BRISK", "BRISK", true, true, RAN_high, keep_N_best, gridRows, gridCols);
		
		case ORB_FREAK_GM_RAN:		return find_features_TYPE(ref, cpy, frame_features, "ORB",   "FREAK", true, true, RAN_high, keep_N_best, gridRows, gridCols);
		case BRISK_FREAK_GM_RAN:	return find_features_TYPE(ref, cpy, frame_features, "BRISK", "FREAK", true, true, RAN_high, keep_N_best, gridRows, gridCols);
		case MSER_FREAK_GM_RAN:		return find_features_TYPE(ref, cpy, frame_features, "MSER",  "FREAK", true, true, RAN_high, keep_N_best, gridRows, gridCols);
		case DENSE_FREAK_GM_RAN:	return find_features_TYPE(ref, cpy, frame_features, "DENSE", "FREAK", true, true, RAN_high, keep_N_best, gridRows, gridCols);

		default:					throw runtime_error("find_features:: Unsupported feature type.\n"
														"Hint: If you compiled the image registration library without non-free OpenCV features,\n"
														"then SIFT/SURF features will not be supported and lead to this error.");
	}
}

/************************************************
 * Utility functions (forward declarations)
 ************************************************/
void ratioTest(vector<vector<DMatch>>& matches,
			   float ratio = 0.65f);

void symmetryTest(const vector<vector<DMatch>>& matches1,
				  const vector<vector<DMatch>>& matches2,
				  vector<DMatch>& symMatches);

void noSymmetryTest(const vector<vector<DMatch>>& matches,
					vector<DMatch>& outMatches);

void ransacTest(const vector<DMatch>& matches,
				const vector<KeyPoint>& keypoints1, 
				const vector<KeyPoint>& keypoints2,
				vector<DMatch>& outMatches,
				double ransacReprojThreshold);

/************************************************
 * GoodFeaturesToTrack - Common base function
 ************************************************/
double find_features_GFTT_GENERIC(
	Ptr<FeatureDetector> detector,
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features, bool filterRANSAC,
	double ransacReprojThreshold, int keep_N_best,
	int gridRows, int gridCols,
	Size winSize = Size(21,21), int maxLevel = 3)
{
	// Sanity check
	CV_Assert(ref_frames.size() == copy_frames.size());
	CV_Assert(detector != NULL);

	// If no limit on the feature count is specified, or if the
	// specified limit is too hight, set it to the max possible.
	if(keep_N_best == 0 || keep_N_best > MAX_NUM_FEATURES)
		keep_N_best = MAX_NUM_FEATURES;

	// Use GradAdaptedFeatureDetector, even if grid os 1x1. It also retains
	// the N best features only, but evenly distributed in case of a grid.
	GridAdaptedFeatureDetector grid_detector(detector, keep_N_best, gridRows, gridCols);

	double error_sum = 0;
	unsigned int error_count = 0;

	// Iterate over all image pairs
	for(unsigned int i = 0; i < ref_frames.size(); i++)
	{
		KeyPointVector ref_kpts;
		grid_detector.detect(ref_frames[i], ref_kpts);

		// Skip images with no features
		if(ref_kpts.size() == 0)
		{
			// Insert an empty FrameFeatures object
			frame_features.push_back(FrameFeatures());
			continue;
		}
		
		Point2fVector ref_pts;
		KeyPoint::convert(ref_kpts, ref_pts);

		vector<uchar> status_string;
		vector<float> error_string;

		//vector<Mat> ref_frame_pyr, copy_frame_pyr;
		//buildOpticalFlowPyramid(ref_frames[i], ref_frame_pyr, winSize, maxLevel);
		//buildOpticalFlowPyramid(copy_frames[i], copy_frame_pyr, winSize, maxLevel);

		Point2fVector copy_pts;
		calcOpticalFlowPyrLK(ref_frames[i], copy_frames[i], ref_pts, copy_pts,
		//calcOpticalFlowPyrLK(ref_frame_pyr, copy_frame_pyr, ref_pts, copy_pts,
			status_string, error_string, winSize, maxLevel);

		KeyPointVector copy_kpts;
		KeyPoint::convert(copy_pts, copy_kpts);

		vector<DMatch> matches;
		for(unsigned int i = 0; i < ref_kpts.size(); i++)
		{
			if(status_string[i])
			{
				DMatch match;
				match.queryIdx = i;
				match.trainIdx = i;
				match.distance = error_string[i];
				matches.push_back(match);
			}
		}

		vector<DMatch> selected_matches;

		if(filterRANSAC)
		{
			// Validate matches using RANSAC
			ransacTest(matches, ref_kpts, copy_kpts,
				selected_matches, ransacReprojThreshold);
		}
		else
		{
			selected_matches.assign(matches.begin(), matches.end());
		}

		FrameFeatures features;
		for(unsigned int i = 0; i < selected_matches.size(); i++)
		{
			features.ref_kpts.push_back(ref_kpts[selected_matches[i].queryIdx]);
			features.copy_kpts.push_back(copy_kpts[selected_matches[i].trainIdx]);
		
			// Sum up the error
			error_sum += selected_matches[i].distance;
		}

		error_count += (unsigned int)selected_matches.size();

		frame_features.push_back(features);
	}

	if(error_count == 0) return -1;
	return error_sum / error_count;
}

/************************************************
 * GoodFeaturesToTrack - Default parameters
 ************************************************/
double find_features_GFTT(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features,
	bool filterRANSAC, double ransacReprojThreshold,
	int keep_N_best, int gridRows, int gridCols)
{
	// Sanity check
	CV_Assert(ref_frames.size() == copy_frames.size());

	Ptr<FeatureDetector> detector =
		new GoodFeaturesToTrackDetector();
	
	return find_features_GFTT_GENERIC(detector,
		ref_frames, copy_frames, frame_features,
		filterRANSAC, ransacReprojThreshold,
		keep_N_best, gridRows, gridCols);
}

/************************************************
 * SIFT/SURF - Common base function
 ************************************************/
double find_features_GENERIC(
	Ptr<FeatureDetector> detector, DescriptorExtractor* extractor,
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features,
	bool filterGoodMatches, bool filterRANSAC,
	double ransacReprojThreshold, int keep_N_best,
	int gridRows, int gridCols)
{
	// Sanity check
	CV_Assert(ref_frames.size() == copy_frames.size());
	CV_Assert(detector != NULL && extractor != NULL);
	
	// If no limit on the feature count is specified, or if the
	// specified limit is too hight, set it to the max possible.
	if(keep_N_best == 0 || keep_N_best > MAX_NUM_FEATURES)
		keep_N_best = MAX_NUM_FEATURES;

	// Use GradAdaptedFeatureDetector, even if grid os 1x1. It also retains
	// the N best features only, but evenly distributed in case of a grid.
	GridAdaptedFeatureDetector grid_detector(detector, keep_N_best, gridRows, gridCols);
	
	double error_sum = 0;
	unsigned int error_count = 0;

	for(unsigned int i = 0; i < ref_frames.size(); i++)
	{
		// Detect key points
		KeyPointVector ref_kpts, copy_kpts;
		grid_detector.detect(ref_frames[i], ref_kpts);
		grid_detector.detect(copy_frames[i], copy_kpts);

		// Skip images without the minimum required number of features
		if(ref_kpts.size() < 4 || copy_kpts.size() < 4)
		{
			// Insert an empty FrameFeatures object
			frame_features.push_back(FrameFeatures());
			continue;
		}
		
		// Extract feature descriptors
		Mat ref_descriptors, copy_descriptors;
		extractor->compute(ref_frames[i], ref_kpts, ref_descriptors);
		extractor->compute(copy_frames[i], copy_kpts, copy_descriptors);
		
		vector<DMatch> matches;

		if(filterGoodMatches)
		{
			//FlannBasedMatcher fMatcher;
			BFMatcher fMatcher;

			// from ref to copy
			// based on k nearest neighbours (with k=2)
			vector<vector<DMatch>> matches_ref;
			fMatcher.knnMatch(ref_descriptors, copy_descriptors, 
				matches_ref, // vector of matches (up to 2 per entry) 
				2);		     // return 2 nearest neighbours

			// from copy to ref
			// based on k nearest neighbours (with k=2)
			vector<vector<DMatch>> matches_copy;
			fMatcher.knnMatch(copy_descriptors, ref_descriptors, 
				matches_copy, // vector of matches (up to 2 per entry)
				2);			  // return 2 nearest neighbours
			
			// Remove matches for which NN ratio is > than threshold
			ratioTest(matches_ref);
			ratioTest(matches_copy);

			// Remove non-symmetrical matches
			symmetryTest(matches_ref, matches_copy, matches);
		}
		else
		{
			//FlannBasedMatcher fMatcher;
			BFMatcher fMatcher;

			// Get most probable matches
			fMatcher.match(ref_descriptors, copy_descriptors, matches);
		}

		vector<DMatch> selected_matches;

		if(filterRANSAC)
		{
			// Validate matches using RANSAC
			ransacTest(matches, ref_kpts, copy_kpts,
				selected_matches, ransacReprojThreshold);
		}
		else
		{
			selected_matches.assign(matches.begin(), matches.end());
		}

		FrameFeatures features;
		for(unsigned int i = 0; i < selected_matches.size(); i++)
		{
			features.ref_kpts.push_back(ref_kpts[selected_matches[i].queryIdx]);
			features.copy_kpts.push_back(copy_kpts[selected_matches[i].trainIdx]);

			// Sum up the error
			error_sum += selected_matches[i].distance;
		}

		error_count += (unsigned int)selected_matches.size();

		frame_features.push_back(features);
	}

	if(error_count == 0) return -1;
	return error_sum / error_count;
}

// Clear matches for which NN ratio is > than threshold
// return the number of removed points 
// (corresponding entries being cleared, i.e. size will be 0)
void ratioTest(vector<vector<DMatch>>& matches, float ratio)
{
	// for all matches
	for (vector<vector<DMatch>>::iterator matchIterator= matches.begin();
			matchIterator!= matches.end(); ++matchIterator)
	{
				// if 2 NN has been identified
				if (matchIterator->size() > 1)
				{
					// check distance ratio
					if ((*matchIterator)[0].distance/(*matchIterator)[1].distance > ratio) {

						matchIterator->clear(); // remove match
					}

				} else
				{ // does not have 2 neighbours

					matchIterator->clear(); // remove match
				}
	}
}

// Insert symmetrical matches in symMatches vector
void symmetryTest(const vector<vector<DMatch>>& matches1,
				const vector<vector<DMatch>>& matches2,
				vector<DMatch>& symMatches)
{
	// for all matches image 1 -> image 2
	for (vector<vector<DMatch>>::const_iterator matchIterator1= matches1.begin();
			matchIterator1!= matches1.end(); ++matchIterator1) {

		if (matchIterator1->size() < 2) // ignore deleted matches 
			continue;

		// for all matches image 2 -> image 1
		for (vector<vector<DMatch>>::const_iterator matchIterator2= matches2.begin();
			matchIterator2!= matches2.end(); ++matchIterator2) {

			if (matchIterator2->size() < 2) // ignore deleted matches 
				continue;

			// Match symmetry test
			if ((*matchIterator1)[0].queryIdx == (*matchIterator2)[0].trainIdx  && 
				(*matchIterator2)[0].queryIdx == (*matchIterator1)[0].trainIdx) {

					// add symmetrical match
					symMatches.push_back(DMatch((*matchIterator1)[0].queryIdx,
													(*matchIterator1)[0].trainIdx,
													(*matchIterator1)[0].distance));
					break; // next match in image 1 -> image 2
			}
		}
	}
}

// Identify good matches using RANSAC
// Return fundemental matrix
void ransacTest(const vector<DMatch>& matches,
				const vector<KeyPoint>& keypoints1, 
				const vector<KeyPoint>& keypoints2,
				vector<DMatch>& outMatches,
				double ransacReprojThreshold)
{
	// Early out
	if(matches.size() == 0) return;

	// Convert keypoints into Point2f
	vector<Point2f> points1, points2;	
	for (vector<DMatch>::const_iterator it= matches.begin();
			it!= matches.end(); ++it) {

			// Get the position of left keypoints
			float x= keypoints1[it->queryIdx].pt.x;
			float y= keypoints1[it->queryIdx].pt.y;
			points1.push_back(Point2f(x,y));
			// Get the position of right keypoints
			x= keypoints2[it->trainIdx].pt.x;
			y= keypoints2[it->trainIdx].pt.y;
			points2.push_back(Point2f(x,y));
	}

	// At least 4 points are needed
	if(points1.size() < 4) return;

	// Comupte the homography using RANSAC
	vector<uchar> inliers(points1.size(),0);
	findHomography(Mat(points1), Mat(points2),
		CV_RANSAC, ransacReprojThreshold, inliers);

	// Extract the surviving (inliers) matches
	vector<uchar>::const_iterator itIn= inliers.begin();
	vector<DMatch>::const_iterator itM= matches.begin();
	// for all matches
	for ( ;itIn!= inliers.end(); ++itIn, ++itM)
	{
		if (*itIn)
		{ // it is a valid match
			outMatches.push_back(*itM);
		}
	}
}

#ifdef INCLUDE_NONFREE

/************************************************
 * SURF with default or specified parameters
 ************************************************/
double find_features_SURF(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features, double hessianThreshold,
	bool filterGoodMatches, bool filterRANSAC,
	double ransacReprojThreshold, int keep_N_best,
	int gridRows, int gridCols)
{
	Ptr<FeatureDetector> detector;
	if(hessianThreshold > 0)
		 detector = new SurfFeatureDetector(hessianThreshold);
	else detector = new SurfFeatureDetector;
	
	SurfDescriptorExtractor extractor;
	
	double error_estimate = find_features_GENERIC(
		detector, &extractor,
		ref_frames, copy_frames, frame_features,
		filterGoodMatches, filterRANSAC,
		ransacReprojThreshold, keep_N_best,
		gridRows, gridCols);

	return error_estimate;
}

/************************************************
 * SURF with automatic parameters calculated
 * based on a target number of features
 ************************************************/
double find_features_SURF_AA(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features, int num_features,
	bool filterGoodMatches, bool filterRANSAC,
	double ransacReprojThreshold, int keep_N_best,
	int gridRows, int gridCols)
{
	Ptr<FeatureDetector> detector(
		new DynamicAdaptedFeatureDetector(
			new SurfAdjuster, num_features - 50,
							  num_features * 50, 10));
	
	SurfDescriptorExtractor extractor;
	
	double error_estimate = find_features_GENERIC(
		detector, &extractor,
		ref_frames, copy_frames, frame_features,
		filterGoodMatches, filterRANSAC,
		ransacReprojThreshold, keep_N_best,
		gridRows, gridCols);

	return error_estimate;
}

/************************************************
 * SIFT with default or adjusted parameters
 ************************************************/
double find_features_SIFT(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features, float factor,
	bool filterGoodMatches, bool filterRANSAC,
	double ransacReprojThreshold, int keep_N_best,
	int gridRows, int gridCols)
{
	Ptr<FeatureDetector> detector;
	if(factor != 1)
		 detector = new SiftFeatureDetector(0, 3, 0.04f / factor, 10.0f * factor);
	else detector = new SiftFeatureDetector;

	SiftDescriptorExtractor extractor;
	
	double error_estimate = find_features_GENERIC(
		detector, &extractor,
		ref_frames, copy_frames, frame_features,
		filterGoodMatches, filterRANSAC,
		ransacReprojThreshold, keep_N_best,
		gridRows, gridCols);

	return error_estimate;
}

#endif

/****************************************************
 * Checkerboard with 'patternsize' internal corners *
 ****************************************************/
double find_features_Checkerboard(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features, Size patternsize)
{
	// Sanity check
	CV_Assert(ref_frames.size() == copy_frames.size());
	
	for(unsigned int i = 0; i < ref_frames.size(); i++)
	{
		Mat ref_gray, copy_gray;
		cvtColor(ref_frames[i], ref_gray, CV_BGR2GRAY);
		cvtColor(copy_frames[i], copy_gray, CV_BGR2GRAY);

		vector<Point2f> ref_corners, copy_corners;

		// Detect chessboard internal corners
		bool ref_pattern_found = findChessboardCorners(ref_gray, patternsize, ref_corners);
		bool copy_pattern_found = findChessboardCorners(copy_gray, patternsize, copy_corners);

		// Skip images where no pattern was found
		if(!ref_pattern_found || !copy_pattern_found)
		{
			// Insert an empty FrameFeatures object
			frame_features.push_back(FrameFeatures());
			continue;
		}

		cornerSubPix(ref_gray, ref_corners, Size(11, 11), Size(-1, -1), TermCriteria(CV_TERMCRIT_EPS + CV_TERMCRIT_ITER, 30, 0.1));
		cornerSubPix(copy_gray, copy_corners, Size(11, 11), Size(-1, -1), TermCriteria(CV_TERMCRIT_EPS + CV_TERMCRIT_ITER, 30, 0.1));

		// Visualization
		/*Mat ref_vis, copy_vis;
		cvtColor(ref_gray, ref_vis, CV_GRAY2BGR);
		cvtColor(copy_gray, copy_vis, CV_GRAY2BGR);
		drawChessboardCorners(ref_vis, patternsize, Mat(ref_corners), ref_pattern_found);
		drawChessboardCorners(copy_vis, patternsize, Mat(copy_corners), copy_pattern_found);
		imshow("Chessboard on reference", ref_vis);
		imshow("Chessboard on copy", copy_vis);
		waitKey(100);*/

		FrameFeatures features;
		KeyPoint::convert(ref_corners, features.ref_kpts);
		KeyPoint::convert(copy_corners, features.copy_kpts);
		
		frame_features.push_back(features);
	}

	return 0;
}

/************************************************
 * FeatureDetector and DescriptorExtractor generated
 * by type string, with default parameters.
 *
 * FeatureDetectors: 
 *   "FAST" – FastFeatureDetector
 *   "STAR" – StarFeatureDetector
 *   "SIFT" – SIFT (nonfree module)
 *   "SURF" – SURF (nonfree module)
 *   "ORB" – ORB
 *   "BRISK" – BRISK
 *   "MSER" – MSER
 *   "GFTT" – GoodFeaturesToTrackDetector
 *   "HARRIS" – GoodFeaturesToTrackDetector with Harris detector enabled
 *   "Dense" – DenseFeatureDetector
 *   "SimpleBlob" – SimpleBlobDetector
 *
 * DescriptorExtractors: 
 *   "SIFT" – SIFT
 *   "SURF" – SURF
 *   "BRIEF" – BriefDescriptorExtractor
 *   "BRISK" – BRISK
 *   "ORB" – ORB
 *   "FREAK" – FREAK
 ************************************************/
double find_features_TYPE(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features,
	const string featureDetectorType,
	const string descriptorExtractorType,
	bool filterGoodMatches, bool filterRANSAC,
	double ransacReprojThreshold, int keep_N_best,
	int gridRows, int gridCols)
{
	Ptr<FeatureDetector> featureDetector = FeatureDetector::create(featureDetectorType);
	Ptr<DescriptorExtractor> desciptorExtractor = DescriptorExtractor::create(descriptorExtractorType);

	if( NULL == featureDetector ||
		NULL == desciptorExtractor ) return -1;
	
	double error_estimate = find_features_GENERIC(
		featureDetector, desciptorExtractor,
		ref_frames, copy_frames, frame_features,
		filterGoodMatches, filterRANSAC,
		ransacReprojThreshold, keep_N_best,
		gridRows, gridCols);

	return error_estimate;
}
