/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#ifndef IMREG_FEATURES_H
#define IMREG_FEATURES_H

// ------------------------------------------------------------
// Define to also include SIFT/SURF from the nonfree module
#define INCLUDE_NONFREE
// ------------------------------------------------------------

#include <vector>

/******************************************************************
 * Utility typedefs
 ******************************************************************/
typedef std::vector<cv::Mat> MatVector;
typedef std::vector<cv::KeyPoint> KeyPointVector;
typedef std::vector<cv::Point2f> Point2fVector;
typedef std::vector<cv::Point3f> Point3fVector;

/******************************************************************
 * All possible feature type algorithms/parameters
 ******************************************************************/
enum FeatureType
{
	NoRegistration = 0,

	// Checkerboard calibration
	CHECKERBOARD_31x20,

	// GoodFeaturesToTrack
	GFTT_RAN,

	// SURF + RANSAC
	SURF_Dflt_RAN,
	SURF_10_RAN, SURF_400_RAN, SURF_800_RAN,
	SURF_AA250_RAN, SURF_AA500_RAN, SURF_AA750_RAN,

	// SURF + GoodMatches + RANSAC
	SURF_Dflt_GM_RAN,
	SURF_10_GM_RAN, SURF_400_GM_RAN, SURF_800_GM_RAN,
	SURF_AA250_GM_RAN, SURF_AA500_GM_RAN, SURF_AA750_GM_RAN,

	// SIFT + RANSAC
	SIFT_Dflt_RAN,
	SIFT_Half_RAN,
	SIFT_Dbl_RAN,

	// SIFT + GoodMatches + RANSAC
	SIFT_Dflt_GM_RAN,
	SIFT_Dflt_BEST500_GM_RAN,
	SIFT_Half_GM_RAN,
	SIFT_Dbl_GM_RAN,

	// ORB + GoodMatches + RANSAC
	ORB_GM_RAN,

	// BRISK + GoodMatches + RANSAC
	BRISK_GM_RAN,

	// (ORB|BRISK|MSER|Dense)+FREAK + GoodMatches + RANSAC
	ORB_FREAK_GM_RAN,
	BRISK_FREAK_GM_RAN,
	MSER_FREAK_GM_RAN,
	DENSE_FREAK_GM_RAN
};

/******************************************************************
 * Structure storing feature points for reference and copy
 ******************************************************************/
struct FrameFeatures
{
	// Features in reference image
	KeyPointVector ref_kpts;

	// Features in corresponding copy image
	KeyPointVector copy_kpts;
};

typedef std::vector<FrameFeatures> FrameFeaturesVector;

// Generic wrapper
double find_features(const MatVector& ref_frames, const MatVector& copy_frames,
					 FeatureType type, FrameFeaturesVector& frame_features,
					 int keep_N_best = 0, int gridRows = 1, int gridCols = 1);

double find_features_GFTT(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features, bool filterRANSAC,
	double ransacReprojThreshold = 3.0, int keep_N_best = 0,
	int gridRows = 1, int gridCols = 1);

#ifdef INCLUDE_NONFREE

double find_features_SURF(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features, double hessianThreshold,
	bool filterGoodMatches, bool filterRANSAC,
	double ransacReprojThreshold = 3.0, int keep_N_best = 0,
	int gridRows = 1, int gridCols = 1);

double find_features_SURF_AA(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features, int num_features,
	bool filterGoodMatches, bool filterRANSAC,
	double ransacReprojThreshold = 3.0, int keep_N_best = 0,
	int gridRows = 1, int gridCols = 1);

double find_features_SIFT(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features, float factor,
	bool filterGoodMatches, bool filterRANSAC,
	double ransacReprojThreshold = 3.0, int keep_N_best = 0,
	int gridRows = 1, int gridCols = 1);

#endif

double find_features_TYPE(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features,
	const std::string featureDetectorType,
	const std::string descriptorExtractorType,
	bool filterGoodMatches, bool filterRANSAC,
	double ransacReprojThreshold = 3.0, int keep_N_best = 0,
	int gridRows = 1, int gridCols = 1);

double find_features_Checkerboard(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FrameFeaturesVector& frame_features, cv::Size patternsize);

#endif // IMREG_FEATURES_H
