/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <opencv2/opencv.hpp>

#ifdef INCLUDE_NONFREE
#include <opencv2/nonfree/features2d.hpp>
#endif


#include <iostream>
#include <list>
#include <vector>
#include <time.h>

#include "params.h"

using namespace std;
using namespace cv;

ImRegParams::ImRegParams(Size& sourceImgSize, Size& targetImgSize)
	: originalSourceImgSize(sourceImgSize),
	  originalTargetImgSize(targetImgSize)
{
	// An ImRegParam object created with this constructor is still invalid
	// (since only the sizes are set), but if stored, at least the sizes are written
	// (they are used by some applications even if the registration is invalid).
}

ImRegParams::ImRegParams(Mat& homography, Mat& distCoeffs,
	Mat& cameraMatrix, Size& sourceImgSize, Size& targetImgSize)
	: originalSourceImgSize(sourceImgSize),
	  originalTargetImgSize(targetImgSize)
{
	homography.copyTo(this->homography);
	distCoeffs.copyTo(this->distCoeffs);
	cameraMatrix.copyTo(this->cameraMatrix);

	// Pre-compute the rectification matrix
	rectification = computeRectificationMatrix(cameraMatrix, homography);
}

ImRegParams::ImRegParams(const ImRegParams &other)
	: originalSourceImgSize(other.originalSourceImgSize),
	  originalTargetImgSize(other.originalTargetImgSize)
{
	// Copy all matrices
	other.homography.copyTo(homography);
	other.distCoeffs.copyTo(distCoeffs);
	other.cameraMatrix.copyTo(cameraMatrix);
	other.rectification.copyTo(rectification);
}

ImRegParams::ImRegParams(ImRegParams &&other)
{
	swap(*this, other);
}

void swap(ImRegParams& first, ImRegParams& second)
{
	// Enable ADL (not necessary in our case, but good practice)
	using std::swap; 

	// By swapping the members of two classes,
	// the two classes are effectively swapped
	swap(first.homography, second.homography); 
	swap(first.distCoeffs, second.distCoeffs);
	swap(first.cameraMatrix, second.cameraMatrix);
	swap(first.rectification, second.rectification);
	swap(first.originalSourceImgSize, second.originalSourceImgSize);
	swap(first.originalTargetImgSize, second.originalTargetImgSize);
}

ImRegParams& ImRegParams::operator=(ImRegParams other)
{
	swap(*this, other);
	return *this;
}

void ImRegParams::initializeNeutral(Size imgSize)
{
	// Initialize with neutral defaults
	homography = Mat::eye(3,3, CV_64F);
	distCoeffs = Mat::zeros(5,1, CV_64F);
	cameraMatrix = Mat::eye(3,3, CV_64F);

	// Set the camera matrix' focal length
	cameraMatrix.at<double>(0,0) = imgSize.width;
	cameraMatrix.at<double>(1,1) = imgSize.width;

	// Set the camera matrix' principal point
	cameraMatrix.at<double>(0,2) = imgSize.width / 2.0;
	cameraMatrix.at<double>(1,2) = imgSize.height / 2.0;

	// Pre-compute the rectification matrix
	rectification = computeRectificationMatrix(cameraMatrix, homography);

	// Initialize size parameters
	originalSourceImgSize = imgSize;
	originalTargetImgSize = imgSize;
}

bool ImRegParams::valid() const
{
	// Check all members
	return	!homography.empty() &&
			!distCoeffs.empty() &&
			!cameraMatrix.empty() &&
			!rectification.empty() &&
			originalSourceImgSize.area() > 0 &&
			originalTargetImgSize.area() > 0;
}

ImRegUndistortRectifyMaps ImRegParams::calcUndistortRectifyMaps() const
{
	// Shortcut for call with original source/target image sizes
	return calcUndistortRectifyMaps(originalSourceImgSize, originalTargetImgSize);
}

ImRegUndistortRectifyMaps ImRegParams::calcUndistortRectifyMaps(
	Size sourceImgSize, Point srcImgOrigin) const
{
	// Shortcut for call with original target image size
	return calcUndistortRectifyMaps(sourceImgSize, originalTargetImgSize, srcImgOrigin);
}

ImRegUndistortRectifyMaps ImRegParams::calcUndistortRectifyMaps(
	Size sourceImgSize, Size targetImgSize, Point srcImgOrigin) const
{
	ImRegUndistortRectifyMaps maps;

	// Sanity check
	if(!valid()) return maps;

	// Get source/target camera matrices, adjusted to the specified sizes and/or origin (if needed)
	const Mat sourceCameraMatrix = getSourceCameraMatrix(sourceImgSize, srcImgOrigin);
	const Mat targetCameraMatrix = getTargetCameraMatrix(targetImgSize);
	
	// Use OpenCV to calculate the combined undistort/recitfy maps
	initUndistortRectifyMap(sourceCameraMatrix, distCoeffs,
		rectification, targetCameraMatrix, targetImgSize, CV_16SC2,
		maps.undistort_rectify_map1, maps.undistort_rectify_map2);

	return maps;
}

const Mat ImRegParams::getSourceCameraMatrix(Size sourceImgSize, Point srcImgOrigin) const
{
	if( sourceImgSize == originalSourceImgSize &&
		srcImgOrigin == Point(0, 0))
		return cameraMatrix;

	return createAdjustedCameraMatrix(cameraMatrix,
		originalSourceImgSize, sourceImgSize, srcImgOrigin);
}

const Mat ImRegParams::getTargetCameraMatrix(Size targetImgSize) const
{
	if(targetImgSize == originalTargetImgSize)
		return cameraMatrix;

	return createAdjustedCameraMatrix(cameraMatrix,
		originalTargetImgSize, targetImgSize);
}

void ImRegParams::write(FileStorage& fs) const
{
	fs << "{";
	fs << "homography" << homography;
	fs << "distCoeffs" << distCoeffs;
	fs << "cameraMatrix" << cameraMatrix;
	fs << "originalSourceImgSize" << originalSourceImgSize;
	fs << "originalTargetImgSize" << originalTargetImgSize;
	fs << "}";
}

void ImRegParams::read(const FileNode& node)
{
	node["homography"] >> homography;
	node["distCoeffs"] >> distCoeffs;
	node["cameraMatrix"] >> cameraMatrix;
	node["originalSourceImgSize"] >> originalSourceImgSize;
	node["originalTargetImgSize"] >> originalTargetImgSize;

	// Pre-compute the rectification matrix
	rectification = computeRectificationMatrix(cameraMatrix, homography);
}

const Mat ImRegParams::computeRectificationMatrix(
	const Mat& cameraMatrix, const Mat& homography)
{
	// Sanity check
	if(cameraMatrix.empty() || homography.empty())
		return Mat();

	// Rectification matrix R = C^{-1} * H * C
	return cameraMatrix.inv() * homography * cameraMatrix;
}

const Mat ImRegParams::createAdjustedCameraMatrix(
	const Mat& cameraMatrix, Size imgSize, Size newSize, Point imgOrigin)
{
	CV_Assert(	// Sanity check
		cameraMatrix.size().width == 3 ||
		cameraMatrix.size().height == 3 ||
		cameraMatrix.type() == CV_64F);

	Mat adjCameraMatrix;
	cameraMatrix.copyTo(adjCameraMatrix);

	// Calculate a scaling factor from the previous image size
	Size2d scalingFactors(			  // to the new image size
		(double)newSize.width / imgSize.width,
		(double)newSize.height / imgSize.height);

	// Adjust the camera matrix' focal length
	adjCameraMatrix.at<double>(0,0) *= scalingFactors.width;
	adjCameraMatrix.at<double>(1,1) *= scalingFactors.height;

	// Adjust the camera matrix' principal point
	adjCameraMatrix.at<double>(0,2) *= scalingFactors.width;
	adjCameraMatrix.at<double>(1,2) *= scalingFactors.height;
	adjCameraMatrix.at<double>(0,2) += imgOrigin.x;
	adjCameraMatrix.at<double>(1,2) += imgOrigin.y;

	return adjCameraMatrix;
}

void write(FileStorage& fs, const string&, const ImRegParams& x)
{
	x.write(fs);
}

void read(const FileNode& node, ImRegParams& x, const ImRegParams& default_value)
{
	if(node.empty())
		 x = default_value;
	else x.read(node);
}
