/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#ifndef IMREG_PARAMS_H
#define IMREG_PARAMS_H

struct ImRegUndistortRectifyMaps
{
	// Combined preprocessed undistortion map
	// (used when transformation is applied)
	cv::Mat undistort_rectify_map1;
	cv::Mat undistort_rectify_map2;
};

class ImRegParams
{
public:
	// Regular constructors
	ImRegParams() { /*empty*/ }
	ImRegParams(cv::Size& sourceImgSize, cv::Size& targetImgSize);
	ImRegParams(cv::Mat& homography, cv::Mat& distCoeffs,
		cv::Mat& cameraMatrix, cv::Size& sourceImgSize, cv::Size& targetImgSize);

	// Copy/move constructors
	ImRegParams(const ImRegParams &other);
	ImRegParams(ImRegParams &&other);

	// Implement copy-and-swap idiom
	friend void swap(ImRegParams& first, ImRegParams& second);

	// Copy/move assignment (other *must* be
	// passed by value for the copy-and-swap idiom)
	ImRegParams& operator=(ImRegParams other);
	
public:
	// Initialize to effectifely not introduce any distortion when applied.
	void initializeNeutral(cv::Size imgSize);

	// Check if this represents valid registration data.
	// If registration methods fail, they will return an empty
	// an dthus invalif object.
	bool valid() const;

	// Calculate the combined distortion & rectification lookup maps,
	// optionally for a new specified source/target image size and/or source image origin.
	ImRegUndistortRectifyMaps calcUndistortRectifyMaps() const;
	ImRegUndistortRectifyMaps calcUndistortRectifyMaps(
		cv::Size sourceImgSize, cv::Point srcImgOrigin = cv::Point(0, 0)) const;
	ImRegUndistortRectifyMaps calcUndistortRectifyMaps(
		cv::Size sourceImgSize, cv::Size targetImgSize, cv::Point srcImgOrigin = cv::Point(0, 0)) const;

public:
	// Inline access to matrices
	inline const cv::Mat getHomography() const			{	return homography;		}
	inline const cv::Mat getDistCoeffs() const			{	return distCoeffs;		}
	inline const cv::Mat getRectification() const		{	return rectification;	}
	inline const cv::Mat getCameraMatrix() const		{	return cameraMatrix;	}
	
	// Inline access to original source/target image size
	inline cv::Size getOriginalSourceImgSize() const	{	return originalSourceImgSize;	}
	inline cv::Size getOriginalTargetImgSize() const	{	return originalTargetImgSize;	}

	// Get source/target camera matrix that is adjusted to
	// new source/target image sizes and/or source image origin, if needed
	const cv::Mat getSourceCameraMatrix(cv::Size sourceImgSize, cv::Point srcImgOrigin = cv::Point(0, 0)) const;
	const cv::Mat getTargetCameraMatrix(cv::Size targetImgSize) const;

	// Serialization
	void write(cv::FileStorage& fs) const;
	void read(const cv::FileNode& node);

protected:
	// Comutes a rectification matrix from camera and homography matrix
	static const cv::Mat computeRectificationMatrix(
		const cv::Mat& cameraMatrix,
		const cv::Mat& homography);

	// Adjusts a camera matrix to a new image size
	static const cv::Mat createAdjustedCameraMatrix(
		const cv::Mat& cameraMatrix,
		cv::Size imgSize, cv::Size newSize,
		cv::Point imgOrigin = cv::Point(0, 0));

protected:
	// Perspective transformation
	cv::Mat homography;

	// Lens distortion coefficients
	cv::Mat distCoeffs;

	// Intrinsic camera parameters
	cv::Mat cameraMatrix;
	
	// Original source/target image size.
	// The above cameraMatrix is dependent on these sizes.
	cv::Size originalSourceImgSize;
	cv::Size originalTargetImgSize;

	// Pre-calculated rectification matrix
	cv::Mat rectification;
};

//These write and read functions must be defined for the serialization in FileStorage to work
void write(cv::FileStorage& fs, const std::string&, const ImRegParams& x);
void read(const cv::FileNode& node, ImRegParams& x,
	const ImRegParams& default_value = ImRegParams());

#endif // IMREG_PARAMS_H
