/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <opencv2/opencv.hpp>
#include <vector>

#include "registration.h"
#include "remapping.h"

using namespace std;
using namespace cv;

bool register_homography(
	const FrameFeaturesVector& frame_features,
	Mat& homography)
{
	// Sanity check
	if(frame_features.size() == 0)
		return false;

	Point2fVector all_ref_pts, all_copy_pts;
	for(unsigned int i = 0; i < frame_features.size(); i++)
	{
		// Sanity check
		CV_Assert(frame_features[i].ref_kpts.size() ==
				  frame_features[i].copy_kpts.size());

		Point2fVector ref_pts, copy_pts;
		KeyPoint::convert(frame_features[i].ref_kpts, ref_pts);
		KeyPoint::convert(frame_features[i].copy_kpts, copy_pts);

		all_ref_pts.insert(all_ref_pts.end(), ref_pts.begin(), ref_pts.end());
		all_copy_pts.insert(all_copy_pts.end(), copy_pts.begin(), copy_pts.end());
	}

	if(all_ref_pts.size() < 4)
		return false;

	homography = findHomography(
		all_ref_pts, all_copy_pts, 0);
		
	return true;
}

double computeReprojectionErrors( const vector<vector<Point3f> >& objectPoints,
								  const vector<vector<Point2f> >& imagePoints,
								  const vector<Mat>& rvecs, const vector<Mat>& tvecs,
								  const Mat& cameraMatrix , const Mat& distCoeffs,
								  vector<float>& perViewErrors)
{
	vector<Point2f> imagePoints2;
	int i, totalPoints = 0;
	double totalErr = 0, err;
	perViewErrors.resize(objectPoints.size());

	for( i = 0; i < (int)objectPoints.size(); ++i )
	{
		projectPoints( Mat(objectPoints[i]), rvecs[i], tvecs[i], cameraMatrix, 
					   distCoeffs, imagePoints2);
		err = norm(Mat(imagePoints[i]), Mat(imagePoints2), CV_L2);
		
		int n = (int)objectPoints[i].size();
		perViewErrors[i] = (float) std::sqrt(err*err/n);
		totalErr        += err*err;
		totalPoints     += n;
	}

	return std::sqrt(totalErr/totalPoints);
}

void getPointSets(const FrameFeaturesVector& frame_features,
	bool mergePointSets,
	vector<Point3fVector>& objectPoints3f,
	vector<Point2fVector>& imagePoints2f)
{
	if(mergePointSets)
	{
		objectPoints3f.resize(1);
		imagePoints2f.resize(1);
	}
	else
	{
		objectPoints3f.clear();
		imagePoints2f.clear();
	}

	for(unsigned int i = 0; i < frame_features.size(); i++)
	{
		CV_Assert(frame_features[i].ref_kpts.size() ==
				  frame_features[i].copy_kpts.size());

		if(frame_features[i].ref_kpts.size() < 8)
			return;

		Point2fVector ref_pts, copy_pts;
		KeyPoint::convert(frame_features[i].ref_kpts, ref_pts);
		KeyPoint::convert(frame_features[i].copy_kpts, copy_pts);

		if(ref_pts.size() >= 8)
		{
			Point3fVector objectPointBuf3f;
			Point2fVector imagePointBuf2f;

			for(unsigned int i = 0; i < ref_pts.size(); i ++)
			{
				objectPointBuf3f.push_back(Point3f(copy_pts[i]));
				imagePointBuf2f.push_back(Point2f(ref_pts[i]));
			}

			if(mergePointSets)
			{
				objectPoints3f[0].insert(objectPoints3f[0].end(), objectPointBuf3f.begin(), objectPointBuf3f.end());
				imagePoints2f[0].insert(imagePoints2f[0].end(), imagePointBuf2f.begin(), imagePointBuf2f.end());
			}
			else
			{
				objectPoints3f.push_back(objectPointBuf3f);
				imagePoints2f.push_back(imagePointBuf2f);
			}
		}
	}
}

bool register_calibrate_camera(
	const FrameFeaturesVector& frame_features, Mat& cameraMatrix, Mat& distCoeffs,
	Size img_size, bool mergePointSets, double& error_estimation, int flags)
{
	// Sanity check
	if(frame_features.size() == 0)
		return false;

	vector<Point3fVector> objectPoints3f;
	vector<Point2fVector> objectPoints2f;
	vector<Point2fVector> imagePoints2f;

	getPointSets(frame_features, mergePointSets,
		objectPoints3f, imagePoints2f);
	
	if( objectPoints3f.size() == 0 ||
		imagePoints2f.size() == 0)
		return false;

	if(flags == 0) // Default flags
	{
		flags |= CV_CALIB_FIX_PRINCIPAL_POINT;
		flags |= CV_CALIB_ZERO_TANGENT_DIST;
		//flags |= CV_CALIB_FIX_ASPECT_RATIO;
		flags |= CV_CALIB_FIX_K2;
		flags |= CV_CALIB_FIX_K3;
		flags |= CV_CALIB_FIX_K4;
		flags |= CV_CALIB_FIX_K5;
		flags |= CV_CALIB_FIX_K6;
	}

	cameraMatrix = Mat::eye(3, 3, CV_64F);
	if( flags & CV_CALIB_FIX_ASPECT_RATIO )
		cameraMatrix.at<double>(0,0) = 1.0;

	distCoeffs = Mat::zeros(8, 1, CV_64F);

	//Find intrinsic and extrinsic camera parameters
	vector<Mat> rvecs, tvecs;
	/*double rms = */calibrateCamera(objectPoints3f, imagePoints2f, img_size, cameraMatrix,
									 distCoeffs, rvecs, tvecs, flags);

	bool ok = checkRange(cameraMatrix) && checkRange(distCoeffs);
	if(!ok) return false;

	vector<float> reprojErrs;
	error_estimation = computeReprojectionErrors(objectPoints3f, imagePoints2f,
							rvecs, tvecs, cameraMatrix, distCoeffs, reprojErrs);
	
	return true;
}

ImRegParams register_frames(
	const MatVector& ref_frames, const MatVector& copy_frames,
	FeatureType lens_feature_type, FeatureType perspective_feature_type,
	int flags, int keep_N_best, int gridRows, int gridCols,
	ImRegErrEstimate *errEstimate, FrameFeaturesVector *found_features)
{
	// Sanity check: Same number of reference and copy frames
	CV_Assert(ref_frames.size() == copy_frames.size());
	
	MatVector reg_ref_frames;

	Size ref_size = ref_frames[0].size();
	Size copy_size = copy_frames[0].size();
	
	if(errEstimate)
	{
		// Initialize error estimates to -1 (not set)
		errEstimate->error_est_lens = -1;
		errEstimate->error_est_features = -1;

		// Initialize the number of features found
		errEstimate->num_features_lens = 0;
		errEstimate->num_features_perspective = 0;
	}

	// Matrices to be determined
	Mat homography, cameraMatrix, distCoeffs;

	// 1) LENS DISTORTION
	if(lens_feature_type != NoRegistration)
	{
		// Find features for lens undistortion
		FrameFeaturesVector undist_features;
		find_features(ref_frames, copy_frames,
			lens_feature_type, undist_features,
			keep_N_best, gridRows, gridCols);

		// Count the number of features found
		if(errEstimate)
		{
			for(unsigned int i = 0; i < undist_features.size(); i++)
				errEstimate->num_features_lens += undist_features[i].ref_kpts.size();
		}

		// Registration of lens distortion
		double errEstLens;
		if(!register_calibrate_camera(undist_features, cameraMatrix, distCoeffs,
									  copy_size, false, errEstLens, flags))
			return ImRegParams(ref_size, copy_size);

		if(errEstimate) errEstimate->error_est_lens = errEstLens;

		// Apply the same distortion to the reference frames,
		// those will be then used for calculating the prespective registration
		ImRegUndistortRectifyMaps maps;
		
		//maps = calcUndistortRectifyMaps(params, ref_size, Size(0,0), copy_size);
		initUndistortRectifyMap(cameraMatrix, distCoeffs, Mat(),
			cameraMatrix, copy_size, CV_16SC2,
			maps.undistort_rectify_map1, maps.undistort_rectify_map2);
		
		reg_ref_frames.resize(ref_frames.size());
		for(unsigned int i = 0; i < ref_frames.size(); i++)
			remap_frame(ref_frames[i], reg_ref_frames[i], maps);
	}
	else
	{
		// Use the unmodified reference frames for calculating the
		// perspective registration in case we do no lens undistortion
		reg_ref_frames.assign(ref_frames.begin(), ref_frames.end());

		// Setup lens distortion values to do nothing
		cameraMatrix = Mat::eye(3, 3, CV_64F);
		distCoeffs = Mat::zeros(8, 1, CV_64F);
	}

	// 2) PERSPECTIVE DISTORTION
	if(perspective_feature_type != NoRegistration)
	{
		// Either used provided features vector,
		// or local temporal vector
		FrameFeaturesVector temp_features;
		FrameFeaturesVector	*reg_features_ptr = &temp_features;
		if(found_features)	 reg_features_ptr = found_features;
		FrameFeaturesVector &reg_features = *reg_features_ptr;

		// Find features for the perspective registration
		double errEstFeatures;
		errEstFeatures = find_features(reg_ref_frames, copy_frames,
			perspective_feature_type, reg_features,
			keep_N_best, gridRows, gridCols);
		
		if(errEstimate)
		{
			errEstimate->error_est_features = errEstFeatures;

			// Count the number of features found
			for(unsigned int i = 0; i < reg_features.size(); i++)
				errEstimate->num_features_perspective += reg_features[i].ref_kpts.size();
		}

		// Registration of perspective distortion
		if(!register_homography(reg_features, homography))
			return ImRegParams(ref_size, copy_size);
	}

	// Create and return an ImRegParams object (will use C++11 move semantics)
	return ImRegParams(homography, distCoeffs, cameraMatrix, ref_size, copy_size);
}
