/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <opencv2/opencv.hpp>

#ifdef INCLUDE_NONFREE
#include <opencv2/nonfree/features2d.hpp>
#endif

#include "remapping.h"
#include "params.h"

using namespace std;
using namespace cv;

void remap_frame(
	const Mat& ref_frame, Mat& out_frame,
	const ImRegParams& params,
	int interpolation, int borderMode)
{
	remap_frame(ref_frame, out_frame,
		params, params.getOriginalTargetImgSize(),
		interpolation, borderMode);
}

void remap_frame(
	const Mat& ref_frame, Mat& out_frame,
	const ImRegParams& params,
	const Size target_size,
	int interpolation, int borderMode)
{
	// Sanity check
	if(!params.valid()) return;

	// Calculate undistort/rectify maps
	// with ref_frame's image size as new source size,
	// and the specified target_size as new target size.
	ImRegUndistortRectifyMaps maps =
		params.calcUndistortRectifyMaps(
			ref_frame.size(), target_size);

	// Remapping
	remap_frame(ref_frame, out_frame,
		maps, interpolation, borderMode);
}

void remap_frame(
	const Mat& ref_frame, Mat& out_frame,
	const ImRegUndistortRectifyMaps& maps,
	int interpolation, int borderMode)
{
	// Sanity check
	if( maps.undistort_rectify_map1.size().area() == 0 ||
		maps.undistort_rectify_map2.size().area() == 0)
		return;

	remap(ref_frame, out_frame,
		maps.undistort_rectify_map1,
		maps.undistort_rectify_map2,
		interpolation, borderMode);
}

void remap_points(
	InputArray sourcePoints, OutputArray targetPoints,
	const ImRegParams& params)
{
	// Shortcut for call with original source/target image sizes
	return remap_points(sourcePoints, targetPoints,
		params.getOriginalSourceImgSize(),
		params.getOriginalTargetImgSize(),
		params);
}

void remap_points(
	InputArray sourcePoints, OutputArray targetPoints,
	const Size sourceImgSize, const Size targetImgSize, 
	const ImRegParams& params)
{
	// Sanity check
	if(!params.valid()) return;

	// Get source/target camera matrices, adjusted to the specified sizes (if needed)
	const Mat sourceCameraMatrix = params.getSourceCameraMatrix(sourceImgSize);
	const Mat targetCameraMatrix = params.getTargetCameraMatrix(targetImgSize);

	// Remapping
	undistortPoints(sourcePoints, targetPoints,
		sourceCameraMatrix,
		params.getDistCoeffs(),
		params.getRectification(),
		targetCameraMatrix);
}
