/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../stdafx.h"

#include "../utility/Util.h"

#include "SimAWB.h"

using namespace std;
using namespace cv;

void Simulation_AWB::grayWorld(Mat& input, Mat& dest)
{
	double meanRed, meanGreen, meanBlue, meanGray;

	Scalar mean, stdDev;
	double scaleRed, scaleGreen, scaleBlue;
	
	//get mean
	meanStdDev(input,mean,stdDev);

	//opencv saves in BGR, read mean values of red, green and blue channel
	meanRed = mean[2];
	meanGreen = mean[1];
	meanBlue = mean[0];
	meanGray = (meanRed + meanGreen + meanBlue) / 3;

	//calculate coefficients
	scaleRed = meanGray/meanRed;
	scaleGreen = meanGray/meanGreen;
	scaleBlue = meanGray/meanBlue;
	
	//split color channels and multiply with coefficients
	vector<Mat> bgr_planes;
	split(input, bgr_planes);

	bgr_planes[2].convertTo(bgr_planes[2],0,scaleRed,0);
	bgr_planes[1].convertTo(bgr_planes[1],0,scaleGreen,0);
	bgr_planes[0].convertTo(bgr_planes[0],0,scaleBlue,0);
	
	//merge channels for output mat
	merge(bgr_planes, dest);
}

void Simulation_AWB::weng(Mat& input, Mat& dest)
{    
	// Set the output image to the input for saftey (so it's not empty)
	input.copyTo( dest ) ;

	int numPartitions = 12 ;

	vector<double> meanCr ; meanCr.reserve(numPartitions) ;
	vector<double> meanCb ; meanCb.reserve(numPartitions) ;
	vector<double> meanAbsDiffCr ; meanAbsDiffCr.reserve(numPartitions) ;
	vector<double> meanAbsDiffCb ; meanAbsDiffCb.reserve(numPartitions) ;
	
	double globalMeanCr = 0;
	double globalMeanCb = 0; 
	double globalAbsDiffMeanCr = 0; 
	double globalAbsDiffMeanCb = 0;

	vector<Mat> ycrcb_planes;
	vector<Point> coordinates;
	
	//convert to YCrCb
	Mat img_ycrcb;
	cvtColor(input, img_ycrcb, CV_BGR2YCrCb);
	
	double minCrDiff = 3 ;
	double minCbDiff = 3 ;

	// split in 12 parts and calculate mean value
	// This is supposed to be so that if one section has a very low mean it can be discarded
	// so it doesn't affect the final result too much. Unfortunately the paper is very unclear on
	// what "too low" means, so it is not clear what to remove. I have set the minimum absolute
	// difference requirement to 0.5. Set it to 0 if you wish to keep everything.
	for( int rows = 0 ; rows < 3 ; rows++ )
	{
		for( int cols = 0 ; cols < 4 ; cols++ )
		{
			//image, width, height
			vector<double> retValues = wengCalcMeanAbsDiff( img_ycrcb, rows, cols ) ;

			if( retValues[2] >= minCrDiff && retValues[3] >= minCbDiff)
			{
				meanCr.push_back( retValues[0] );
				meanCb.push_back( retValues[1] ) ;
				meanAbsDiffCr.push_back( retValues[2] ) ;
				meanAbsDiffCb.push_back( retValues[3] ) ;
			}
		}
	}
	
	//sum up values and calc global mean value
	globalMeanCr = accumulate( meanCr.begin(), meanCr.end(), 0.0 ) / meanCr.size() ;
	globalMeanCb = accumulate( meanCb.begin(), meanCb.end(), 0.0 ) / meanCb.size() ;
	globalAbsDiffMeanCr = accumulate( meanAbsDiffCr.begin(), meanAbsDiffCr.end(), 0.0 ) / meanAbsDiffCr.size() ;
	globalAbsDiffMeanCb = accumulate( meanAbsDiffCb.begin(), meanAbsDiffCb.end(), 0.0 ) / meanAbsDiffCb.size() ;
	
	double tcond1 = globalMeanCb + globalAbsDiffMeanCb * Util::signum( globalMeanCb ) ;
	double tcond2 = 1.5 * globalAbsDiffMeanCb  ;
	double tcond3 = globalMeanCr + globalAbsDiffMeanCr * Util::signum( globalMeanCr ) ;
	double tcond4 = 1.5 * globalAbsDiffMeanCr ;
	
	//select reference points w/ postive Near-White-Region condition
	for( int i = 0 ; i < img_ycrcb.rows ; i++ )
	{
		for( int j = 0 ; j < img_ycrcb.cols ; j++ )
		{
			double tmp1 = img_ycrcb.at<Vec3b>( i, j )[2] ;
			double tmp2 = img_ycrcb.at<Vec3b>( i, j )[1] ;
			if( (  abs( tmp1 - tcond1 ) < tcond2 ) 
			   && (abs( tmp2 - tcond3 ) < tcond4 ) )
			{
					//Save coordinates of reference points
					coordinates.push_back(Point(i,j));
			}
		}
	}
	
	vector<Point> top10coordinates ;
	top10coordinates = wengGetReferencePoints(img_ycrcb,coordinates) ;
	
	// If no reference points could be found, return.
	if(top10coordinates.size() == 0)
		return ;
	
	double meanValBlue, meanValRed, meanValGreen;
	meanValBlue  = 0 ;
	meanValRed   = 0 ;
	meanValGreen = 0 ;

	double maxPartialLum = 0 ;
	double meanPartialLum = 0 ;
	//calc mean value of rgb-values of top10 reference points
	for(unsigned int n = 0 ; n < top10coordinates.size() ; n++)
	{
		double bVal = input.at<Vec3b>(top10coordinates[n].x,top10coordinates[n].y)[0] ;
		double gVal = input.at<Vec3b>(top10coordinates[n].x,top10coordinates[n].y)[1] ;
		double rVal = input.at<Vec3b>(top10coordinates[n].x,top10coordinates[n].y)[2] ;
		
		meanValBlue += bVal ;
		meanValGreen += gVal ;
		meanValRed += rVal ;
		
		double currentLum = 0.299 * rVal + 0.587 * gVal + 0.114 * bVal ;
		
		if(currentLum > maxPartialLum)
			maxPartialLum = currentLum ;
		
		meanPartialLum += currentLum ;
	}

	meanPartialLum /= top10coordinates.size();
	meanValBlue /= top10coordinates.size();
	meanValGreen /= top10coordinates.size();
	meanValRed /= top10coordinates.size();

	// Find the maximum luminance value
	vector<Mat> ycbcr_planes ;
	split(img_ycrcb, ycbcr_planes) ;
	
	double minVal, maximumLuminance ;
	Point minLoc, maxLoc ;
	
	minMaxLoc( ycbcr_planes[0], &minVal, &maximumLuminance, &minLoc, &maxLoc ) ;

	Scalar mean1, stdDev1 ;
	meanStdDev(img_ycrcb, mean1, stdDev1);
	
	//calculate scale coefficents
	double scaleRed, scaleGreen, scaleBlue;
	
	maximumLuminance = meanPartialLum ;
	
	scaleBlue = maximumLuminance/meanValBlue;
	scaleGreen = maximumLuminance/meanValGreen;
	scaleRed = maximumLuminance/meanValRed;
	
	//split color channels
	vector<Mat> bgr_planes;
	split(input, bgr_planes);

	//multiply color channels w/ scale coefficients
	bgr_planes[0] *= scaleBlue ;
	bgr_planes[1] *= scaleGreen ;
	bgr_planes[2] *= scaleRed ;
	
	//merge color channels for dstImage
	merge(bgr_planes, dest);
}

vector<double> Simulation_AWB::wengCalcMeanAbsDiff(Mat img_ycrcb, int partitionH, int partitionW){

	vector<double> retValues ; 

	int width, height;
	width = img_ycrcb.size().width;
	height = img_ycrcb.size().height;
	
	//rect(x value, y value, width, height)
	Rect calculationMask(width /4  * partitionW, height / 3 * partitionH, width / 4, height / 3 ) ;
	Mat partImage = img_ycrcb( calculationMask ) ;
	
	Scalar mean, stdDev, diffMean, diffStdDev;
	Mat absDiffMat ;
	// Calculate mean of the image section
	meanStdDev(partImage,mean,stdDev); // Verified, works.
	
	// Take the absolute difference of the image section and the mean of each channel
	absdiff(partImage, mean, absDiffMat) ; // Verified, works.
	
	// Calculate the mean of the differences
	meanStdDev(absDiffMat, diffMean, diffStdDev) ; // Verified, works.
	
	retValues.push_back( mean[1] ) ; //MeanCr value in 0
	retValues.push_back( mean[2] ) ; //MeanCb value in 1
	retValues.push_back( diffMean[1] ) ; //AbsDiffCr value in 2
	retValues.push_back( diffMean[2] ) ; //AbsDiffCb value in 3
	
	return retValues;
}

vector<Point> Simulation_AWB::wengGetReferencePoints(Mat& img_ycrcb, vector<Point>& coordinates)
{
	vector<cv::Point> top10coordinates;
	
	BrightnessPoints brightPts ;
	brightPts.reserve(10) ;
	
	for(unsigned int k = 0 ; k < coordinates.size() ; k++ )
	{
		// Make a pair consisting of the brightness value, and the coordinate
		pair<int, Point> pr = make_pair(
			img_ycrcb.at<Vec3b>(coordinates[k].x,coordinates[k].y)[0],
			coordinates[k]) ;
		
		brightPts.push_back(pr) ; 
	}

	if(brightPts.size() == 0)
		return top10coordinates ;
	else

	//sort works in ascending order.
	sort(brightPts.begin(), brightPts.end(), 
		[](const std::pair<int,cv::Point> &left, const std::pair<int,cv::Point> &right) -> bool
		{ return left.first < right.first; });
	
	//last 10% mean value. Take 1/10 of the brightness vector
	int last10percent = (int)( 0.1 * brightPts.size() ) + 1 ;
	
	top10coordinates.resize(last10percent);

	int numBrightnessValues = (int)brightPts.size();
	int startingIndex = numBrightnessValues - last10percent ;
	
	//extract top10 coordinates
	for(int j= startingIndex ; j < numBrightnessValues ; j++ )
		top10coordinates.push_back( brightPts[j].second ) ;
	
	return top10coordinates ;
}
