/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../stdafx.h"

#include "../utility/VideoReader.h"
#include "../utility/ThreadCout.h"
#include "../utility/Version.h"

#include "../common/Scenario.h"

#include "SimSpatial.h"

using namespace std;
using namespace cv;
using namespace boost::filesystem;

// Static member initialization
double Simulation_Spatial::fallbackSamplingFrameRate = 25;

Simulation_Spatial::Simulation_Spatial(void)
	: aspect_ratio_mode(Scenario::AR_RESIZE),
	  extendBorderPx(0), fadeBorderPx(0),
	  samplingFrameRate(fallbackSamplingFrameRate)
{
	// Detect CUDA support
	int numCUDADevs = cv::gpu::getCudaEnabledDeviceCount();
	if(numCUDADevs > 0) hasCUDASupport = true;
	else hasCUDASupport = false;
}

Simulation_Spatial::~Simulation_Spatial(void)
{
}

void Simulation_Spatial::setup(
	string transform_file, Size frame_size, 
	Size display_resolution, Size camera_resolution,
	Scenario::AspectRatioMode ar_mode,
	unsigned int extendBorderPx, unsigned int fadeBorderPx)
{
	// If a spatial transformation file is specified...
	if(!transform_file.empty())
	{
		// Read the specified transform file...
		readTransformFile(transform_file);

		// Get the transforms original source/target size...
		Size transform_source_size = spatial_registration.front().reg_params.getOriginalSourceImgSize();
		Size transform_target_size = spatial_registration.front().reg_params.getOriginalTargetImgSize();

		assert(transform_source_size.area() > 0);
		assert(transform_target_size.area() > 0);

		// Get the transforms original source/target aspect ratios
		float transform_source_ar = (float)transform_source_size.width / transform_source_size.height;
		float transform_target_ar = (float)transform_target_size.width / transform_target_size.height;

		// Handle special values
		if(display_resolution.width == -1) display_resolution.width = frame_size.width;
		if(display_resolution.height == -1) display_resolution.height = frame_size.height;
		if(display_resolution.width == -2) display_resolution.width = transform_source_size.width;
		if(display_resolution.height == -2) display_resolution.height = transform_source_size.height;

		// Validity check
		if(display_resolution.width < 0) throw runtime_error("Spatial distortion simulation: Negative width in display resolution.");
		if(display_resolution.height < 0) throw runtime_error("Spatial distortion simulation: Negative height in display resolution.");

		// Fill display resolution's width/height with source frame size if not specified,
		// but force the aspect ratio of the transformation's source size 
		if(display_resolution.area() == 0) fillMissing(display_resolution, frame_size, transform_source_ar);

		// Handle special values
		if(camera_resolution.width == -1) camera_resolution.width = frame_size.width;
		if(camera_resolution.height == -1) camera_resolution.height = frame_size.height;
		if(camera_resolution.width == -2) camera_resolution.width = transform_target_size.width;
		if(camera_resolution.height == -2) camera_resolution.height = transform_target_size.height;
		
		// Validity check
		if(camera_resolution.width < 0) throw runtime_error("Spatial distortion simulation: Negative camera resolution width.");
		if(camera_resolution.height < 0) throw runtime_error("Spatial distortion simulation: Negative camera resolution height.");

		// Fill camera resolution's width/height with display resolution if not specified
		// but force the aspect ratio of the transformation's target size 
		if(camera_resolution.area() == 0) fillMissing(camera_resolution, display_resolution, transform_target_ar);
	}
	
	// If no transform file is specified...
	else
	{
		// Handle special values
		if(display_resolution.width == -1) display_resolution.width = frame_size.width;
		if(display_resolution.height == -1) display_resolution.height = frame_size.height;
		if(display_resolution.width == -2) throw runtime_error("Spatial distortion simulation: Cannot set display resolution width to calibration width, since no transform file was specified.");
		if(display_resolution.height == -2) throw runtime_error("Spatial distortion simulation: Cannot set display resolution height to calibration height, since no transform file was specified.");

		// Validity check
		if(display_resolution.width < 0) throw runtime_error("Spatial distortion simulation: Negative width in display resolution.");
		if(display_resolution.height < 0) throw runtime_error("Spatial distortion simulation: Negative height in display resolution.");

		// Fill display resolution's width/height with source frame size if not specified
		if(display_resolution.area() == 0) fillMissing(display_resolution, frame_size);

		// Handle special values
		if(camera_resolution.width == -1) camera_resolution.width = frame_size.width;
		if(camera_resolution.height == -1) camera_resolution.height = frame_size.height;
		if(camera_resolution.width == -2) throw runtime_error("Spatial distortion simulation: Cannot set camera resolution width to calibration width, since no transform file was specified.");
		if(camera_resolution.height == -2) throw runtime_error("Spatial distortion simulation: Cannot set camera resolution height to calibration height, since no transform file was specified.");

		// Validity check
		if(camera_resolution.width < 0) throw runtime_error("Spatial distortion simulation: Negative camera resolution width.");
		if(camera_resolution.height < 0) throw runtime_error("Spatial distortion simulation: Negative camera resolution height.");

		// Fill camera resolution's width/height with display resolution if not specified
		if(camera_resolution.area() == 0) fillMissing(camera_resolution, display_resolution);

		// Create an identity transform based on the display size
		loadIdentityTransformation(display_resolution);
	}

	// Store all parameters
	this->displayResolution = display_resolution;
	this->cameraResolution = camera_resolution;
	this->aspect_ratio_mode = ar_mode;
	this->extendBorderPx = extendBorderPx;
	this->fadeBorderPx = fadeBorderPx;
}

void Simulation_Spatial::loadIdentityTransformation(Size initializationSize)
{
	SpatialRegWithBg identityReg;

	// Simply add one neutral (identity) transformation
	identityReg.reg_params.initializeNeutral(initializationSize);
	spatial_registration.push_back(identityReg);

	// Initialize loop indices
	loop_indices.push_back(0);
}

void Simulation_Spatial::readTransformFile(std::string spatialTransformFile)
{
	path transformFilePath = spatialTransformFile;
	transformFile = transformFilePath.make_preferred().string();

	FileStorage fs(transformFile, FileStorage::READ);
	if(!fs.isOpened()) throw runtime_error("Spatial transform file: path does not exist (" + transformFile + ")");

	fs["SamplingFrameRate"] >> samplingFrameRate;
	if(samplingFrameRate == 0)
	{
		samplingFrameRate = fallbackSamplingFrameRate;
		TLOG("Spatial transform file does not provide a sampling frame rate, using fallback of "
			<<  fallbackSamplingFrameRate << " fps");
	}
	else TLOG("Spatial transform file: sampling rate = " <<  samplingFrameRate << " fps");

	fs["BackgroundFile"] >> bgFile;
	if(!bgFile.empty())
	{
		path parent_folder = transformFilePath.parent_path();
		path bgFilePath = parent_folder / path(bgFile);
		bgFile = bgFilePath.make_preferred().string();

		TLOG("Spatial transform file: background = " <<  bgFilePath.filename());
	}

	FileNode spatial_node = fs["SpatialRegistration"];
	for(FileNodeIterator it = spatial_node.begin();
		it != spatial_node.end(); ++it)
	{
		const FileNode &node = *it;

		SpatialRegWithBg regWithBg;
		node >> regWithBg.reg_params;
				
		spatial_registration.push_back(regWithBg);
	}
		
	if(spatial_registration.empty())
		throw runtime_error("Spatial transform file does not contain any spatial registration information");

	// Initialize loop indices
	int numSpatialRegs = (int)spatial_registration.size();
	for(int i = 0; i < numSpatialRegs; i++) loop_indices.push_back(i);
	for(int i = numSpatialRegs - 2; i > 0; i--) loop_indices.push_back(i);
}

void Simulation_Spatial::loadBgVideoToInMemPNGs()
{
	// If there is no background data, skip this step
	if(bgFile.empty()) return;

	VideoReader capture(bgFile);
	if(!capture.isOpened()) throw runtime_error(
		"Error: Spatial background file could not be opened (" + bgFile + ").");

	// Set properties on the file-capture
	capture.set(CV_CAP_PROP_CONVERT_RGB, 1);

	// Setup PNG compression parameters
	// to apply highst compression rate
	vector<int> params(2);
	params[0] = CV_IMWRITE_PNG_COMPRESSION;
	params[1] = 9;

	// Iterate over all frames
	Mat bgFrame, resizedBgFrame;

	unsigned int frameIdx = 0;
	unsigned int lastLogFrameIdx = 0;

	unsigned int numSptialRegEntries =
		(unsigned int)spatial_registration.size();

	while(frameIdx < numSptialRegEntries)
	{
		// Read input video frame
		capture >> bgFrame;
		if(bgFrame.empty())
			break;
		
		// Resize background to specified display resolution
		resize(bgFrame, resizedBgFrame, cameraResolution);

		// Encode the frame as PNG and store in in-memory
		auto &pngBuffer = spatial_registration[frameIdx].pngCompressedBgImg;
		imencode(".png", resizedBgFrame, pngBuffer, params);

		// Assure minimum required memory usage
		pngBuffer.shrink_to_fit();

		frameIdx++;

		// Log progress every 10%, but not more
		// often than every 50 spatial frames.
		if(frameIdx > lastLogFrameIdx + 50)
		{
			int progressPerc = (int)((frameIdx / (float)numSptialRegEntries) * 100);
			if(progressPerc % 10 == 0)
			{
				TLOG(progressPerc << "% done");
				lastLogFrameIdx = frameIdx;
			}
		}
	}

	if(frameIdx != numSptialRegEntries)
	{
		TLOG("Warning: The number of spatial distortion\n"
			 "entries (" << numSptialRegEntries << ") and "
			 "background frames (" << frameIdx << ") differs.");
	}
}

Mat Simulation_Spatial::distort(Mat inputFrame, double time) const
{
	unsigned int spatialRegIdxFromTime = (unsigned int)(time * samplingFrameRate);
	unsigned int spatialRegIdx = loop_indices[spatialRegIdxFromTime % loop_indices.size()];
	const SpatialRegWithBg &spatial_reg = spatial_registration[spatialRegIdx];

	Mat dist_frame;

	// We can only compute a distortion if
	// we have valid image registration data
	if(spatial_reg.reg_params.valid())
	{
		Mat displayFrame;
		if(inputFrame.size() == displayResolution)
		{
			// No adjustments necessary if input size
			// equals simulated display resolution
			displayFrame = inputFrame;
		}
		else
		{
			// Simulate a certain display resolution by
			// resampling the input with this resolution.
			// Consider selected aspect-ratio mode.
			createDisplayFrame(inputFrame, displayResolution,
							   aspect_ratio_mode, displayFrame);
		}

		// Optionally extend frame by N extrapolated border pixels each side
		Mat extendedInputFrame;
		if(extendBorderPx > 0)
			addExtrapolatedBorder(displayFrame, extendBorderPx, extendedInputFrame);
		else extendedInputFrame = displayFrame;

		// Optionally fade to black over N border pixels each side
		if(fadeBorderPx > 0)
			darkenBorder(extendedInputFrame, fadeBorderPx * 2);

		// Adjust the distorion data to our source/target frame sizes
		// and the additionally added border, and create resampling maps
		ImRegUndistortRectifyMaps maps = spatial_reg.reg_params.
			calcUndistortRectifyMaps(displayResolution, cameraResolution,
									 Point(extendBorderPx, extendBorderPx));

		// Distort the frame spatially by using the resampling maps
		remap_frame(extendedInputFrame, dist_frame, maps, INTER_LINEAR);
	}
	else
	{
		// If we don't have valid distortion data,
		// we have to return a black frame of camera size
		dist_frame = Mat::zeros(cameraResolution, inputFrame.type());
	}

	// If there is a background image, decode it
	if(spatial_reg.pngCompressedBgImg.size() > 0)
	{
		Mat background = imdecode(
			spatial_reg.pngCompressedBgImg, CV_LOAD_IMAGE_COLOR);

		if(!background.empty())
		{
			// Add frame and background
			dist_frame += background;
		}
		else cerr << "Warning: Background frame " << spatialRegIdx
				  << " could not be inserted." << endl;
	}

	return dist_frame;
}

void Simulation_Spatial::fillMissing(
	Size &input_size, const Size &ref_size, double ar)
{
	float ref_ar = (float)ref_size.width / ref_size.height;
	if(ar == 0) ar = ref_ar;
	
	if(input_size.width == 0 && input_size.height == 0)
	{
		if(ref_ar == ar)
		{
			input_size = ref_size;
		}
		else if(ar > ref_ar)
		{
			input_size.width = int(ref_size.height * ar);
			input_size.height = ref_size.height;
		}
		else
		{
			input_size.width = ref_size.width;
			input_size.height = int(ref_size.width / ar);
		}
	}
	else if(input_size.width == 0)
	{
		input_size.width = int(input_size.height * ar);
	}
	else if(input_size.height == 0)
	{
		input_size.height = int(input_size.width / ar);
	}
}

void Simulation_Spatial::createDisplayFrame(
	Mat &src, const Size display_res,
	Scenario::AspectRatioMode ar_mode, Mat &out)
{
	float source_ar = (float)src.size().width / src.size().height;
	float display_ar = (float)display_res.width / display_res.height;

	// If the aspect ratio is the same, we can simply resize
	if(source_ar == display_ar) ar_mode = Scenario::AR_RESIZE;

	switch(ar_mode)
	{
		// Simple resizing, does not preserve aspect ratio
		case Scenario::AR_RESIZE:
		{
			// The resizing function automatically makes sure
			// that the output frame is of correct size and type.
			resize(src, out, display_res, INTER_LINEAR);
			break;
		}
		
		// Letterboxing, preserves aspect ratio
		case Scenario::AR_LETTERBOX:
		{
			// (Re-)Allocate output matrix if necessary and
			// completely fill the output matrix with black.
			if(out.size() != display_res || out.type() != src.type())
				 out = Mat(display_res, src.type(), Scalar::all(0));
			else out.setTo(Scalar::all(0));
			
			// Add black bars at top and bottom
			if(source_ar > display_ar)
			{
				int ar_preserving_height = int(display_res.width / source_ar);
				resize(src, src, Size(display_res.width, ar_preserving_height));

				int diff = display_res.height - ar_preserving_height;
				Rect roi(0, diff / 2, display_res.width, display_res.height - diff);
				src.copyTo(out(roi));
			}
			// Add black bars left and right
			else
			{
				int ar_preserving_width = int(display_res.height * source_ar);
				resize(src, src, Size(ar_preserving_width, display_res.height));

				int diff = display_res.width - ar_preserving_width;
				Rect roi(diff / 2, 0, display_res.width - diff, display_res.height);
				src.copyTo(out(roi));
			}

			break;
		}

		// Cropping, preserves aspect ratio
		case Scenario::AR_CROP:
		{
			// (Re-)Allocate output matrix if necessary and
			// completely fill the output matrix with black.
			if(out.size() != display_res || out.type() != src.type())
				 out = Mat(display_res, src.type(), Scalar::all(0));
			else out.setTo(Scalar::all(0));
			
			// Crop source at top and bottom
			if(source_ar < display_ar)
			{
				int ar_preserving_height = int(display_res.width / source_ar);
				resize(src, src, Size(display_res.width, ar_preserving_height));

				int diff = ar_preserving_height - display_res.height;
				Rect roi(0, diff / 2, src.size().width, src.size().height - diff);
				src(roi).copyTo(out);
			}
			// Crop source at left and right
			else
			{
				int ar_preserving_width = int(display_res.height * source_ar);
				resize(src, src, Size(ar_preserving_width, display_res.height));

				int diff = ar_preserving_width - display_res.width;
				Rect roi(diff / 2, 0, src.size().width - diff, src.size().height);
				src(roi).copyTo(out);
			}

			break;
		}

		// Invalid mode
		default: throw runtime_error(
			"Simulation_Spatial: Invalid aspect ratio mode");
	}
}

void Simulation_Spatial::addExtrapolatedBorder(
	const Mat &src, unsigned int borderWidthPx, Mat &out)
{
	// Calculate output image size
	Size out_size(src.size().width  + borderWidthPx * 2,
				  src.size().height + borderWidthPx * 2);

	// (Re-)Allocate output image if size or type mismatches
	if(out.size() != out_size || out.type() != src.type())
		out = Mat::zeros(out_size, src.type());

	// Copy src image into the center of the output image
	Rect roi(borderWidthPx, borderWidthPx, src.size().width, src.size().height);
	src.copyTo(out(roi));

	// Extrapolate border rows and columns
	for(unsigned int i = 0; i < borderWidthPx; i++)
	{
		unsigned int offset = borderWidthPx - i;

		unsigned int topRowIdx = offset;
		unsigned int btmRowIdx = out.rows - 1 - offset;
		unsigned int leftColIdx = offset;
		unsigned int rghtColIdx = out.cols - 1 - offset;
			
		Mat newTopRow = out.row(topRowIdx - 1);
		Mat newBtmRow = out.row(btmRowIdx + 1);
		Mat newLeftCol = out.col(leftColIdx - 1);
		Mat newRghtCol = out.col(rghtColIdx + 1);

		out.row(topRowIdx).copyTo(newTopRow);
		out.row(btmRowIdx).copyTo(newBtmRow);
		out.col(leftColIdx).copyTo(newLeftCol);
		out.col(rghtColIdx).copyTo(newRghtCol);

		boxFilter(newTopRow, newTopRow, -1, Size(3,1));
		boxFilter(newBtmRow, newBtmRow, -1, Size(3,1));
		boxFilter(newLeftCol, newLeftCol, -1, Size(1,3));
		boxFilter(newRghtCol, newRghtCol, -1, Size(1,3));
	}
}

void Simulation_Spatial::darkenBorder(
	Mat &image, unsigned int borderWidthPx)
{
	// Darken border rows and columns
	for(unsigned int i = 0; i < borderWidthPx; i++)
	{
		unsigned int offset = borderWidthPx - i;

		unsigned int topRowIdx = offset - 1;
		unsigned int btmRowIdx = image.rows - offset;
		unsigned int leftColIdx = offset - 1;
		unsigned int rghtColIdx = image.cols - offset;
			
		Mat currTopRow = image.row(topRowIdx);
		Mat currBtmRow = image.row(btmRowIdx);
		Mat currLeftCol = image.col(leftColIdx);
		Mat currRghtCol = image.col(rghtColIdx);

		float darkening_factor = (1.0f / (borderWidthPx + 1)) * offset;

		currTopRow *= darkening_factor;
		currBtmRow *= darkening_factor;
		currLeftCol *= darkening_factor;
		currRghtCol *= darkening_factor;
	}
}
