/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../stdafx.h"

#include "../common/FrameSequence.h"
#include "../common/Frame.h"

#include "../simulation/SimDisplayPWM.h"

#include "SimTemporal.h"

using namespace std;
using namespace cv;

Simulation_Temporal::Simulation_Temporal(void)
{
	// Detect CUDA support
	int numCUDADevs = cv::gpu::getCudaEnabledDeviceCount();
	if(numCUDADevs > 0) hasCUDASupport = true;
	else hasCUDASupport = false;
}

Simulation_Temporal::~Simulation_Temporal(void)
{
}

void Simulation_Temporal::setup(
	double src_start_time, double src_end_time,
	double cam_start_time, double cam_end_time,
	double src_fps, double cam_fps,
	double display_refresh_rate, double display_g2g_time,
	double exposure_time_ms, double readout_time_ms,
	bool vsync, string dataSrcPathStr)
{
	// Create source frames
	createSourceFrameSequence(
		src_fps, src_start_time * 1000, src_end_time * 1000,
		vsync ? 1 : 0, display_refresh_rate, display_g2g_time);
	
	// Useful default if no exposure time is specified
	if(exposure_time_ms == 0)
		exposure_time_ms = (1000.0 / cam_fps) / 2.0;

	// Validity check
	if(readout_time_ms > exposure_time_ms) throw runtime_error(
		"The sensor readout (" + to_string(readout_time_ms) + " ms) time " +
		"must be shorter than the exposure time (" + to_string(exposure_time_ms) + " ms).");

	// Store camera-frame related parameters
	this->cam_fps = cam_fps;
	this->cam_start_time = cam_start_time * 1000;
	this->cam_end_time = cam_end_time * 1000;
	this->exposure_time = exposure_time_ms;
	this->readout_time = readout_time_ms;

	// Store data source path
	dataSrcPath = dataSrcPathStr;
	dataSrcPath = dataSrcPath.make_preferred();
}

unsigned int Simulation_Temporal::getCamFrameCount()
{
	int firstFrameNo = static_cast<int>((cam_start_time / 1000) * cam_fps);
	int lastFrameNo = static_cast<int>((cam_end_time / 1000) * cam_fps);

	return lastFrameNo > firstFrameNo ?
		   lastFrameNo - firstFrameNo : 0;
}

void Simulation_Temporal::createSourceFrameSequence(
	double src_fps, double start_time, double end_time, int vsync,
	double display_refresh_rate, double display_g2g_time)
{
	double last_frame_end_time = start_time;

	for(unsigned int nbr = 0; true; nbr++)
	{
		//determines the ideal start and end times based on the frame number
		double frame_start_time = start_time + (nbr / src_fps) * 1000;
		double frame_end_time = start_time + ((nbr + 1) / src_fps) * 1000;
		if (frame_end_time > end_time) break;

		//if we want to make vsync adjustments
		if (vsync) {

			//adjust the end time
			simulate_vsync(frame_end_time, display_refresh_rate);
			
			frame_start_time = last_frame_end_time;		// the current frame must start at the time when the last frame finished
			last_frame_end_time = frame_end_time;		//update $last_frame_end_time$ for the next iteration
		}

		//make the source frame and add to the sequence
		Frame src_frame(nbr, frame_start_time, frame_end_time, display_g2g_time);
		src_frame_seq.add(src_frame);
	}

	// Validate the sequence
	src_frame_seq.validate();
}

//rounds the time to the nearest multiple of the display refresh period
void Simulation_Temporal::simulate_vsync(double& time, double display_refresh_rate)
{ 
	double refresh_time_ms = 1000.0/display_refresh_rate;
	
	int n = static_cast<int>(abs(time) / refresh_time_ms);

	if (time >= 0) {
		if (time - n * refresh_time_ms > 0.5 * refresh_time_ms)
			time = refresh_time_ms * (n+1);
		else 				
			time = refresh_time_ms * n;
	}

	else {
		if (time + n * refresh_time_ms < -0.5 * refresh_time_ms)
			time = -1 * refresh_time_ms * (n+1);
		else
			time = -1 * refresh_time_ms * n;
	}
}

Simulation_Temporal::BlendDescription Simulation_Temporal::getBlendDescForFrame(
	unsigned int nbr, unsigned int row_count) const
{
	//determine the frame presentation times
	double frame_start_time = cam_start_time + (nbr / cam_fps) * 1000;
	double frame_end_time = cam_start_time + ((nbr + 1) / cam_fps) * 1000;

	if(frame_end_time > cam_end_time)
		return BlendDescription();
	
	BlendDescription blendDesc(row_count);
	FrameSequence<Frame>::SeqAccessCache cache;

	double row_increment_time = readout_time/row_count;
		
	//the exposure time of the frame is different for each rows due to the rolling shutter
	//N.B: we assume that the start time of the frame is the same as the exposure start time of the bottom row
	double row_exposure_start_time, row_exposure_end_time;

	for (unsigned int i = 0; i < row_count; i++)
	{
		row_exposure_start_time = frame_start_time - readout_time + i * row_increment_time;
		row_exposure_end_time = row_exposure_start_time + exposure_time;
			
		//get the source frames that were presented during the exposure period
		vector<Frame> influencing_src_frames = src_frame_seq.getFrameRange_seq(
			row_exposure_start_time, row_exposure_end_time, cache);
			
		for(auto src_frame : influencing_src_frames)
		{
			auto influence_factors = getInfluenceFactors(
				row_exposure_start_time, row_exposure_end_time,
				displayPWM, src_frame);

			SrcFrameInfluence src_frm_infl;
			src_frm_infl.number = src_frame.getNbr();
			src_frm_infl.share = influence_factors.first;
			src_frm_infl.power = influence_factors.second;
			
			// Add to blend description of this row
			blendDesc[i].push_back(src_frm_infl);
		}
	}

	return blendDesc;
}

pair<float, float> Simulation_Temporal::getInfluenceFactors(
	double exposure_start_time, double exposure_end_time,
	const Simulation_DisplayPWM &displayPWM,
	const Frame &src_frame)
{
	double src_fade_time = src_frame.getFadeTime();
	double src_start_time = src_frame.getStartTime();
	double src_end_time = src_frame.getEndTime();

	double src_full_start_time = src_start_time + src_fade_time;
	double src_full_end_time = src_end_time - src_fade_time;

	if (src_start_time >= exposure_end_time || src_end_time <= exposure_start_time)
		return make_pair(0.0f, 0.0f);

	double influence_amount = 0;

	//sections without fading
	double src_influence_start = max(exposure_start_time, src_full_start_time);
	double src_influence_end = min(exposure_end_time, src_full_end_time);
	if (src_influence_end - src_influence_start > 0) influence_amount += src_influence_end - src_influence_start;

	//fading in
	double a = max(src_start_time, exposure_start_time) - src_start_time;
	double b = min(src_full_start_time, exposure_end_time) - src_start_time;
	if (b-a > 0) influence_amount += (b*b - a*a)/(2*src_fade_time);
			
	//fading out
	a = max(src_full_end_time, exposure_start_time) - src_full_end_time;
	b = min(src_end_time, exposure_end_time) - src_full_end_time;
	if (b-a > 0) influence_amount += (a*a - b*b)/(2*src_fade_time) - a + b;
		
	double influence_factor = influence_amount / (exposure_end_time - exposure_start_time);
	if(influence_factor > 1.0) influence_factor = 1.0; // Might be due to rounding errors

	// Power factor
	src_influence_start = max(exposure_start_time, src_start_time);
	src_influence_end = min(exposure_end_time, src_end_time);

	double full_power = 1.0 * (src_influence_end - src_influence_start);
	double actual_power = displayPWM.integrate(src_influence_start, src_influence_end);
	double power_factor = actual_power / full_power;

	// Return share and power
	return make_pair(float(influence_factor),
					 float(power_factor));
}

set<unsigned int> Simulation_Temporal::getInvoledFrameNumbers(BlendDescription blendDesc)
{
	set<unsigned int> involed_frames;

	// Iterate over all frame rows of the blending description
	// to find all involved source frames.
	for(auto &row : blendDesc)
	{
		// Iterate over all source frame shares that this row is blended from
		for(auto &srcFrameInfluence : row)
		{
			// Insert the involved source frame number into set
			involed_frames.insert(srcFrameInfluence.number);
		}
	}

	return involed_frames;
}

cv::Mat Simulation_Temporal::blendFrames(
	const BlendDescription &blendDesc,
	const std::map<unsigned int, cv::Mat> &srcFrames)
{
	// Create a black frame of source frame size
	const Mat &anyFrame = srcFrames.begin()->second;
	Mat blended = Mat::zeros(anyFrame.size(), anyFrame.type());

	for(int row = 0; row < blended.rows; ++row)
	{
		for(const SrcFrameInfluence &infl : blendDesc[row])
		{
			// Get source frame
			const auto it = srcFrames.find(infl.number);

			// If blend data is incomplete and the
			// source frame missing, skip this share.
			if(it == srcFrames.end()) continue;

			// Get the Mat object of the frame,
			// skip shares with empty Mat objects.
			const Mat &src_frame = it->second;
			if(src_frame.empty()) continue;

			// Add source frame to blend
			//blended.row(row) += src_frame.row(row) * infl.share * infl.power;

			const uchar* src_frame_row = src_frame.ptr(row);
			uchar*   blended_frame_row = blended.ptr(row);

			for(int col = 0; col < blended.cols * 3; ++col)
			{
				*blended_frame_row += (uchar)(*src_frame_row * infl.share * infl.power);

				blended_frame_row++;
				src_frame_row++;
			}
		}
	}

	return blended;
}
