/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../stdafx.h"

#include "../utility/VideoReader.h"
#include "../utility/BufferedVideoReader.h"
#include "../utility/SynchronizedList.h"
#include "../utility/SynchronizedQueue.h"
#include "../utility/SynchronizedPriorityQueue.h"
#include "../utility/ThreadCout.h"
#include "../utility/Version.h"

#include "../common/Scenario.h"
#include "../common/FrameSequence.h"
#include "../common/Frame.h"

#include "SimSpatial.h"
#include "SimBayerCFA.h"
#include "SimDisplayPWM.h"
#include "SimTemporal.h"
#include "SimAWB.h"
#include "SimAGC.h"

#include "Simulation.h"

using namespace std;
using namespace cv;

CamMark::CamMark(void)
	: maxQueueSize(10),
	  maxOrderingQueueSize(25),
	  maxAllowedBlendRange(10),
	  scenario(Version())
{
	// Detect CUDA support
	int numCUDADevs = cv::gpu::getCudaEnabledDeviceCount();
	if(numCUDADevs > 0) hasCUDASupport = true;
	else hasCUDASupport = false;
}

CamMark::~CamMark(void)
{
}

float CamMark::getProgress()
{
	return progress;
}

bool CamMark::doneProcessing()
{
	return finishedProcessing ;
}

void CamMark::cancel()
{
	if(noMoreSrcFrames == false)
	{
		TLOG( "Cancelling processing..." );

		// Just pretend there are no more source frames,
		// and let the processing finish regularly.
		noMoreSrcFrames = true;
	}
}

void CamMark::terminate()
{
	// TODO
	throw runtime_error("NOT YET IMPLEMENTED");
}

void CamMark::threadCallWrapper(boost::function<void(void)> workerMemFunc)
{
	try
	{
		workerMemFunc();
	}
	catch(boost::thread_interrupted)
	{
		ThreadCout::tcout.display("Thread Interrupted.");
	}
	catch(...)
	{
		ThreadCout::tcout.display("An exception occurred. ");
		boost::recursive_mutex::scoped_lock lock(exceptionMutex);
		exceptionPtrQueue.push(current_exception());
	}
}

void CamMark::rethrowIfPending()
{
	boost::recursive_mutex::scoped_lock lock(exceptionMutex);
	
	if(!exceptionPtrQueue.empty())
	{
		exception_ptr pendingException = exceptionPtrQueue.front();
		exceptionPtrQueue.pop();

		rethrow_exception(pendingException);
	}
}

bool CamMark::isExceptionPending()
{
	boost::recursive_mutex::scoped_lock lock(exceptionMutex);
	return !exceptionPtrQueue.empty();
}

void CamMark::startSimulation(string src_file, string out_file,
							  const Scenario &scenarioToSimulate)
{
	ThreadCout::tcout.setThreadName("appmain");

	// Store passed scenario
	scenario = scenarioToSimulate;

	// For videos, use XVID as default codec
	nCodec = CV_FOURCC('X', 'V', 'I', 'D');

	// Handle single frame (PNG) output differently
	boost::filesystem::path out_path(out_file);
	string out_file_ext = out_path.extension().string();
	Util::ci_string ci_ext(out_file_ext.begin(), out_file_ext.end());
	if(".png" == ci_ext)
	{
		// Add a frame counter to the filename
		string::size_type lastSep = Util::findLastDirectorySeperator(out_file);
		if(lastSep == string::npos) lastSep = 0; else lastSep++;
		out_file.insert(lastSep, "%06d_");

		// No video codec for single frames
		nCodec = 0;
	}
	
	nextSrcFrame = 0 ;
	nextCamFrame = 0 ;

	progress = 0.0f;
	finishedProcessing = false ;

	this->src_file = src_file;
	this->out_file = out_file;

	// Begin by starting the preprocessing thread, which in turn starts all processing threads.
	boost::function<void(void)> preprocessFn = boost::bind( &CamMark::prepare, this );
	preprocessThread = boost::thread(boost::bind( &CamMark::threadCallWrapper, this, preprocessFn )) ;
}

void CamMark::finishSimulation()
{
}

void CamMark::startProcessingThreads()
{
	// Get number of cores for multithreading
	int numCores = boost::thread::hardware_concurrency();
	TLOG("Detected CPU cores: " << numCores);

	// The reader, writer and stage 2 processing
	// always use one thread only (no concurrency)
	const int numStaticThreads = 3;

	// Always create at least N more threads than cores
	// (over-provisioning) for optimal performance
	const int overprovisioning = 2;

	// Get the number of threads available for asynchronous processing
	int numAvailable = numCores - numStaticThreads + overprovisioning;
	if(numAvailable < 1) numAvailable = 1;

	// Log statistics
	int numThreadsTotal = numStaticThreads + numAvailable;
	TLOG("Creating " << numAvailable << " processing thread(s).");
	TLOG("Creating a total of " << numThreadsTotal << " threads.");

	// Bind processing function pointers
	boost::function<void(void)> readerFn = boost::bind( &CamMark::videoReader, this );
	boost::function<void(void)> stage1ProcessingFn = boost::bind( &CamMark::stage1Processing, this );
	boost::function<void(void)> stage2ProcessingFn = boost::bind( &CamMark::stage2Processing, this );
	boost::function<void(void)> writerFn = boost::bind( &CamMark::videoWriter, this );

	// Each stage 1 processing thread needs an entry in this vector
	stage1NeededFrames.resize(numAvailable);

	// Create a single reader thread
	readerThread = boost::thread(boost::bind(&CamMark::threadCallWrapper, this, readerFn)) ;

	// Create stage 1 processing threads
	for(int i = 0 ; i < numAvailable; i++)
		stage1ThreadGroup.create_thread(boost::bind( &CamMark::threadCallWrapper, this, stage1ProcessingFn )) ;
	
	// Create a single stage 2 processing thread
	stage2Thread = boost::thread(boost::bind( &CamMark::threadCallWrapper, this, stage2ProcessingFn )) ;

	// Create a single writer thread
	writerThread = boost::thread(boost::bind( &CamMark::threadCallWrapper, this, writerFn )) ;
}

void CamMark::prepare()
{
	// Used to perform any necessary preparation
	// before then starting all processing threads.

	ThreadCout::tcout.setThreadName("preparation");
	TLOG("Thread started");

	// Open source video file-capture
	capture.open(src_file);
	if(!capture.isOpened()) throw runtime_error(
		"Could not open source video file for reading (" + src_file + ")");

	// Set properties on the file-capture
	capture.set(CV_CAP_PROP_CONVERT_RGB, 1);

	// Read and store frame rate property
	src_fps = capture.getFPS();

	// Is source frame rate overwritten?
	if(scenario.override_src_FPS > 0)
	{
		src_fps = scenario.override_src_FPS;
		TLOG("Automatically detected source video frame rate is overwritten with " << scenario.override_src_FPS << " fps");
	}

	// If the source frame rate cannot be determined automatically, quit and
	// report a note to specify the correct frame rate through a parameter.
	if(ISNAN(src_fps) || src_fps == 0) throw runtime_error(
		"Source video frame rate cannot be determined from the specified source video.\n"
		"Use the OverrideSourceFPS tag in the XML scenario file to specify the correct frame rate.");

	// If not specified, set display's refresh rate to source fps
	if(scenario.temporal.display_refresh_rate == 0) scenario.temporal.display_refresh_rate = src_fps;
	TLOG("Final display refresh rate is " << scenario.temporal.display_refresh_rate << " Hz");

	// If not specified, set camera's fps to source fps
	if(scenario.temporal.cam_FPS == 0) scenario.temporal.cam_FPS = src_fps;
	TLOG("Final camera frame rate is " << scenario.temporal.cam_FPS << " fps");

	// Get source frame size
	Size frameSize = capture.getSize();

	// Sanity check
	if(frameSize.area() == 0) throw runtime_error(
		"Unable to detect source frame size. Is the specified source file valid?");

	// Setup spatial distortion
	if(scenario.applySpatialDistortion)
	{
		// Setup remaining parameters
		spatial.setup(scenario.spatial.transform_file, frameSize,
					  scenario.spatial.display_resolution,
					  scenario.spatial.camera_resolution,
					  scenario.spatial.ar_mode,
					  scenario.spatial.extendBorderPx,
					  scenario.spatial.fadeBorderPx);

		// The display/camera resolution might have been completed by the setup call
		scenario.spatial.display_resolution = spatial.getFinalDisplayResolution();
		scenario.spatial.camera_resolution = spatial.getFinalCameraResolution();

		// Output final display/camera resolution
		TLOG("Final display resolution is " << scenario.spatial.display_resolution.width
								   << " x " << scenario.spatial.display_resolution.height);
		TLOG("Final camera resolution is  " << scenario.spatial.camera_resolution.width
								   << " x " << scenario.spatial.camera_resolution.height);

		// Load all frame background images into memory (as PNG compressed images)
		if(scenario.spatial.backgroundInsertion)
		{
			TLOG("Loading spatial background data...");
			spatial.loadBgVideoToInMemPNGs();
			TLOG("Loading spatial background data finished.");
		}
	}
	else
	{
		// If no spatial distortion is applied,
		// set display/camera to source frame size.
		scenario.spatial.display_resolution = frameSize;
		scenario.spatial.camera_resolution = frameSize;
	}

	// Read and store frame count
	numSrcFrames = capture.getFrameCount();

	// If the number of frames could not be determined,
	// do a preprocessing run to count them all.
	if(0 == numSrcFrames)
	{
		TLOG("Warning: Unable to determine source video frame count.\n"
			 "Thus, the video needs to be completely decoded once, counting its frames.");
	
		TLOG("Counting source video frames...");

		Mat dummyFrame;
		while(true)
		{
			// Decode frames until we reach the end
			capture >> dummyFrame;
			if(dummyFrame.empty()) break;
			
			// Increase count
			numSrcFrames++;
			
			// Log an update every 5000 frames
			if(numSrcFrames % 5000 == 0)
				TLOG("Found " << numSrcFrames << " frames, still counting...");
		}

		TLOG("Found " << numSrcFrames << " source video frames.");

		if(capture.seekToFrame(0) > 1) throw runtime_error(
			"Unable to seek to the beginning of the video after counting frames.");
	}

	// Time is relative to the source video time,
	// so the source video starts at zero seconds
	double srcStartTime = 0.0;
	double srcEndTime = numSrcFrames / src_fps;

	numSrcFramesToPrepend = 0;
	numSrcFramesToAppend = 0;

	// Log source video information
	TLOG("Source video information: " << numSrcFrames << " frames, " << src_fps << " fps");
	TLOG("Calculated source video length: " << srcEndTime << " sec");

	// If not specified, set end of source video as capture end time
	if (scenario.cam_end_time <= 0)
		scenario.cam_end_time = srcEndTime;

	// Check if the capture end time exceeds the source video's end time
	if(scenario.cam_end_time > srcEndTime)
	{
		double timeToAppend = scenario.cam_end_time - srcEndTime;

		TLOG("NOTE: The capture end time exceeds the end of the source video.\n"
			 "   Appending the source video with " + to_string(timeToAppend) + " sec of black frames.");

		srcEndTime += timeToAppend;
		numSrcFramesToAppend = (unsigned int)(timeToAppend * src_fps);
		numSrcFrames += numSrcFramesToAppend;
	}

	// Check if the capture start time preceeds the source video's start time
	if(scenario.cam_start_time < 0)
	{
		double timeToPrepend = -scenario.cam_start_time;

		TLOG("NOTE: The capture start time preceeds the start of the source video.\n"
			 "   Prepending the source video with " + to_string(timeToPrepend) + " sec of black frames.");

		scenario.cam_start_time = 0;
		scenario.cam_end_time += timeToPrepend;

		srcEndTime += timeToPrepend;
		numSrcFramesToPrepend = (unsigned int)(timeToPrepend * src_fps);
		numSrcFrames += numSrcFramesToPrepend;
	}

	// Check for values that can't be corrected
	if(scenario.cam_start_time >= srcEndTime) throw runtime_error(
		"The capture start time exceeds the source video's end time.");
	
	if(scenario.cam_start_time > scenario.cam_end_time) throw runtime_error(
		"The capture start exceeds the capture end time.");

	// Setup temporal distortion
	if(scenario.applyTemporalDistortion)
	{
		temporal.setup(
			srcStartTime, srcEndTime,
			scenario.cam_start_time, scenario.cam_end_time,
			src_fps, scenario.temporal.cam_FPS,
			scenario.temporal.display_refresh_rate, scenario.temporal.display_g2g_time,
			scenario.temporal.exposure_time_ms, scenario.temporal.readout_time_ms,
			scenario.temporal.vsync, scenario.temporal.blend_file);

		if(scenario.temporal.display_pwm_duty_cycle < 1.0)
		{
			Simulation_DisplayPWM displayPWM;
			displayPWM.setup(scenario.temporal.display_pwm_duty_cycle,
							 scenario.temporal.display_pwm_frequency);
			
			temporal.setDisplayPWM(displayPWM);
		}
	}
	
	if(scenario.applyTemporalDistortion)
		 numCamFrames = temporal.getCamFrameCount();
	else numCamFrames = (unsigned int)((scenario.cam_end_time - scenario.cam_start_time) * src_fps);
	
	TLOG("Calculated output frame count: " << numCamFrames);

	// Create output video file writer
	writer.open(out_file, nCodec, scenario.temporal.cam_FPS, scenario.spatial.camera_resolution);
	if(!writer.isOpened()) throw runtime_error(
		"Could not open output video file for writing (" + out_file + ")");

	noMoreSrcFrames = false;
	noMoreCamFrames = false;

	// Now start the processing threads
	startProcessingThreads();
}

void CamMark::videoReader()
{
	// Used to decode the source video
	// and fill the queues of stage 1

	ThreadCout::tcout.setThreadName("reader");
	TLOG("Thread started");

	// Read N frames ahead of the frames currently needed by stage 1
	const unsigned int readAheadCount = 25;

	// If temporal distortion is applied, these values are constantly re-calculated;
	// otherwise they never change and can thus be initialized once before starting.
	unsigned int loweststage1FrameNeeded, higheststage1FrameNeeded;

	if(scenario.applyTemporalDistortion == false)
	{
		// Without temporal distortion, still consider the specified capturing
		// time range and initialize the needed source frames accordingly.
		loweststage1FrameNeeded = (unsigned int)(scenario.cam_start_time * src_fps);
		higheststage1FrameNeeded = (unsigned int)(scenario.cam_end_time * src_fps);
	}
	else
	{
		// With temporal distortion, just initialize with values marking them as
		// un-initialized; otherwise, compilers complain (potentially un-inilialized).
		loweststage1FrameNeeded = UINT_MAX;
		higheststage1FrameNeeded = 0;
	}

	while(true)
	{
		// If temporal distortion is applied, all stage 1 processing threads
		// specify which source frames they need for the current capture frame.
		// The highest and lowest needed is thus constantly updated.
		if( scenario.applyTemporalDistortion )
		{
			// Lock mutex
			stage1NeededFramesMutex.lock();

			loweststage1FrameNeeded = UINT_MAX;
			higheststage1FrameNeeded = 0;

			for(const auto &thread_range : stage1NeededFrames)
			{
				// Find lowest frame that is needed by stage 1
				if(thread_range.lowest < loweststage1FrameNeeded)
					loweststage1FrameNeeded = thread_range.lowest;
			
				// Find highest frame that is needed by stage 1
				if(thread_range.highest > higheststage1FrameNeeded)
					higheststage1FrameNeeded = thread_range.highest;
			}

			// Release mutex
			stage1NeededFramesMutex.unlock();
		}
		
		// Determine if we need to decode the next source frame
		bool decodeNextSrcFrame;
		
		if( scenario.applyTemporalDistortion )
		{
			// If temporal distortion is applied, stage 1 controls the pipeline:
			// If the next source frame number is lower or equal to the highest frame
			// number that is needed by stage 1 plus a read-ahead count of N frames,
			// decode the next source frame and enqueue it for stage 1 processing.
			 decodeNextSrcFrame = nextSrcFrame <= higheststage1FrameNeeded + readAheadCount;
		}
		else
		{
			// If no temporal distortion is applied, the reader controls the pipeline.
			// We have to implement some flow control, so only decode new frames
			// if the stage 1 input list is not full.
			decodeNextSrcFrame = stage1InputList.getSize() < maxQueueSize;

			// But also check if we reached the end of the specified source frame range
			if(nextSrcFrame > higheststage1FrameNeeded)
			{
				TLOG( "Reached capture end time." );
				noMoreSrcFrames = true;
			}
		}

		// If we should not decode a new frame, wait for some time...
		if(decodeNextSrcFrame == false)
			boost::this_thread::sleep_for(boost::chrono::milliseconds(100));

		// ...but if we should decode and still have source frames:
		else if(!noMoreSrcFrames)
		{
			FrameWithIdx srcFrame;
			
			// Set next source frame number
			srcFrame.idx = nextSrcFrame++;
			
			// Did we reach the specifed number of source frames?
			if(srcFrame.idx >= numSrcFrames)
			{
				// If yes, we stop by setting the no-more-src-frames flag
				TLOG( "Reached end of source video." );
				noMoreSrcFrames = true;
			}
			else
			{
				// Check if we are before/past the actual source frame boundaries
				if( srcFrame.idx < numSrcFramesToPrepend ||
					srcFrame.idx >= numSrcFrames - numSrcFramesToAppend)
				{
					// In this case, we just insert a black frame
					srcFrame.fr = Mat::zeros(capture.getSize(), CV_8UC3);
				}
				else
				{
					// Otherwise, we decode the next source frame
					capture >> srcFrame.fr;
				}

				// If we read an empty frame from the source video (we reached its end)
				// although we haven't reached the calculated frame count yet,
				// we deliver a black frame instead and log a warning message.
				if(srcFrame.fr.empty())
				{
					// Show warning if not already done so
					static bool emptyFrameWarningShown = false;
					if(emptyFrameWarningShown == false)
					{
						unsigned int actualSrcVideoFrame = srcFrame.idx - numSrcFramesToPrepend;
						unsigned int actualLastSrcVideoFrame = numSrcFrames - numSrcFramesToPrepend - numSrcFramesToAppend;
						TLOG( "WARNING: Source frames " + to_string(actualSrcVideoFrame) + " to "
														+ to_string(actualLastSrcVideoFrame) + " are empty.\n"
							"   This can happen if the source frame rate was detected wrong,\n"
							"   and thus a wrong source frame count was calculated.\n"
							"   Consider using the OverrideSourceFPS tag in the\n"
							"   XML scenario file to specify the correct frame rate.\n"
							"   Inserting black frames to fill the remaining frames." );

						emptyFrameWarningShown = true;
					}
					
					// Insert black frame
					srcFrame.fr = Mat::zeros(capture.getSize(), CV_8UC3);
				}

				// Only enqueue a frame for stage 1 if its number
				// is not lower than the lowest number needed in stage 1.
				// This happens if only a section of the input frames is used
				if(srcFrame.idx >= loweststage1FrameNeeded)
				{
					// If we do not apply temporal distortion, the capture frame index
					// is the same as the source frame index. However, stage 2 which
					// orders the output capture frames requires them to start at zero.
					if(scenario.applyTemporalDistortion == false)
						srcFrame.idx -= loweststage1FrameNeeded;

					// Enqueue it for stage 1
					stage1InputList.Load(srcFrame);
				}
			}
		}

		// When temporal distortion is applied, stage 1 threads cannot remove the
		// frames from the stage 1 output list on their own. Instead, all frames
		// in the stage 1 output list that have a lower frame numbers than the
		// lowest frame that is needed by stage 1 are removed from the list here.
		if(scenario.applyTemporalDistortion && loweststage1FrameNeeded != UINT_MAX)
		{
			// Lock the list mutex and get raw access to the list
			std::list<FrameWithIdx> &list = stage1InputList.LockMutexAndGetList();

			auto it = list.begin();
			while(it != list.end())
			{
				// If the current element is to be deleted, delete it and update iterator
				if(it->idx < loweststage1FrameNeeded) it = list.erase(it);
				else ++it; // Otherwise just increment the iterator
			}

			// Unlock the list mutex
			stage1InputList.UnlockMutex();
		}

		// If either we ran out of source frames, or there is no
		// further cam frame blending information, we have to finish
		// the existing processing and shutdown the processing pipeline.
		if(noMoreSrcFrames || noMoreCamFrames)
		{
			// Since the reader thread is without work now,
			// let it call the shutdown procedure, which is
			// blocking until all other threads are finished.
			finishProcessingThreads();

			// The reader thread is the last thread that is still
			// running, so break out of the while-loop and exit as well.
			break;
		}
	}
}

void CamMark::finishProcessingThreads()
{
	// Finish stage 1 processing
	TLOG( "Waiting for stage 1 threads to finish processing..." );
	while(!stage1ThreadsFinished())
		boost::this_thread::sleep_for(boost::chrono::milliseconds(100));
	
	// Shut down stage 1 threads
	TLOG( "Shutting down stage 1 threads..." );
	stage1ThreadGroup.interrupt_all();
	stage1ThreadGroup.join_all();

	// Finish stage 2 processing
	TLOG( "Waiting for stage 2 thread to finish processing..." );
	while(!stage2InputPrioQueue.isEmpty())
		boost::this_thread::sleep_for(boost::chrono::milliseconds(100));

	// Shut down stage 2 thread
	TLOG( "Shutting down stage 2 thread..." );
	stage2Thread.interrupt();
	stage2Thread.join();
	
	// Finish writer
	TLOG( "Waiting for writer thread to finish..." );
	while(!writerInputQueue.isEmpty())
		boost::this_thread::sleep_for(boost::chrono::milliseconds(100));

	// Shut down writer thread
	TLOG( "Shutting down writer thread..." );
	writerThread.interrupt();
	writerThread.join();

	// Log success and set the finished-flag
	TLOG( "Processing finished." );
	finishedProcessing = true;
}

bool CamMark::stage1ThreadsFinished()
{
	if(scenario.applyTemporalDistortion == false)
	{
		// If no temporal processing is applied, frames are drawn from the
		// stage 1 input list sequencially, so we are done once it is empty.
		return stage1InputList.isEmpty();
	}

	// Don't forget to test if stage 1 threads already finished due to an
	// end of frame blending information. In this case, stage 1 threads
	// exited already, and the following conditions will never be met.
	if(noMoreCamFrames) return true;

	// Lock the stage 1 input list mutex
	std::list<FrameWithIdx> &list = stage1InputList.LockMutexAndGetList();

	// Find the highest available frame number in the stage 1 input list. Actually,
	// this value shouldn't change since stage 1 should be done by the time of this call.
	unsigned int highestAvailFrame = 0;
	for(const FrameWithIdx &curr : list)
		if(curr.idx > highestAvailFrame)
			highestAvailFrame = curr.idx;
	
	// Unlock mutex
	stage1InputList.UnlockMutex();

	// Lock the stage 1 neede frames mutex
	stage1NeededFramesMutex.lock();

	// Find the lowest of all highest frames
	// needed by any of the stage 1 thread
	unsigned int lowestOfAllHighestNeeded = UINT_MAX;
	for(const auto &thread_range : stage1NeededFrames)
		if(thread_range.highest < lowestOfAllHighestNeeded)
			lowestOfAllHighestNeeded = thread_range.highest;

	// Release mutex
	stage1NeededFramesMutex.unlock();

	// If the lowest of all highest needed frames
	// of stage 1 is higher than the highest available
	// frame from the stage 1 input list, stage 1
	// cannot generate any new frames and is done.
	bool stage1Done = lowestOfAllHighestNeeded > highestAvailFrame;
	
	return stage1Done;
}

void CamMark::stage1Processing()
{
	static atomic<unsigned int> thId(0);
	unsigned int this_thread_id = thId++;
	ThreadCout::tcout.setThreadName(
		"stage1:" + to_string(this_thread_id));

	TLOG("Thread started");

	// Let all threads exit the while-loop
	// as soon as one thread sets this flag
	while(noMoreCamFrames == false)
	{
		// While stage 2 input queue is full, sleep for some time
		while(stage2InputPrioQueue.getSize() > maxOrderingQueueSize)
			boost::this_thread::sleep_for(boost::chrono::milliseconds(100));

		FrameWithIdx camFrame;

		if( scenario.applyTemporalDistortion )
		{
			// Guaranteed to be atomic
			camFrame.idx = nextCamFrame++;

			// Get blend data for the new frame
			Simulation_Temporal::BlendDescription blendDesc =
				temporal.getBlendDescForFrame(camFrame.idx,
					scenario.spatial.camera_resolution.height);

			// An empty blend description is considered the no-more-camera-frames
			// marker, so set a flag for the reader and exit this thread.
			if(blendDesc.size() == 0)
			{
				TLOG( "Reached capture end time." );
				
				// Set flag and break out of while-loop
				noMoreCamFrames = true;
				break;
			}

			// Get the frame number of all involed frames
			set<unsigned int> involved_frames =
				Simulation_Temporal::getInvoledFrameNumbers(blendDesc);

			// Validate the involved frames and remove any
			// invalid ones. This is important so we don't
			// wait for a frame that will never be delivered.
			if(!validateFramesToBlend(involved_frames))
			{
				// If validation failed, insert an empty frame
				stage2InputPrioQueue.Enqueue( camFrame );
				continue;
			}

			// Get this blend's lowest/highest needed frames numbers
			unsigned int lowestFrameNeeded = *involved_frames.begin();
			unsigned int highestFrameNeeded = *involved_frames.rbegin();
			
			// Report this thread's lowest/highest needed frames numbers to
			// the reader thread. Based on this information, it schedules
			// new source frames and removes frames that are no longer used.
			stage1NeededFramesMutex.lock();
			stage1NeededFrames[this_thread_id].lowest = lowestFrameNeeded;
			stage1NeededFrames[this_thread_id].highest = highestFrameNeeded;
			stage1NeededFramesMutex.unlock();

			// Assemble all frames to blend. This method
			// blocks until all involed frames are available.
			map<unsigned int, Mat> srcFramesToBlend =
				assembleFramesToBlend(involved_frames);
			
			// Should always be true since assembleFramesToBlend doesn't return otherwise
			CV_DbgAssert(srcFramesToBlend.size() == involved_frames.size());
			
			// If we came here, blend data is valid and all frames to blend
			// are assembled, so blend the source frames into one cam frame.
			camFrame.fr = temporal.blendFrames(blendDesc, srcFramesToBlend);
		}
		else // No temporal distortion
		{
			// Get the next avaliable frame
			camFrame = stage1InputList.GetAndRemoveFront();
		}

		// Apply spatial distortion?
		if(scenario.applySpatialDistortion)
		{
			double time = camFrame.idx / scenario.temporal.cam_FPS;
			camFrame.fr = spatial.distort(camFrame.fr, time);
		}
		
		// Apply defocus?
		if(scenario.applyDefocus)
		{
			// Gaussian blur method
			if(scenario.defocus.method == Scenario::DEFOCUS_GAUSS)
			{
				GaussianBlur(camFrame.fr, camFrame.fr,
					Size(scenario.defocus.kernel_size, scenario.defocus.kernel_size),
					scenario.defocus.sigma, scenario.defocus.sigma);
			}
			else throw runtime_error("Simulation: unknown defocus method specified.");
		}
		
		// Apply CFA?
		if(scenario.applyCFADistortion)
		{
			// Apply optional AA filter?
			if(scenario.cfa.aafilter)
				Simulation_BayerCFA::applyAAFilter(camFrame.fr);
			
			// Simple bayer method
			if(scenario.cfa.method == Scenario::CFA_SIMPLE_BAYER)
				Simulation_BayerCFA::applySimpleBayerCFA(camFrame.fr);
			else
				throw runtime_error("Simulation: unknown CFA method specified.");
		}

		// Apply automatic white balance (AWB)
		if(scenario.applyAWBDistortion)
		{
			Mat awbTemp ;
			switch( scenario.awb.method )
			{
			case Scenario::AWB_GRAYWORLD: Simulation_AWB::grayWorld( camFrame.fr, awbTemp ); break;
			case Scenario::AWB_WENG: Simulation_AWB::weng( camFrame.fr, awbTemp ); break;
			}
			camFrame.fr = awbTemp;
		}
		
		// Insert the result frame into the input map of stage 2
		stage2InputPrioQueue.Enqueue( camFrame );
	}

	// This thread is done
	TLOG("Thread exiting");
}

bool CamMark::validateFramesToBlend(
	set<unsigned int> &involved_frames)
{
	/*****************************************
	 * Check if there are any involved frames
	 *****************************************/

	if(involved_frames.empty())
	{
		TLOG("WARNING: Empty frame blending data.");
		return false;
	}

	/********************************************************
	 * Find the lowest frame number that is currently in use
	 ********************************************************/

	// Lock mutex
	stage1NeededFramesMutex.lock();

	unsigned int lowestFrameInUse = UINT_MAX;
	for(const auto &thread_range : stage1NeededFrames)
		if(thread_range.lowest < lowestFrameInUse)
			lowestFrameInUse = thread_range.lowest;

	// Release mutex
	stage1NeededFramesMutex.unlock();

	/********************************************
	 * If there is a lowest frame, verify that
	   none of the new involved frames in lower.
	 ********************************************/
	
	// Get lowest frame involved in this blend
	unsigned int lowestInvolvedFrame = *involved_frames.begin();
	
	if( lowestFrameInUse < UINT_MAX &&
		lowestInvolvedFrame < lowestFrameInUse )
	{
		// Log a warning message
		string str_involed = Util::enumerate_as_string(involved_frames);
		TLOG( "WARNING: Frames to be blended reference past frames.\n"
				"   The involved frames are: " << str_involed << ".\n"
				"   The last frame not in the past is: " << lowestFrameInUse << ".\n"
				"   Excluding invalid frames from blending.");
			
		// Remove invalid frames
		for(const unsigned int idx : involved_frames)
			if(idx < lowestFrameInUse)
				involved_frames.erase(idx);

		// Log which frames remained
		str_involed = Util::enumerate_as_string(involved_frames);
		TLOG( "CLEANED involved frames are: " << str_involed );
	}

	// If no frame is left, return false
	if(involved_frames.empty()) return false;
	
	/*********************************************************
	 * Check if the given range of blending frames is allowed
	 *********************************************************/

	// Update lowest frame involved in this blend and get highest
	lowestInvolvedFrame = *involved_frames.begin();
	unsigned int highestInvolvedFrame = *involved_frames.rbegin();

	if(highestInvolvedFrame - lowestInvolvedFrame > maxAllowedBlendRange)
	{
		// Log a warning message
		string str_involed = Util::enumerate_as_string(involved_frames);
		TLOG( "WARNING: Frames to be blended exceed allowed range.\n"
				"   The involved frames are: " << str_involed << ".\n"
				"   Max. allowed blend range is " << maxAllowedBlendRange << " frames.\n"
				"   Excluding invalid frames from blending.");

		// Calculate the highest allowed frame for this blend
		unsigned int highestAllowed = lowestInvolvedFrame + maxAllowedBlendRange;

		// Remove invalid frames
		for(const unsigned int idx : involved_frames)
			if(idx > highestAllowed)
				involved_frames.erase(idx);

		// Log which frames remained
		str_involed = Util::enumerate_as_string(involved_frames);
		TLOG( "CLEANED involved frames are: " << str_involed );
	}

	// Return true only if there are frames left
	return !involved_frames.empty();
}

map<unsigned int, Mat> CamMark::assembleFramesToBlend(
	const set<unsigned int> &involved_frames)
{
	// This will contain our result
	map<unsigned int, Mat> framesMap;

	// Create a copy of the set of involved frames.
	// We will continously remove frames we already found.
	set<unsigned int> frames_needed(involved_frames);

	// Gather all source frames to be blended
	while(true)
	{
		// Lock the list mutex
		std::list<FrameWithIdx> &list = stage1InputList.LockMutexAndGetList();

		// Iterate through the input list
		for(const FrameWithIdx &curr : list)
		{
			// Check if the current frame is needed. If it is...
			if( frames_needed.find(curr.idx) != frames_needed.end())
			{
				// ...put it into the map and...
				framesMap[curr.idx] = curr.fr;

				// ...remove it from the set of needed frames.
				frames_needed.erase(curr.idx);
			}
		}

		// Unlock the list mutex
		stage1InputList.UnlockMutex();

		// If we found all needed frames, break and return
		if(frames_needed.empty()) break;

		// Wait until new frames arrive before we try 
		// again in the next iteration of the while-loop.
		stage1InputList.TimedWaitForNewData(1000);
	}

	return framesMap;
}

void CamMark::stage2Processing()
{
	// This stage orders the incoming frames and
	// applies operations that need to be done in order,
	// before frames are passed to the video writer.
	ThreadCout::tcout.setThreadName("stage2");
	TLOG("Thread started");

	// AGC calculations need to be applied in order since the
	// result for frame i always depends on frame i+1.
	Simulation_AGC agc;
	if(scenario.applyAGCDistortion)
	{
		agc.setup( // Setup AGC
			scenario.agc.method,
			scenario.agc.frameCountLower, scenario.agc.frameCountUpper,
			scenario.agc.threshLower, scenario.agc.threshUpper,
			scenario.agc.adjustLower, scenario.agc.adjustUpper);
	}

	// The first frame that should arrive is 0
	unsigned int next_frame_in_order = 0;

	while(true)
	{
		// While writer input queue is full, sleep for some time
		while(writerInputQueue.getSize() > maxQueueSize)
			boost::this_thread::sleep_for(boost::chrono::milliseconds(100));

		// Wait for new data on the input map
		stage2InputPrioQueue.TimedWaitForNewData(100);

		/**********************************************
		 * Gather all frames that are in-order already
		 **********************************************/

		// Once a previously missing frame arrives, it might result in
		// a long sequence of frames in order that are then available.
		// To not block the map's mutex, we first collect all new
		// in-order frames into a vector and afterwards process it.
		vector<Mat> new_frames_in_order;

		// Lock the mutex and get direct access to the map
		auto &prio_queue = stage2InputPrioQueue.LockMutexAndGetAccess();

		// (Repeatedly) test if the lowest frame from
		// the priority queue is the next frame in order.
		while(!prio_queue.empty())
		{
			const FrameWithIdx &lowest = prio_queue.top();

			// If bug-free, this is always true.
			CV_DbgAssert(lowest.idx >= next_frame_in_order);

			// If the lowest frame is next in order...
			if(lowest.idx == next_frame_in_order)
			{
				// Skip empty frames
				if(lowest.fr.empty())
				{
					// Log a warning
					TLOG("WARNING: Skipping output frame " << lowest.idx << " since it is empty.");
				}
				else
				{
					// Store frame in our temporary vector
					new_frames_in_order.push_back(lowest.fr);
				}
				
				// Remove the lowest element from the map
				prio_queue.pop();

				// Determine next frame in order and continue
				next_frame_in_order++;
				continue;
			}
			
			break;
		}

		// Unlock the mutex
		stage2InputPrioQueue.UnlockMutex();

		/*******************************************
		 * Do processing on frames in correct order
		 *******************************************/
		for(Mat &currFrame : new_frames_in_order)
		{
			// Apply automatic gain control (AGC)
			if(scenario.applyAGCDistortion)
				agc.processFrame(currFrame);

			// Enqueue it for writing
			writerInputQueue.Enqueue(currFrame);
		}
	}
}

void CamMark::videoWriter()
{
	// Pulls a frame from the queue and writes it to a video in order
	ThreadCout::tcout.setThreadName("writer");
	TLOG("Thread started");

	unsigned int numWrittenFrames = 0;
	
	while(true)
	{
		// Pull a frame from the queue
		const Mat &frame = writerInputQueue.Dequeue() ;

		// Write frame
		writer << frame;
			
		// Increment the written frame counter and calculate progress
		numWrittenFrames++;
		progress = (float)numWrittenFrames / numCamFrames;
	}
}
