/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#ifndef __SIMULATION__
#define __SIMULATION__

class CamMark
{
public:
	CamMark(void);
	~CamMark(void);

	// Processing
	void startSimulation(std::string src_file,
						 std::string out_file,
						 const Scenario &scenarioToSimulate);

	void finishSimulation() ;

	float getProgress() ;
	bool doneProcessing() ;

	void cancel();
	void terminate();

	/**
	* Rethrows pending exceptions.
	*/
	void rethrowIfPending();
 
	/**
	* Returns true iff an exception is pending.
	*/
	bool isExceptionPending();

	inline bool hasPendingLogMessages()
	{	return ThreadCout::tcout.getPendingMsgCount() > 0;	}

	inline void showLogMessages(std::ostream &os = std::cout)
	{	ThreadCout::tcout.render(os, true);	}
	
public:
	struct FrameWithIdx
	{
		FrameWithIdx() { }
		FrameWithIdx(unsigned int idx) : idx(idx) { }

		cv::Mat fr;
		unsigned int idx;

		bool operator> (const FrameWithIdx &rhs) const
		{	return idx > rhs.idx;	}
	};

	struct FrameRange
	{
		FrameRange()
			: lowest(0), highest(0) { }

		unsigned int lowest;
		unsigned int highest;
	};

protected:
	// Thread management
	void startProcessingThreads();
	void threadCallWrapper(boost::function<void(void)> workerMemFunc);
	void finishProcessingThreads();
	bool stage1ThreadsFinished();
	
	// Processing methods
	void prepare();
	void videoReader();
	void stage1Processing();
	void stage2Processing();
	void videoWriter();

	// Processing utility methods
	bool validateFramesToBlend(
		std::set<unsigned int> &involved_frames);

	std::map<unsigned int, cv::Mat> assembleFramesToBlend(
		const std::set<unsigned int> &involved_frames);

protected:
	// Maximum queue sizes
	const unsigned int maxQueueSize;

	// Maximum size for the stage 2 input map
	const unsigned int maxOrderingQueueSize;

	// Currently hard-coded, but can be changed to any value.
	// Note that the higher the value, the more memory is required.
	const unsigned int maxAllowedBlendRange;

protected:
	boost::thread preprocessThread;
	boost::thread readerThread;
	boost::thread_group stage1ThreadGroup;
	boost::thread stage2Thread;
	boost::thread writerThread;
	
	SynchronizedList<FrameWithIdx> stage1InputList;
	SynchronizedPriorityQueue<FrameWithIdx> stage2InputPrioQueue;
	SynchronizedQueue<cv::Mat> writerInputQueue;
	
	boost::mutex stage1NeededFramesMutex;
	std::vector<FrameRange> stage1NeededFrames;
	
	std::atomic<unsigned int> nextSrcFrame;
	std::atomic<unsigned int> nextCamFrame;
	
	std::atomic<bool> noMoreSrcFrames;
	std::atomic<bool> noMoreCamFrames;

	std::atomic<float> progress;

	unsigned int numSrcFrames;
	unsigned int numCamFrames;

	unsigned int numSrcFramesToPrepend;
	unsigned int numSrcFramesToAppend;

	double src_fps;

	bool finishedProcessing;

protected:
	std::string src_file;
	std::string out_file;

	Scenario scenario;
	unsigned int nCodec;

	VideoReader capture;
	cv::VideoWriter writer;

	Simulation_Spatial spatial;
	Simulation_Temporal temporal;

private:
	bool hasCUDASupport;
 
	boost::recursive_mutex exceptionMutex;
	std::queue<std::exception_ptr> exceptionPtrQueue;
};

#endif // __SIMULATION__