/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../stdafx.h"

#include "VideoReader.h"
#include "BufferedVideoReader.h"

using namespace std;
using namespace cv;

BufferedVideoReader::BufferedVideoReader(FrameVector::size_type buffer_size)
	: VideoReader(), nextFrameNr(0), default_buffer(buffer_size)
{
}

BufferedVideoReader::BufferedVideoReader(int device, FrameVector::size_type buffer_size)
	: VideoReader(device), nextFrameNr(0), default_buffer(buffer_size)
{
}

BufferedVideoReader::BufferedVideoReader(const string &filename, FrameVector::size_type buffer_size)
	: VideoReader(filename), nextFrameNr(0), default_buffer(buffer_size)
{
}

BufferedVideoReader::~BufferedVideoReader(void)
{
}

unsigned int BufferedVideoReader::seekToFrame(unsigned int frame_nr)
{
	// Use base class seeking
	unsigned int afterSeek = VideoReader::seekToFrame(frame_nr);

	// Also update the nextFrameNr variable
	nextFrameNr = afterSeek;

	return afterSeek;
}

Mat BufferedVideoReader::getFrame(unsigned int frame_nr)
{
	return getFrame(frame_nr, default_buffer);
}

Mat BufferedVideoReader::getFrame(unsigned int frame_nr, FrameVector &buffer)
{
	// See if the frame is in the buffer
	unsigned int bufferIdx = frame_nr % buffer.size();
	if(buffer[bufferIdx].frame_nr == frame_nr)
		return buffer[bufferIdx].frame;
	
	// If it is in the past and not in the
	// buffer any longer, throw an exception
	else if(frame_nr < nextFrameNr)
	{
		stringstream msg;
		msg << "Error in BufferedVideoReader::getFrame():\n"
			<< "   The requested frame #" << frame_nr << " is not in the buffer any longer,\n"
			<< "   but since the video is sequencial, it cannot be retrieved again.\n"
			<< "   Increase the buffer size to fix this issue.";
		
		throw runtime_error(msg.str());
	}
	
	// Otherwise, if it is upcoming,
	// read new frames until we reach it
	else while(nextFrameNr == 0 || frame_nr > nextFrameNr - 1)
	{
		// If readNextFrame returns false,
		// we reached the end of the video,
		// which is indicated by an empty frame
		if(false == readNextFrame())
			return Mat();
	}

	// These conditions should hold now...
	CV_Assert(buffer[bufferIdx].frame_nr == frame_nr);
	CV_Assert(buffer[bufferIdx].frame.empty() == false);

	// ... so we can return the newly loaded frame
	return buffer[bufferIdx].frame;
}

bool BufferedVideoReader::readNextFrame()
{
	unsigned int bufferIdx = nextFrameNr % default_buffer.size();

	default_buffer[bufferIdx].frame_nr = nextFrameNr;
	read(default_buffer[bufferIdx].frame);
	
	bool eof = default_buffer[bufferIdx].frame.empty();
	nextFrameNr++;

	return !eof;
}