/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

//
// This code is based on an example from: "Chapter 18 - An Introduction to Thread"
// from the book "Introduction to the C++ Boost Libraries - Volume I - Foundations".
// URL: https://www.quantnet.com/threads/c-multithreading-in-boost.10028/
//

#pragma once

#ifndef __SYNCHRONIZED_LIST__
#define __SYNCHRONIZED_LIST__

// Queue class that has thread synchronisation
template <typename T>
class SynchronizedList
{
private:
	
	std::list<T> m_list ; // Use STL list to store data
	boost::mutex m_mutex ; // The mutex to synchronize the list on
	boost::condition_variable m_cond ; // The condition to wait for
	
public:
	// Direct access to the raw list
	std::list<T>& LockMutexAndGetList()
	{
		m_mutex.lock();
		return m_list;
	}
	
	void UnlockMutex()
	{	m_mutex.unlock();	}

	// Add data to the list and notify others
	void Load(const T& data)
	{
		// Acquire lock on the queue
		boost::unique_lock<boost::mutex> lock(m_mutex) ;
		
		// Add the data to the queue
		m_list.push_back(data) ;
		
		// Notify others that data is ready
		m_cond.notify_all() ;
	}
	
	// Get data from the queue. Wait for data if not available. Then remove it.
	T GetAndRemoveFront()
	{
		// Acquire lock on the queue
		boost::unique_lock<boost::mutex> lock(m_mutex) ;
		
		// When there is no data, wait till someone fills it.
		// Lock is automatically released in the wait
		// and obtained again after the wait.
		while (m_list.size() == 0) m_cond.wait(lock) ;
		
		// Retrieve the data from the queue
		T result = m_list.front() ; m_list.pop_front() ;
		return result ;
	}

	// Wait for new data
	void WaitForNewData()
	{
		// Acquire lock on the queue
		boost::unique_lock<boost::mutex> lock(m_mutex) ;
		
		// Wait until someone fills new data.
		// Lock is automatically released in the wait
		// and obtained again after the wait.
		m_cond.wait(lock) ;
	}

	// Wait for new data
	void TimedWaitForNewData(unsigned int timeout_ms)
	{
		// Acquire lock on the queue
		boost::unique_lock<boost::mutex> lock(m_mutex) ;
		
		// Wait until someone fills new data or the timeout expires.
		// Lock is automatically released in the wait
		// and obtained again after the wait.
		m_cond.timed_wait(lock, boost::posix_time::milliseconds(timeout_ms)) ;
	}

	unsigned int getSize()
	{
		boost::unique_lock<boost::mutex> lock(m_mutex) ;
		return (unsigned int)m_list.size() ;
	}
	
	bool isEmpty()
	{
		boost::unique_lock<boost::mutex> lock(m_mutex) ;
		return m_list.empty() ;
	}
};

#endif //__SYNCHRONIZED_LIST__
