/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

//
// This code is based on an example from: "Chapter 18 - An Introduction to Thread"
// from the book "Introduction to the C++ Boost Libraries - Volume I - Foundations".
// URL: https://www.quantnet.com/threads/c-multithreading-in-boost.10028/
//

#pragma once

#ifndef __SYNCHRONIZED_PRIORITY_QUEUE__
#define __SYNCHRONIZED_PRIORITY_QUEUE__

// Priority queue class that has thread synchronization
template <typename T>
class SynchronizedPriorityQueue
{
public:
	// Use STL priority queue to store data, with a vector as underlying container.
	// As comparer, we use greater<T>, to make the lowerst element the highest priority.
	typedef std::priority_queue<T, std::vector<T>, std::greater<T>> QueueType;

private:
	// The protected priority queue
	QueueType m_queue; 
	
	// The mutex to synchronize on
	boost::mutex m_mutex;

	// The condition to wait for new data
	boost::condition_variable m_cond;
	
public:
	
	// Add data to the queue and notify others
	void Enqueue(const T& data)
	{
		// Acquire lock on the queue
		boost::unique_lock<boost::mutex> lock(m_mutex);
		
		// Add the data to the queue
		m_queue.push(data);
		
		// Notify others that data is ready
		m_cond.notify_all();
	}
	
	// Get data from the queue. Wait for data if not available
	T Dequeue()
	{
		// Acquire lock on the queue
		boost::unique_lock<boost::mutex> lock(m_mutex);
		
		// When there is no data, wait till someone fills it.
		// Lock is automatically released in the wait and obtained
		// again after the wait
		while (m_queue.size() == 0) m_cond.wait(lock);
		
		// Retrieve the data from the queue
		T result = m_queue.top(); m_queue.pop();
		return result;
	}

	// Direct access to the priority queue
	// with manual locking of the mutex
	QueueType& LockMutexAndGetAccess()
	{
		m_mutex.lock();
		return m_queue;
	}
	
	// Manual unlocking of the mutex
	void UnlockMutex()
	{	m_mutex.unlock();	}

	// Wait for new data
	bool TimedWaitForNewData(unsigned int timeout_ms)
	{
		// Acquire lock on the queue
		boost::unique_lock<boost::mutex> lock(m_mutex) ;
		
		// Wait until someone fills new data or the timeout expires.
		// Lock is automatically released in the wait and obtained again after the wait.
		return m_cond.timed_wait(lock, boost::posix_time::milliseconds(timeout_ms)) ;
	}

	unsigned int getSize()
	{
		boost::unique_lock<boost::mutex> lock(m_mutex);
		return (unsigned int)m_queue.size();
	}

	bool isEmpty()
	{
		boost::unique_lock<boost::mutex> lock(m_mutex);
		return m_queue.empty();
	}
};

#endif //__SYNCHRONIZED_PRIORITY_QUEUE__
