/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#ifndef __UTIL__
#define __UTIL__

namespace Util {

// Finds the last occurrence of either '/' or '\' in a string
std::string::size_type
	findLastDirectorySeperator(const std::string file);

// Splits the filename from the filepath
void splitFilePathAndName(const std::string file,
	std::string &path, std::string &name);

// Goes up by one line on the console (opposite of \n)
void stdOut_goUpByOneLine();

// Writes a formatted timestap to a std::ostream
void writeDateTime(std::ostream &os);

// ---------------------------------------------
// Generate enumeration as string of container.
// Variants to enumerate first/second member.
// ---------------------------------------------

template<class T>
std::string enumerate_as_string(
	T container, std::string separator = ", ")
{
	auto it = container.begin();
	std::string enum_string = std::to_string(*it);
	for(++it; it != container.end(); ++it)
		enum_string += separator + std::to_string(*it);

	return enum_string;
}

template<class T>
std::string enumerate_first_as_string(
	T container, std::string separator = ", ")
{
	auto it = container.begin();
	std::string enum_string = std::to_string(it->first);
	for(++it; it != container.end(); ++it)
		enum_string += separator + std::to_string(it->first);

	return enum_string;
}

template<class T>
std::string enumerate_second_as_string(
	T container, std::string separator = ", ")
{
	auto it = container.begin();
	std::string enum_string = std::to_string(it->second);
	for(++it; it != container.end(); ++it)
		enum_string += separator + std::to_string(it->second);

	return enum_string;
}

// ---------------------------------------------
// String trimming
// ---------------------------------------------

// trim from start
inline std::string& ltrim(std::string &s)
{
	s.erase(s.begin(), std::find_if(s.begin(), s.end(),
		std::not1(std::ptr_fun<int, int>(std::isspace))));

	return s;
}

// trim from end
inline std::string& rtrim(std::string &s)
{
	s.erase(std::find_if(s.rbegin(), s.rend(),
		std::not1(std::ptr_fun<int, int>(std::isspace))).base(), s.end());

	return s;
}

// trim from both ends
inline std::string& trim(std::string &s)
{
	return ltrim(rtrim(s));
}

// ---------------------------------------------
// Branchless, type-safe C++ signum function. Taken from:
// http://stackoverflow.com/questions/1903954/is-there-a-standard-sign-function-signum-sgn-in-c-c
// ---------------------------------------------

template <typename T> inline //constexpr (if C++11 is supported)
int signum(T x, std::false_type /*is_signed*/) {
	return T(0) < x;
}

template <typename T> inline //constexpr (if C++11 is supported)
int signum(T x, std::true_type /*is_signed*/) {
	return (T(0) < x) - (x < T(0));
}

template <typename T> inline //constexpr (if C++11 is supported)
int signum(T x) {
	return signum(x, std::is_signed<T>());
}

// ---------------------------------------------
// This code is from Guru of the Week #29
// http://www.gotw.ca/gotw/029.htm
// ---------------------------------------------

struct ci_char_traits : public std::char_traits<char> {
	static bool eq(char c1, char c2) { return toupper(c1) == toupper(c2); }
	static bool ne(char c1, char c2) { return toupper(c1) != toupper(c2); }
	static bool lt(char c1, char c2) { return toupper(c1) <  toupper(c2); }
	static int compare(const char* s1, const char* s2, size_t n) {
		while( n-- != 0 ) {
			if( toupper(*s1) < toupper(*s2) ) return -1;
			if( toupper(*s1) > toupper(*s2) ) return 1;
			++s1; ++s2;
		}
		return 0;
	}
	static const char* find(const char* s, int n, char a) {
		while( n-- > 0 && toupper(*s) != toupper(a) ) {
			++s;
		}
		return s;
	}
};

typedef std::basic_string<char, ci_char_traits> ci_string;

// ---------------------------------------------
// A custom iterator that allows to iterate only over the values of a map.
// From: http://john-ahlgren.blogspot.com/2013/10/how-to-iterate-over-values-of-stdmap.html
// ---------------------------------------------

template <typename Iter>
class second_iterator : public std::iterator<std::bidirectional_iterator_tag, typename Iter::value_type::second_type>
{
public:
	second_iterator() {}
	second_iterator(Iter j) : i(j) {}

	second_iterator& operator++() { ++i; return *this; }
	second_iterator operator++(int) { auto tmp = *this; ++(*this); return tmp; }
	second_iterator& operator--() { --i; return *this; }
	second_iterator operator--(int) { auto tmp = *this; --(*this); return tmp; }

	bool operator==(second_iterator j) const { return i == j.i; }
	bool operator!=(second_iterator j) const { return !(*this == j); }

	typename second_iterator::reference operator*() { return i->second; }
	typename second_iterator::pointer operator->() { return &i->second; }

protected:
	Iter i;
};

template <typename Iter>
inline second_iterator<Iter> make_second_iterator(Iter j)
{	return second_iterator<Iter>(j);	}

}

#endif //__UTIL__