/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

//
// This code is based on an implementation found on Stack Overflow.
// URL: http://stackoverflow.com/questions/2941491/compare-versions-as-strings
//

#pragma once

#ifndef __VERSION_UTIL__   // Don't use __VERSION__ as include guard, as
#define __VERSION_UTIL__   // it is already used/defined by some compilers.

class Version
{
	// An internal utility structure just used to make
	// the std::copy in the constructor easy to write.
	struct VersionDigit
	{
		int value;
		operator int() const { return value; }
	};

	friend std::istream& operator>>
		(std::istream& str, Version::VersionDigit& digit);

	friend std::ostream& operator<<
		(std::ostream& stream, const Version& version);

	public:
		Version() // Default constructor
		{	versionInfo.push_back(0);	}

		Version(int major, int minor = 0, int patch = 0)
		{	versionInfo.push_back(major);
			versionInfo.push_back(minor);
			versionInfo.push_back(patch);	}

		// Constructor that parses a version string
		Version(std::string const& versionStr)
		{
			// To make processing easier in VersionDigit, prepend a '.'
			std::stringstream versionStream(std::string(".") + versionStr);

			// Assign all parts of the version number into the version info vector
			versionInfo.assign(std::istream_iterator<VersionDigit>(versionStream),
							   std::istream_iterator<VersionDigit>());
		}

		// Compare two version numbers
		bool operator< (Version const& rhs) const
		{
			std::vector<int> padded_lhs(versionInfo.begin(), versionInfo.end());
			std::vector<int> padded_rhs(rhs.versionInfo.begin(), rhs.versionInfo.end());

			while(padded_lhs.size() < padded_rhs.size()) padded_lhs.push_back(0);
			while(padded_rhs.size() < padded_lhs.size()) padded_rhs.push_back(0);

			return std::lexicographical_compare(
				padded_lhs.begin(), padded_lhs.end(),
				padded_rhs.begin(), padded_rhs.end());
		}

		// Print the version to an ostream
		void print(std::ostream& stream, std::string delim = ".") const
		{
			for (const auto& i : versionInfo)
			{
				if (&i != &versionInfo[0]) stream << delim;
				stream << i;
			}
		}

	private:
		std::vector<int> versionInfo;
};

// Read a single digit from the version
inline std::istream& operator>>(std::istream& stream, Version::VersionDigit& digit)
{
	stream.get();
	stream >> digit.value;
	return stream;
}

// Output a version object to an ostream
inline std::ostream& operator<<(std::ostream& stream, const Version& version)
{
	version.print(stream);
	return stream;
}

#endif //__VERSION_UTIL__