/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../stdafx.h"

#include "../utility/Util.h"

#include "VideoReader.h"

using namespace std;
using namespace cv;

unsigned int VideoReader::MAX_NUM_DIGITS = 12;

VideoReader::VideoReader()
	: fps(0), frameCount(0)
{
}

VideoReader::VideoReader(int device)
	: fps(0), frameCount(0)
{
	// Call open method that also
	// extracts frame size and count
	open(device);
}

VideoReader::VideoReader(const string &filename)
	: fps(0), frameCount(0)
{
	// Call open method that also
	// extracts frame size and count
	open(filename);
}

VideoReader::~VideoReader(void)
{
}

bool VideoReader::open(int device)
{
	if(!cv_capture.open(device))
		return false;

	// Retrieve and store some properties from the capture object
	size.width =  (int)cv_capture.get(CV_CAP_PROP_FRAME_WIDTH);
	size.height = (int)cv_capture.get(CV_CAP_PROP_FRAME_HEIGHT);
	frameCount =  (unsigned int)cv_capture.get(CV_CAP_PROP_FRAME_COUNT);

	return true;
}

bool VideoReader::open(const string &filename)
{
	// FIRST, try to open the media as a sequence of single frames
	string path, name;
	Util::splitFilePathAndName(filename, path, name);

	// Try if this is an image sequence with a variable length padding
	for(unsigned int numDigits = MAX_NUM_DIGITS; numDigits > 0; numDigits--)
	{
		// Create a padding of numDigits zeros
		string padding(numDigits, '0');
		
		// Check if the filename contains this padding
		if(name.find(padding) != std::string::npos)
		{
			// Create the according pattern
			stringstream pattern; pattern << "%";
			pattern << setw(2) << setfill('0') << numDigits << "d";

			// Replace the padding with the pattern in the filename
			string newName = name;
			replaceSubstring(newName, padding, pattern.str());
		
			// Try to open the media as an image sequence
			if(cv_capture.open(path + newName)) break;
		}
	}

	// If opening the media as image sequence was successful, set
	// a temporary flag to later distinguish it from a single image.
	bool isImageSequence = cv_capture.isOpened();

	// SECOND, if we could not open the media as an image sequence,
	// try to directly open the media, which works for video and image files.
	if(isImageSequence == false) cv_capture.open(filename);

	// We the media is now opened, retrieve some properties
	if(cv_capture.isOpened()) 
	{
		// Since the OpenCV capture object does not provide the size of image-based media,
		// we test if the source is image-based and read the size from a Mat object instead.
		Mat image = imread(filename);
		bool isImageBased = !image.empty();

		if(isImageBased)
		{
			// Use imread to open the media as an image,
			// and read the size from the Mat object.
			size = image.size();
		}
		else
		{
			// Otherwise, read the width/height from the capture object
			size.width =  (int)cv_capture.get(CV_CAP_PROP_FRAME_WIDTH);
			size.height = (int)cv_capture.get(CV_CAP_PROP_FRAME_HEIGHT);
		}

		// Get the frame count from the capture object,
		// unless it is a single image -- then just set to 1.
		if(isImageBased && !isImageSequence) frameCount = 1;
		else frameCount = (unsigned int)cv_capture.get(CV_CAP_PROP_FRAME_COUNT); 

		// Success
		return true;
	}
	// ------------------------------------------------------------------------
	// NOTE: At least for OpenCV 2.4.10, there is a BUG with PNG files:
	// For some reason, the OpenCV capture object opens a single PNG file
	// ONLY IF IT'S FILE NAME CONTAINS AT LEAST ONE NUMBER.
	// For JPEG files, everything works fine.
	//
	// To detect this bug, we try to open the media as an image.
	// If this worked although opening it through a cature object failed,
	// we throw an exception giving valuable information about this.
	// ------------------------------------------------------------------------
	else if(!imread(filename).empty()) throw runtime_error(
		"VideoReader: Although the specified media file '" + filename
		+ "' is a valid image file, OpenCV cannot open it using the CvCapture object "
		+ "which is used to conveniently read videos, image sequences, or single images "
		+ "with the same interface. For PNG images, there is a known bug OpenCV that "
		+ "requires the file name of the PNG file to contain at least one number. "
		+ "Please try again with adding a number to any part of the filename, "
		+ "or use another image file type (e.g., JPEG).");

	// Media could not be opened
	return false;
}

void VideoReader::close()
{
	cv_capture.release();
}

bool VideoReader::isOpened()
{
	return cv_capture.isOpened();
}

double VideoReader::get(int propId)
{
	switch(propId)
	{
		// Special cases
		case CV_CAP_PROP_FPS:			return getFPS();
		case CV_CAP_PROP_FRAME_WIDTH:	return size.width;
		case CV_CAP_PROP_FRAME_HEIGHT:	return size.height;
		case CV_CAP_PROP_FRAME_COUNT:	return frameCount;

		// General case
		default: return cv_capture.get(propId);
	}
}

bool VideoReader::set(int propId, double value)
{
	switch(propId)
	{
		// Special cases
		case CV_CAP_PROP_POS_FRAMES:
		{
			unsigned int frame_nr = (unsigned int)value;
			return seekToFrame(frame_nr) == frame_nr;
		}
		case CV_CAP_PROP_FPS:
			return setFPS(value) == value;
		
		// These properties cannot be changed
		case CV_CAP_PROP_FRAME_WIDTH:
		case CV_CAP_PROP_FRAME_HEIGHT:
		case CV_CAP_PROP_FRAME_COUNT:
			return false;

		// General case
		default: return cv_capture.set(propId, value);
	}
}

double VideoReader::getFPS()
{
	if(fps == 0 && isOpened())
		fps = cv_capture.get(CV_CAP_PROP_FPS);

	return fps;
}

double VideoReader::setFPS(double value)
{
	fps = value;
	
	if(isOpened())
	{
		cv_capture.set(CV_CAP_PROP_FPS, fps);

		double new_fps = cv_capture.get(CV_CAP_PROP_FPS);
		if(new_fps != fps)
			cerr << "WARNING: VideoReader: Setting the video's FPS was not successful." << endl;

		fps = new_fps;
	}

	return fps;
}

unsigned int VideoReader::seekToFrame(unsigned int frame_nr)
{
	// Seek by setting the according property
	cv_capture.set(CV_CAP_PROP_POS_FRAMES, frame_nr);

	// Check success by retrieving the current position
	unsigned int afterSeek = static_cast<unsigned int>(cv_capture.get(CV_CAP_PROP_POS_FRAMES));
	if(afterSeek != frame_nr)
		cerr << "WARNING: VideoReader: Seeking the video was not successful." << endl;

	return afterSeek;
}

void VideoReader::replaceSubstring(string& str, const string& oldStr, const string& newStr)
{
	size_t pos = 0;
	while((pos = str.find(oldStr, pos)) != std::string::npos)
	{
		str.replace(pos, oldStr.length(), newStr);
		pos += newStr.length();
	}
}
