/*
	Copyright (C) 2015  University of Mannheim

	This file is part of CamMark.

	CamMark is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	CamMark is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CamMark.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#ifndef __SORTEDINTERVALVECTOR__
#define __SORTEDINTERVALVECTOR__

//#include <math.h> // for log2
#include "Util.h"

template <class value_type, class key_type>
class sorted_interval_vector
{
public:
	struct RightOpenInterval
	{
		RightOpenInterval(key_type _lower, key_type _upper)
			: lower(_lower), upper(_upper) { }

		inline bool cmp_lower(const key_type &val) const {	return lower <= val;	}
		inline bool cmp_upper(const key_type &val) const {	return val < upper;		}

		inline bool contains(const key_type &val) const  {	return cmp_lower(val) &&
															   cmp_upper(val);	}
		key_type lower;
		key_type upper;
	};

public:
	// Data typedefs
	typedef std::pair<const RightOpenInterval, value_type> item_type;
	typedef std::vector<value_type> value_vector;
	typedef std::vector<item_type> item_vector;

	// Iterator typedefs
	typedef typename item_vector::iterator item_iterator;
	typedef typename item_vector::const_iterator item_const_iterator;
	
	typedef Util::second_iterator<item_iterator> iterator;
	typedef Util::second_iterator<item_const_iterator> const_iterator;

public:
	struct SeqAccessCache
	{
		friend class sorted_interval_vector<value_type, key_type>;
	
		SeqAccessCache()
			: initialized(false) { }

	private:
		bool initialized;
		item_const_iterator range_first;
		item_const_iterator range_last;
	};

public:
	// Construction
	sorted_interval_vector(
		size_t seq2bin_search_threshold = 25);

	// Insertion
	void insert(key_type lower, key_type upper,
				const value_type &value);

	// This data structure only works on SORTED elements
	void sort();

	// Check if the elements are sorted BOTH based
	// on the interval start as well as end values
	bool validate();
	
	// Shortcut for a call to sort() with subsequent validate()
	bool sortAndValidate()	{	sort(); return validate();	}

	// Returns the number of elements
	size_t size() const	{	return items.size();	}

	// Access through iterators that iterate over the actual values only
	iterator begin()		{	return Util::make_second_iterator(items.begin());	}
	iterator end()			{	return Util::make_second_iterator(items.end());		}
	const_iterator cbegin()	{	return Util::make_second_iterator(items.cbegin());	}
	const_iterator cend()	{	return Util::make_second_iterator(items.cend());	}
	
	// Operators
	value_type operator[] (size_t idx) const
	{	return items[idx].second;	}

	// Accesses
	value_vector get(key_type val) const;
	value_vector get_seq(key_type val, SeqAccessCache &cache) const;

	// Range access
	value_vector get_range(key_type lower, key_type upper) const;
	value_vector get_range_seq(key_type lower, key_type upper, SeqAccessCache &cache) const;
	
protected:
	// Sequencial approaches
	item_const_iterator getFirstUpperBoundGreaterThan_Sequencial(
		const key_type &val, item_const_iterator search_hint,
		size_t num_steps, size_t &steps_done) const;
	item_const_iterator getLastLowerBoundLowerThan_Sequencial(
		const key_type &val, item_const_iterator search_hint,
		size_t num_steps, size_t &steps_done) const;

	// Binary search approaches
	item_const_iterator getFirstUpperBoundGreaterThan_BinSearch(
		const key_type &val) const;
	item_const_iterator getLastLowerBoundLowerThan_BinSearch(
		const key_type &val) const;

	// Combined search approaches
	item_const_iterator getFirstUpperBoundGreaterThan_SequencialThenBinSearch(
		const key_type &val, item_const_iterator search_hint) const;
	item_const_iterator getLastLowerBoundLowerThan_SequencialThenBinSearch(
		const key_type &val, item_const_iterator search_hint) const;

protected:
	// The data
	item_vector items;

	// Theshold on the number of sequencial search steps
	// that are performed before a binary search is started
	// on the end of an interval.
	size_t seq2bin_search_threshold;

private:
	// An empty result vector utility object
	static value_vector emptyResult;
};

// Static member definition
template <class value_type, class key_type>
typename sorted_interval_vector<value_type, key_type>::value_vector
sorted_interval_vector<value_type, key_type>::emptyResult;

template <class value_type, class key_type>
sorted_interval_vector<value_type, key_type>::sorted_interval_vector(
	size_t _seq2bin_search_threshold)
	: seq2bin_search_threshold(_seq2bin_search_threshold)
{
}

template <class value_type, class key_type>
void
sorted_interval_vector<value_type, key_type>::insert(
	key_type lower, key_type upper, const value_type &value)
{
	items.push_back(std::make_pair(
		RightOpenInterval(lower, upper), value));
}

template <class value_type, class key_type>
void
sorted_interval_vector<value_type, key_type>::sort()
{
	// Sort based on interval START value
	std::sort(items.begin(), items.end(),
		[](const item_type &a, const item_type &b) -> bool
		{	return a.first.lower > b.first.lower; });
}

template <class value_type, class key_type>
bool
sorted_interval_vector<value_type, key_type>::validate()
{
	// Make sure the items are BOTH sorted based
	// on the interval START as well as  END values
	auto it = items.begin();
	while(it < items.end() - 1)
	{
		const item_type &curr = *it; ++it;
		const item_type &next = *it;

		if( curr.first.upper > next.first.upper ||
			curr.first.lower > next.first.lower )
			return false;
	}

	// Calculate an optimal seq2bin_search threshold,
	// being the log2-value of the number of elements.
	// This way, the worst-case runtime for a lookup is 2*log2(N).
	seq2bin_search_threshold = (size_t)(log(items.size()) / log(2)); // log(n)/log(2) is log2.

	return true;
}

template <class value_type, class key_type>
typename sorted_interval_vector<value_type, key_type>::value_vector
sorted_interval_vector<value_type, key_type>::get(
	key_type val) const
{
	return get_range(val, val);
}

template <class value_type, class key_type>
typename sorted_interval_vector<value_type, key_type>::value_vector
sorted_interval_vector<value_type, key_type>::get_seq(
	key_type val, SeqAccessCache &cache) const
{
	return get_range_seq(val, val, cache);
}

template <class value_type, class key_type>
typename sorted_interval_vector<value_type, key_type>::value_vector
sorted_interval_vector<value_type, key_type>::get_range(
	key_type lower, key_type upper) const
{
	// Trivial check (the following
	// code assume items not to be empty)
	if(items.empty()) return emptyResult;

	item_const_iterator range_first = getFirstUpperBoundGreaterThan_BinSearch(lower);
	if(range_first == items.end()) return emptyResult;

	item_const_iterator range_last = getLastLowerBoundLowerThan_SequencialThenBinSearch(upper, range_first);
	if(range_last == items.end()) return emptyResult;

	value_vector result;
	for(auto it = range_first; it <= range_last; ++it)
		result.push_back(it->second);
	
	return result;
}

template <class value_type, class key_type>
typename sorted_interval_vector<value_type, key_type>::value_vector
sorted_interval_vector<value_type, key_type>::get_range_seq(
	key_type lower, key_type upper, SeqAccessCache &cache) const
{
	// Trivial check (the following
	// code assume items not to be empty)
	if(items.empty()) return emptyResult;

	// If not done yet, initialize cache
	if(!cache.initialized)
	{
		cache.range_first = items.begin();
		cache.range_last = items.begin();
	}

	item_const_iterator range_first = getFirstUpperBoundGreaterThan_SequencialThenBinSearch(lower, cache.range_first);
	if(range_first == items.end()) return emptyResult;
	else cache.range_first = range_first;

	item_const_iterator range_last = getLastLowerBoundLowerThan_SequencialThenBinSearch(upper, cache.range_last);
	if(range_last == items.end()) return emptyResult;
	else cache.range_last = range_last;

	value_vector result;
	for(auto it = range_first; it <= range_last; ++it)
		result.push_back(it->second);

	return result;
}

template <class value_type, class key_type>
typename sorted_interval_vector<value_type, key_type>::item_const_iterator
sorted_interval_vector<value_type, key_type>::getFirstUpperBoundGreaterThan_Sequencial(
	const key_type &val, item_const_iterator search_hint,
	size_t num_steps, size_t &steps_done) const
{
	item_const_iterator curr_pos = search_hint;

	// Upper bound is greater than val for our current position,
	// so we have to iterate backwards to find the first
	if(curr_pos->first.cmp_upper(val) == true)
	{
		steps_done = 0;
		while(steps_done++ < num_steps)
		{
			if(curr_pos == items.begin()) return curr_pos;
			item_const_iterator next_pos = curr_pos - 1;
		
			if(next_pos->first.cmp_upper(val) == false)
				return curr_pos;
		
			curr_pos = next_pos;
		}
	}
	// Upper bound is lower than val for our current position,
	// so we have to iterate forwards to find the first
	else
	{
		steps_done = 0;
		while(steps_done++ < num_steps)
		{
			item_const_iterator next_pos = curr_pos + 1;
			if(next_pos == items.end()) return items.end();
			
			if(next_pos->first.cmp_upper(val) == true)
				return next_pos;
		
			curr_pos = next_pos;
		}
	}

	return items.end();
}

template <class value_type, class key_type>
typename sorted_interval_vector<value_type, key_type>::item_const_iterator
sorted_interval_vector<value_type, key_type>::getLastLowerBoundLowerThan_Sequencial(
	const key_type &val, item_const_iterator search_hint,
	size_t num_steps, size_t &steps_done) const
{
	item_const_iterator curr_pos = search_hint;

	// Lower bound is lower than val for our current position,
	// so we have to iterate forwards to find the first
	if(curr_pos->first.cmp_lower(val) == true)
	{
		steps_done = 0;
		while(steps_done++ < num_steps)
		{
			item_const_iterator next_pos = curr_pos + 1;
			if(next_pos == items.end()) return curr_pos;
		
			if(next_pos->first.cmp_lower(val) == false)
				return curr_pos;
		
			curr_pos = next_pos;
		}
	}
	// Lower bound is greater than val for our current position,
	// so we have to iterate backwards to find the first
	else
	{
		steps_done = 0;
		while(steps_done++ < num_steps)
		{
			if(curr_pos == items.begin()) return items.end();
			item_const_iterator next_pos = curr_pos - 1;
			
			if(next_pos->first.cmp_lower(val) == true)
				return next_pos;
		
			curr_pos = next_pos;
		}
	}

	return items.end();
}

template <class value_type, class key_type>
typename sorted_interval_vector<value_type, key_type>::item_const_iterator
sorted_interval_vector<value_type, key_type>::getFirstUpperBoundGreaterThan_BinSearch(
	const key_type &val) const
{
	item_const_iterator first_elem = items.begin();
	item_const_iterator last_elem = items.end() - 1;

	// Is the first element already what we are looking for?
	if(first_elem->first.cmp_upper(val) == true)
		return first_elem;

	item_const_iterator left = first_elem + 1;
	item_const_iterator right = last_elem;
	
	while(left <= right)
	{
		item_const_iterator middle = left + ((right - left) / 2);
		item_const_iterator left_of_middle = middle - 1;

		if(middle->first.cmp_upper(val) == true)
		{
			if(left_of_middle->first.cmp_upper(val) == false)
				return middle;

			right = middle - 1;
		}
		else
		{
			left = middle + 1;
		}
	}

	return items.end();
}

template <class value_type, class key_type>
typename sorted_interval_vector<value_type, key_type>::item_const_iterator
sorted_interval_vector<value_type, key_type>::getLastLowerBoundLowerThan_BinSearch(
	const key_type &val) const
{
	item_const_iterator first_elem = items.begin();
	item_const_iterator last_elem = items.end() - 1;

	// Is the last element already what we are looking for?
	if(last_elem->first.cmp_lower(val) == true)
		return last_elem;

	item_const_iterator left = first_elem;
	item_const_iterator right = last_elem - 1;
	
	while(left <= right)
	{
		item_const_iterator middle = left + ((right - left) / 2);
		item_const_iterator right_of_middle = middle + 1;

		if(middle->first.cmp_lower(val) == true)
		{
			if(right_of_middle->first.cmp_lower(val) == false)
				return middle;

			left = middle + 1;
		}
		else
		{
			right = middle - 1;
		}
	}

	return items.end();
}

template <class value_type, class key_type>
typename sorted_interval_vector<value_type, key_type>::item_const_iterator
sorted_interval_vector<value_type, key_type>::getFirstUpperBoundGreaterThan_SequencialThenBinSearch(
	const key_type &val, item_const_iterator search_hint) const
{
	// Try a N steps of a linear serach, starting at search_hint
	size_t steps_done;
	item_const_iterator found =
		getFirstUpperBoundGreaterThan_Sequencial(val, search_hint,
			seq2bin_search_threshold, steps_done);

	// If this did not yield a result...
	if(found == items.end() && steps_done > seq2bin_search_threshold)
	{
		//... start a binary search
		found = getFirstUpperBoundGreaterThan_BinSearch(val);
	}
	
	// Return the result
	return found;
}

template <class value_type, class key_type>
typename sorted_interval_vector<value_type, key_type>::item_const_iterator
sorted_interval_vector<value_type, key_type>::getLastLowerBoundLowerThan_SequencialThenBinSearch(
	const key_type &val, item_const_iterator search_hint) const
{
	// Try a N steps of a linear serach, starting at search_hint
	size_t steps_done;
	item_const_iterator found =
		getLastLowerBoundLowerThan_Sequencial(val, search_hint,
			seq2bin_search_threshold, steps_done);

	// If this did not yield a result...
	if(found == items.end() && steps_done > seq2bin_search_threshold)
	{
		//... start a binary search
		found = getLastLowerBoundLowerThan_BinSearch(val);
	}
	
	// Return the result
	return found;
}

#endif //__SORTEDINTERVALVECTOR__
